#
#  @file   AmdResult.py
#  
#  Description:
#
#  Type your description here
#
#  Copyright (C) 2007 Paraxip Technologies.
#  All Rights Reserved.
#

import paraxip.sipua.user.cpa.CpaResult as CpaResult

#
# getAmdResult - Mapping from internal CPA result to Amd CPA results
#

#  Base Code  Result
#  0          No Detection
#  1          Answering Machine
#  2          Call Forward No Answer
#  3          Call Forward Busy
#  4          Fax machine
#  5          Modem 
#  6          Positive Voice
#  7          Silence Detected
#  8          Special Information Tone 

#  Base Code  Secondary Code  Result
#  8          200             SIT - No Circuit
#  8          201             SIT - Reorder
#  8          202             SIT - Vacant Code
#  8          203             SIT - Intercept
#  8          204             SIT - Ineffective Other
#  6          210             Cadence Break
#  7          211             No Ringback
#  0          212             Unknown
  
# @returns tuple of (base_code-int, base_result-str, secondary_code-int=None, 
#                    secondary_result-str=None)
def getAmdResult(in_szCpaResult=None,
    in_uiCalleeSipErrorCode=None,
    in_uiCpaSipErrorCode=None,
    in_uiRejectCallerResponseCode=None,
    in_bPreConnectTimeoutExpired=False):
  __pychecker__ = 'maxreturns=0 unusednames=t'

  # The logic for returning the Amd result is this following:
  # - if there is a CPA result then use it
  # - if there is a callee SIP error code and the reject caller response
  #   code is already defined then use the reject caller response code
  # - if there is no callee SIP error code and no cpa result then:
  #   - return "Unkown" if CPA SIP error code is defined (error with the CPA 
  #     endpoint)
  #   - return "Silence Detected" if reject caller response code is 
  #     603 (the callee hanged up)
  #   - return "Unknown"

  if (in_szCpaResult is not None):
    # Post-connect results
    if (in_szCpaResult == CpaResult.CpaResult.Human):
      return (6, 'Positive Voice', 210, 'Cadence Break')
    elif (in_szCpaResult == CpaResult.CpaResult.AnsweringMachine):
      return (1, 'Answering Machine', None, None)
    elif (in_szCpaResult == CpaResult.CpaResult.Fax):
      return (4, 'Fax machine', None, None)
    
    #
    # @tag The interface has changed, CpaResult.CpaResult.SIT_{NC, IC, VC, RO}
    #      do not exist anymore at the Cpa level; they will never be received.
    #      All the SITs will become reported as "204, 'SIT - Ineffective Other'".
    #
    
    # Pre-connect results
    elif (in_szCpaResult == CpaResult.CpaResult.Busy):
      return (3, 'Call Forward Busy', None, None)
    elif (in_szCpaResult == CpaResult.CpaResult.ReOrder):
      return (8, 'Special Information Tone', 201, 'SIT - Reorder')
    elif (in_szCpaResult == CpaResult.CpaResult.SIT_NC):
      return (8, 'Special Information Tone', 200, 'SIT - No Circuit')
    elif (in_szCpaResult == CpaResult.CpaResult.SIT_IC):
      return (8, 'Special Information Tone', 203, 'SIT - Intercept')
    elif (in_szCpaResult == CpaResult.CpaResult.SIT_VC):
      return (8, 'Special Information Tone', 202, 'SIT - Vacant Code')
    elif (in_szCpaResult == CpaResult.CpaResult.SIT_RO):
      return (8, 'Special Information Tone', 201, 'SIT - Reorder')
    elif (in_szCpaResult == CpaResult.CpaResult.SIT):
    # Generic SIT used in UK
      return (8, 'Special Information Tone', 204, 'SIT - Ineffective Other')
    elif (in_szCpaResult == CpaResult.CpaResult.SIT_Unknown):
      return (8, 'Special Information Tone', 204, 'SIT - Ineffective Other')

  if ((in_uiCalleeSipErrorCode is not None) and
      (in_uiRejectCallerResponseCode is not None)):
    if ((in_uiRejectCallerResponseCode >= 400) and
        (in_uiRejectCallerResponseCode < 500)):
      if (in_uiRejectCallerResponseCode == 404): # Not Found. Permanent situation
        return (8, 'Special Information Tone', 202, 'SIT - Vacant Code')
      elif (in_uiRejectCallerResponseCode == 408): # Request Timeout
        return (2, 'Call Forward No Answer', None, None)
      elif (in_uiRejectCallerResponseCode == 480): # Temporary Unavailable
        return (8, 'Special Information Tone', 201, 'SIT - Reorder')
      elif (in_uiRejectCallerResponseCode == 486): # Busy Here
        return (3, 'Call Forward Busy', None, None)
      else:
        return (0, 'No Detection', 212, 'Unknown')

    elif ((in_uiRejectCallerResponseCode >= 500) and
        (in_uiRejectCallerResponseCode < 600)):
      return (0, 'No Detection', 212, 'Unknown')

    elif ((in_uiRejectCallerResponseCode >= 600) and
        (in_uiRejectCallerResponseCode < 700)):
      if (in_uiRejectCallerResponseCode == 600): # Busy Everywhere
        return (8, 'Special Information Tone', 200, 'SIT - No Circuit')
      else:
        return (0, 'No Detection', 212, 'Unknown')

  if (in_uiCpaSipErrorCode is not None):
    return (0, 'No Detection', 212, 'Unknown') # Internal error/server congestion
  elif ((in_uiRejectCallerResponseCode is not None) and
      (in_uiRejectCallerResponseCode == 603)):
    # The callee hanged up
    return (7, 'Silence Detected', 211, 'No Ringback')
  elif (in_bPreConnectTimeoutExpired):
    return (2, 'Call Forward No Answer', None, None)
  else:
    return (0, 'No Detection', 212, 'Unknown')

# @input tuple of (base_code-int, base_result-str, secondary_code-int=None, 
#                    secondary_result-str=None)
def getStringFromAmdResult(in_amdResult):
  # HTTP content newlines are carriage return then line feed ("\r\n")
  if(in_amdResult[2] is None):
    return '%d %s\r\n' % (in_amdResult[0], in_amdResult[1]) 
  else:
    return '%d %s\r\n%d %s\r\n' % in_amdResult

def test_module():
  res = getAmdResult(CpaResult.CpaResult.Human)
  assert res[0] == 6
  
  getStringFromAmdResult(res)


## !!MIRROR FILE, DO NOT MODIFY!!
