#
#  @file   ParameterInfo.py
#  @author Sebastien Trottier
#  date    Thu May 28 2008
#  
#  Description:
#
#  Copyright (C) 2008 Paraxip Technologies.
#  All Rights Reserved.
#

import paraxip.infra.PyLoggingFunctions as PyLoggingFunctions
from netborderPyInfra import *

# Force declaration of class attributes in __init__ method
__pychecker__ = "initattr"

fileScopeLogger = Logger("netborder.ParameterInfo.py")

# Info realted to a parameter
# type: Python 'type' instance
# defaultValue: any type
# mandatory: bool
# description: str
# deprecated: bool
# aliases: sequence of str - parameter names that are synonym to this parameter

class ParameterInfo:
  def __init__(self, type, defaultValue=None, mandatory=True, 
      description='', deprecated=False, aliases=None):
    __pychecker__ = 'unusednames=t'
    t = TraceScope(fileScopeLogger, "ParameterInfo.__init__")
    
    self._type = type
    self._defaultValue = defaultValue
    self._isMandatory = mandatory
    self._description = description
    self._isDeprecated = deprecated
    self._aliases = aliases

class ParameterInfoDB:

  def __init__(self):
    # key: param name, value: ParameterInfo
    self._paramInfoDict = {}
    # TODO name for ParameterInfoDB

  # Impl. the membership test operators ("in" and "not in")
  def __contains__(self, paramName):
    return paramName in self._paramInfoDict

  def addParameterInfoForName(self, name, type, defaultValue=None, 
      mandatory=True, description='', deprecated=False, aliases=None):
    __pychecker__ = 'unusednames=t'
    t = TraceScope(fileScopeLogger, "ParameterInfoDB.addParameterInfoForName")

    assert name not in self._paramInfoDict

    self._paramInfoDict[name] = ParameterInfo(type, defaultValue, mandatory,
        description, deprecated, aliases)

  # Returns a config with default parameters for this ParameterInfoDB
  def getDefaultConfig(self):
    __pychecker__ = 'unusednames=t'
    t = TraceScope(fileScopeLogger, "ParameterInfoDB.getDefaultConfig")

    defaultConfig = {}
    for paramName, paramInfo in self._paramInfoDict.items():
      if paramInfo._defaultValue is not None:
        defaultConfig[paramName] = paramInfo._defaultValue
    return defaultConfig

  # Check config for:
  # 1. Missing mandatory parameters
  # 2. Unknown parameters
  # 3. Parameters of invalid type
  # 4. Deprecated parameters (warning)
  #
  # fileName is used only in logging statements
  #
  # returns False on error

  def validateConfig(self, config, fileName):
    __pychecker__ = 'unusednames=t'
    t = TraceScope(fileScopeLogger, "ParameterInfoDB.validateConfig")
    # 1. Missing mandatory parameters
    for paramName, paramInfo in self._paramInfoDict.items():
      if paramInfo._isMandatory and paramName not in config:  #!!
          PyLoggingFunctions.log_error(fileScopeLogger, 
              'Parameter \'%s\' missing in file \'%s\'' % (paramName, fileName))
          return False

    for paramName, paramValue in config.items():
      # 2. Unknown parameters
      # TODO say for which 'Call' this parameter is unknown, LevinsteinDistance
      if paramName not in self._paramInfoDict:
        PyLoggingFunctions.log_error(fileScopeLogger, 'Parameter \'%s\' in file \'%s\' is unknown' % (paramName, fileName))
        return False
      # 3. Parameters of invalid type
      expectedType = self._paramInfoDict[paramName]._type
      if paramValue.__class__ != expectedType:
        # Bug 1429: upgrade param value to expected type if possible
        # int (integers) can be upgraded to float
        if expectedType == float and paramValue.__class__ == int:
          config[paramName] = float(paramValue)
        # All parameter values can be upgraded to str (string)
        elif expectedType == str:
          config[paramName] = str(paramValue)
        else:
          PyLoggingFunctions.log_error(fileScopeLogger, 'Parameter \'%s\' '\
              'in file \'%s\' is not of valid type. '\
              'Expected type: %s, provided type: %s' % (paramName, 
                fileName, expectedType.__name__, paramValue.__class__.__name__))
          return False
      # 4. Deprecated parameters (warning)
      if self._paramInfoDict[paramName]._isDeprecated:
        PyLoggingFunctions.log_warn(fileScopeLogger, 'Parameter \'%s\' in '\
          'file \'%s\' is deprecated. '\
          'Description: %s' % (paramName, fileName, \
            self._paramInfoDict[paramName]._description))

    # Everything's OK
    return True

  # From parameters of in_config, returns a config of parameter aliases 
  #   if they are present in provided info DB
  #
  # in_config follows the param info DB of self, the returned config the param 
  #   info DB of in_parameterInfoDB
  #
  # Ex for a config with {'a':b, 'c':d}: if parameter 'a' has aliases a1 and a2,
  #   if a2 is defined in in_parameterInfoDB but not a1 and, and parameter 'c' 
  #   has no alias, this function will return {'a2':b}
  #
  def getAliasesConfig(self, in_config, in_parameterInfoDB):
    __pychecker__ = 'unusednames=t'
    t = TraceScope(fileScopeLogger, "ParameterInfoDB.getAliasesConfig")

    out_config = {}
    for paramName, paramValue in in_config.items():

      PyLoggingFunctions.log_debug(fileScopeLogger, \
          'getAliasesConfig: considering parameter \'%s\'' % (paramName))

      if paramName in self._paramInfoDict:
        aliases = self._paramInfoDict[paramName]._aliases
        if aliases is not None:

          PyLoggingFunctions.log_debug(fileScopeLogger, \
              'getAliasesConfig: known parameter with aliases')

          for aliasName in aliases:

            PyLoggingFunctions.log_debug(fileScopeLogger, \
                'getAliasesConfig: considering alias \'%s\'' % (aliasName))

            # "in" operator below invokes "__contains__"
            if aliasName in in_parameterInfoDB:

              PyLoggingFunctions.log_debug(fileScopeLogger, \
                  'getAliasesConfig: adding parameter \'%s\' with value %s' % \
                  (aliasName, str(paramValue)))

              out_config[aliasName] = paramValue

    return out_config 

def test_module():
  # Define an parameterInfoDB
  parameterInfoDB = ParameterInfoDB()
  parameterInfoDB.addParameterInfoForName(
      'mandatoryStrParamWithDefaultValue',
      type=str,
      mandatory=True,
      defaultValue='defaultValue')
  parameterInfoDB.addParameterInfoForName(
      'mandatoryStrParamWithoutDefaultValue',
      type=str,
      mandatory=True)
  parameterInfoDB.addParameterInfoForName(
      'optionalStrParam',
      type=str,
      mandatory=False)
  parameterInfoDB.addParameterInfoForName(
      'deprecatedStrParam',
      type=str,
      mandatory=False,
      deprecated=True)
  parameterInfoDB.addParameterInfoForName(
      'floatParam',
      type=float,
      mandatory=False)
  
  # Test 1: generate default config from DB
  defaultConfig = parameterInfoDB.getDefaultConfig()
  # Should be only one item in default Config
  # -> mandatoryStrParamWithDefaultValue
  # with a value of 'defaultValue'
  assert len(defaultConfig.items()) == 1
  assert 'mandatoryStrParamWithDefaultValue' in defaultConfig
  assert defaultConfig['mandatoryStrParamWithDefaultValue'] == 'defaultValue'

  # Test 2: attempt to validate a config with missing mandatory parameters
  #   (logs an ERROR)
  assert not parameterInfoDB.validateConfig({}, 'nofile')

  # Test 3: attempt to validate a config with unknown parameters
  test3Config = dict(defaultConfig) # Take a copy of the default config
  test3Config['mandatoryStrParamWithoutDefaultValue'] = 'givenValue'
  test3Config['unknownParamName'] = 1
  assert not parameterInfoDB.validateConfig(test3Config, 'nofile')

  # Test 4: attempt to validate a config with param of invalid type
  test4Config = dict(defaultConfig) # Take a copy of the default config
  test4Config['mandatoryStrParamWithoutDefaultValue'] = 'givenValue'
  test4Config['floatParam'] = 'stringValue'
  assert not parameterInfoDB.validateConfig(test4Config, 'nofile')

  # Test 5: validate a config with a int value for a param expected 
  #         to be a float
  test5Config = dict(defaultConfig) # Take a copy of the default config
  test5Config['mandatoryStrParamWithoutDefaultValue'] = 'givenValue'
  test5Config['floatParam'] = 1
  assert parameterInfoDB.validateConfig(test5Config, 'nofile')
  # Param should have been converted to a float
  assert test5Config['floatParam'].__class__ == float

  # Test 6: validate a config with a non-string value for a param expected 
  #         to be a string 
  test6Config = dict(defaultConfig) # Take a copy of the default config
  test6Config['mandatoryStrParamWithoutDefaultValue'] = 1
  assert parameterInfoDB.validateConfig(test6Config, 'nofile')
  # Param should have been converted to string
  assert test6Config['mandatoryStrParamWithoutDefaultValue'].__class__ == str

  # Test 7: validate a config that is valid without need for param value 
  #         upgrades
  test7Config = dict(defaultConfig) # Take a copy of the default config
  test7Config['mandatoryStrParamWithoutDefaultValue'] = 'givenValue'
  test7Config['floatParam'] = 1.0
  assert parameterInfoDB.validateConfig(test7Config, 'nofile')

  # Test 8: getAliasesConfig test
  # new param DB
  parameterInfoDB2 = ParameterInfoDB()
  parameterInfoDB2.addParameterInfoForName('a1', type=bool)
  parameterInfoDB2.addParameterInfoForName('a2', type=bool)
  
  # additional param in original DB
  parameterInfoDB.addParameterInfoForName(\
      'a',
      type=bool,
      aliases=('a1', 'a2', 'nonexistent'))
  
  orig_config = {'a':True, 'c':False}
  new_config = parameterInfoDB.getAliasesConfig(orig_config, parameterInfoDB2)
  # expecting {'a1':True, 'a2':True}
  assert len(new_config) == 2
  assert new_config['a1']
  assert new_config['a2']

# runtests
if __name__ == "__main__":
  test_module()

## !!MIRROR FILE, DO NOT MODIFY!!
