#
#  @file   SipCallEngineStateInOutCommon.py
#  @author Sebastien Trottier
#  date    Wed Aug 8 2007
#  
#  Description: States common to In and Out legs
#
#  Copyright (C) 2007 Paraxip Technologies.
#  All Rights Reserved.
#

from paraxip.sipua.internal.SipCallEngineState import *
from paraxip.sipua.internal.SipCallEngineEvent import *
import paraxip.infra.PyLoggingFunctions as PyLoggingFunctions
from netborderPyInfra import *
from SipCallEngineExtending import *

# Force declaration of class attributes in __init__ method
__pychecker__ = "initattr"

#
# Connected state
#

class SipCallEngineConnectedStateInOut(SipCallEngineStateInOut):
  def __init__(self, in_stateMachine):
    SipCallEngineStateInOut.__init__(self, in_stateMachine)
    
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "SipCallEngineConnectedStateInOut.__init__")
  
  def getName(self):
    return "CONNECTED"
  
  def processEvent(self, in_event, out_pystrNewStateName):
    __pychecker__ = 'maxreturns=0'
    
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::processEvent %s",
              self.__class__.__name__, in_event.__class__.__name__)
    
    out_pystrNewStateName.set(self.getName())
    
    if (isinstance(in_event, CallToDialogProvideSdpAnswerEvent)):
      # Providing an SDP answer following offer (ex. we sent an INVITE without SDP
      # and we just got the offer with the 183/200 response (resip::onOffer())
      self._stateMachine.sipLeg.provideSdpAnswer(in_event.getSdpAnswer())
      
      # Stay in the same state
      return True

    elif (isinstance(in_event, CallToDialogReInviteEvent)):
      sdpOffer = in_event.getSdpOffer()
      if (sdpOffer is not None):
        self._stateMachine.sipLeg.reInvite(sdpOffer)
      else:
        self._stateMachine.sipLeg.reInviteWithoutSDP()
      
      out_pystrNewStateName.set("REINVITING")
      return True
    
    elif (isinstance(in_event, CallToDialogSendInfoMsgEvent)):
      self._stateMachine.sipLeg.sendInfoMsg(
          in_event.getMimeType(),
          in_event.getMimeSubType(),
          in_event.getBodyText())

      # Stay in the same state
      return True
    
    elif (isinstance(in_event, CallToDialogHangupEvent)):
      self._stateMachine.sipLeg.hangup()
      
      out_pystrNewStateName.set("DISCONNECTING")
      return True
    
    elif (isinstance(in_event, ResipSdpOfferedEvent)):
      # Received a re-INVITE with SDP
      
      # IMPORTANT boost::python converter for resip::SDP is defined in 
      #   PyResipStack extension module (src: sip/resiprocate/resip/stack)
      #   MUST be imported before calling getSdpContentsCopy()
      __pychecker__ = 'unusednames=PyResipStack'
      import PyResipStack
      toCallEvent = DialogToCallSdpOfferedEvent(self._stateMachine.dialogId,
                                                self._stateMachine.callId,
                                                in_event.getSipMessageCopy(),
                                                in_event.getSdpContentsCopy())
      if (not self.enqueueEventForCall(toCallEvent)):
        return False
      
      out_pystrNewStateName.set("REINVITED")
      return True
    
    elif (isinstance(in_event, ResipSdpOfferRequiredEvent)):
      # Received a re-INVITE without SDP
      
      toCallEvent = DialogToCallSdpOfferRequiredEvent(
          self._stateMachine.dialogId,
          self._stateMachine.callId,
          in_event.getSipMessageCopy())
      if (not self.enqueueEventForCall(toCallEvent)):
        return False
      
      out_pystrNewStateName.set("REINVITED")
      return True
    
    elif (isinstance(in_event, ResipInfoReceivedEvent)):
      toCallEvent = DialogToCallInfoEvent(self._stateMachine.dialogId,
                                          self._stateMachine.callId,
                                          in_event.getSipMessageCopy())
      if (not self.enqueueEventForCall(toCallEvent)):
        return False

      # Stay in the same state
      return True
    
    elif (isinstance(in_event, ResipInfoSentSucceededEvent)):
      toCallEvent = DialogToCallInfoSentSucceededEvent(
        self._stateMachine.dialogId,
        self._stateMachine.callId,
        in_event.getSipMessageCopy())
      if (not self.enqueueEventForCall(toCallEvent)):
        return False

      # Stay in the same state
      return True
    
    elif (isinstance(in_event, ResipInfoSentFailedEvent)):
      toCallEvent = DialogToCallInfoSentFailedEvent(
        self._stateMachine.dialogId,
        self._stateMachine.callId,
        in_event.getSipMessageCopy())
      if (not self.enqueueEventForCall(toCallEvent)):
        return False

      # Stay in the same state
      return True
    
    elif (isinstance(in_event, ResipConnectedConfirmedEvent)):
      # Stay in the same state
      return True

    elif (isinstance(in_event, ResipTerminatedEvent)):
      toCallEvent = DialogToCallTerminatedEvent(self._stateMachine.dialogId,
                                                self._stateMachine.callId)
      if (not self.enqueueEventForCall(toCallEvent)):
        return False
      
      out_pystrNewStateName.set("DISCONNECTED")
      return True    
    
    elif (isinstance(in_event, ResipAckNotReceivedEvent)):
      toCallEvent = DialogToCallAckNotReceivedEvent(self._stateMachine.dialogId,
                                                    self._stateMachine.callId)
      if (not self.enqueueEventForCall(toCallEvent)):
        return False
      
      # Stay in the same state
      return True

    elif (isinstance(in_event, ResipNotifyReceivedEvent)):
      toCallEvent = DialogToCallNotifyEvent(self._stateMachine.dialogId,
                                            self._stateMachine.callId,
                                            in_event.getEventName())
      if (not self.enqueueEventForCall(toCallEvent)):
        return False

      # Stay in the same state
      return True

    else:
      PyLoggingFunctions.log_warn(self.getLogger(),
               "%s::processEvent received an invalid event %s",
               self.__class__.__name__, in_event.__class__.__name__)
    
    return True

#
# Reinviting state
#

class SipCallEngineReinvitingStateInOut(SipCallEngineStateInOut):
  def __init__(self, in_stateMachine):
    SipCallEngineStateInOut.__init__(self, in_stateMachine)
    
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "SipCallEngineReinvitingStateInOut.__init__")
  
  def getName(self):
    return "REINVITING"
  
  def processEvent(self, in_event, out_pystrNewStateName):
    __pychecker__ = 'maxreturns=0'
    
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::processEvent %s",
              self.__class__.__name__, in_event.__class__.__name__)
    
    out_pystrNewStateName.set(self.getName())
    
    if (isinstance(in_event, CallToDialogProvideSdpAnswerEvent)):
      # Providing an SDP answer following offer (ex. we sent an re-INVITE without SDP
      # and we just got the offer with the 183/200 response (resip::onOffer())
      self._stateMachine.sipLeg.provideSdpAnswer(in_event.getSdpAnswer())
      
      out_pystrNewStateName.set("CONNECTED")
      return True

    elif (isinstance(in_event, CallToDialogSendInfoMsgEvent)):
      # Send info message with passed-in args
      self._stateMachine.sipLeg.sendInfoMsg(
        in_event.getMimeType(),
        in_event.getMimeSubType(),
        in_event.getBodyText())
      
      # Stay in the same state
      return True
    
    elif (isinstance(in_event, CallToDialogHangupEvent)):
      self._stateMachine.sipLeg.hangup()
      
      out_pystrNewStateName.set("DISCONNECTING")
      return True
    
    elif (isinstance(in_event, ResipSdpOfferedEvent)):
      # Received an offer for the re-INVITE without SDP we sent
      
      # IMPORTANT boost::python converter for resip::SDP is defined in 
      #   PyResipStack extension module (src: sip/resiprocate/resip/stack)
      #   MUST be imported before calling getSdpContentsCopy()
      __pychecker__ = 'unusednames=PyResipStack'
      import PyResipStack
      toCallEvent = DialogToCallSdpOfferedEvent(self._stateMachine.dialogId,
                                                self._stateMachine.callId,
                                                in_event.getSipMessageCopy(),
                                                in_event.getSdpContentsCopy())
      if (not self.enqueueEventForCall(toCallEvent)):
        return False
      
      # Stay in the same state
      return True
    
    elif (isinstance(in_event, ResipSdpAnsweredEvent)):
      # Received an answer for the re-INVITE with SDP we sent
      
      # IMPORTANT boost::python converter for resip::SDP is defined in 
      #   PyResipStack extension module (src: sip/resiprocate/resip/stack)
      #   MUST be imported before calling getSdpContentsCopy()
      __pychecker__ = 'unusednames=PyResipStack'
      import PyResipStack
      toCallEvent = DialogToCallSdpAnsweredEvent(self._stateMachine.dialogId,
                                                 self._stateMachine.callId,
                                                 in_event.getSipMessageCopy(),
                                                 in_event.getSdpContentsCopy())
      if (not self.enqueueEventForCall(toCallEvent)):
        return False
      
      out_pystrNewStateName.set("CONNECTED")
      return True
    
    elif (isinstance(in_event, ResipFailedEvent)):
      # Our SDP was rejected
      
      toCallEvent = DialogToCallFailedEvent(self._stateMachine.dialogId,
                                            self._stateMachine.callId,
                                            in_event.getSipMessageCopy())
      if (not self.enqueueEventForCall(toCallEvent)):
        return False
      
      out_pystrNewStateName.set("CONNECTED") # re-INVITE is rejected but we are still connected
      return True
    
    elif (isinstance(in_event, ResipInfoReceivedEvent)):
      toCallEvent = DialogToCallInfoEvent(self._stateMachine.dialogId,
                                          self._stateMachine.callId,
                                          in_event.getSipMessageCopy())
      if (not self.enqueueEventForCall(toCallEvent)):
        return False

      # Stay in the same state
      return True
    
    elif (isinstance(in_event, ResipInfoSentSucceededEvent)):
      toCallEvent = DialogToCallInfoSentSucceededEvent(
        self._stateMachine.dialogId,
        self._stateMachine.callId,
        in_event.getSipMessageCopy())
      if (not self.enqueueEventForCall(toCallEvent)):
        return False

      # Stay in the same state
      return True
    
    elif (isinstance(in_event, ResipInfoSentFailedEvent)):
      toCallEvent = DialogToCallInfoSentFailedEvent(
        self._stateMachine.dialogId,
        self._stateMachine.callId,
        in_event.getSipMessageCopy())
      if (not self.enqueueEventForCall(toCallEvent)):
        return False

      # Stay in the same state
      return True
    
    elif (isinstance(in_event, ResipTerminatedEvent)):
      toCallEvent = DialogToCallTerminatedEvent(self._stateMachine.dialogId,
                                                self._stateMachine.callId)
      if (not self.enqueueEventForCall(toCallEvent)):
        return False
      
      out_pystrNewStateName.set("DISCONNECTED")
      return True
    
    else:
      PyLoggingFunctions.log_warn(self.getLogger(),
               "%s::processEvent received an invalid event %s",
               self.__class__.__name__, in_event.__class__.__name__)
    
    return True

#
# Reinvited state
#

class SipCallEngineReinvitedStateInOut(SipCallEngineStateInOut):
  def __init__(self, in_stateMachine):
    SipCallEngineStateInOut.__init__(self, in_stateMachine)
    
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "SipCallEngineReinvitedStateInOut.__init__")
  
  def getName(self):
    return "REINVITED"
  
  def processEvent(self, in_event, out_pystrNewStateName):
    __pychecker__ = 'maxreturns=0'
    
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::processEvent %s",
              self.__class__.__name__, in_event.__class__.__name__)
    
    out_pystrNewStateName.set(self.getName())
    
    if (isinstance(in_event, CallToDialogProvideSdpOfferEvent)):
      # Providing an SDP offer in response to a re-INVITE without SDP
      self._stateMachine.sipLeg.provideSdpOffer(in_event.getSdpOffer())
      
      # Stay in the same state
      return True

    elif (isinstance(in_event, CallToDialogProvideSdpAnswerEvent)):
      # Providing an SDP answer in response to a re-INVITE with SDP
      self._stateMachine.sipLeg.provideSdpAnswer(in_event.getSdpAnswer())
      
      out_pystrNewStateName.set("CONNECTED") # Go back to connected state
      return True

    elif (isinstance(in_event, CallToDialogSendInfoMsgEvent)):
      # Send info message with passed-in args
      self._stateMachine.sipLeg.sendInfoMsg(
        in_event.getMimeType(),
        in_event.getMimeSubType(),
        in_event.getBodyText())
      
      # Stay in the same state
      return True
    
    elif (isinstance(in_event, CallToDialogRejectEvent)):
      # Reject the re-INVITE
      self._stateMachine.sipLeg.reject(in_event.getStatusCode(),
                                       in_event.getSipMessage(),
                                       in_event.getReason())

      out_pystrNewStateName.set("CONNECTED") # Go back to connected state
      return True
    
    elif (isinstance(in_event, CallToDialogHangupEvent)):
      self._stateMachine.sipLeg.hangup()
      
      out_pystrNewStateName.set("DISCONNECTING")
      return True
    
    elif (isinstance(in_event, ResipSdpAnsweredEvent)):
      # Received an answer for the offer we sent in response to a
      # re-INVITE without SDP received
      
      # IMPORTANT boost::python converter for resip::SDP is defined in 
      #   PyResipStack extension module (src: sip/resiprocate/resip/stack)
      #   MUST be imported before calling getSdpContentsCopy()
      __pychecker__ = 'unusednames=PyResipStack'
      import PyResipStack
      toCallEvent = DialogToCallSdpAnsweredEvent(self._stateMachine.dialogId,
                                                 self._stateMachine.callId,
                                                 in_event.getSipMessageCopy(),
                                                 in_event.getSdpContentsCopy())
      if (not self.enqueueEventForCall(toCallEvent)):
        return False
      
      out_pystrNewStateName.set("CONNECTED") # Go back to connected state
      return True
    
    elif (isinstance(in_event, ResipAckNotReceivedEvent)):
      toCallEvent = DialogToCallAckNotReceivedEvent(self._stateMachine.dialogId,
                                                    self._stateMachine.callId)
      if (not self.enqueueEventForCall(toCallEvent)):
        return False
      
      out_pystrNewStateName.set("CONNECTED") # Go back to connected state
      return True

    elif (isinstance(in_event, ResipInfoReceivedEvent)):
      toCallEvent = DialogToCallInfoEvent(self._stateMachine.dialogId,
                                          self._stateMachine.callId,
                                          in_event.getSipMessageCopy())
      if (not self.enqueueEventForCall(toCallEvent)):
        return False

      # Stay in the same state
      return True
    
    elif (isinstance(in_event, ResipInfoSentSucceededEvent)):
      toCallEvent = DialogToCallInfoSentSucceededEvent(
        self._stateMachine.dialogId,
        self._stateMachine.callId,
        in_event.getSipMessageCopy())
      if (not self.enqueueEventForCall(toCallEvent)):
        return False

      # Stay in the same state
      return True
    
    elif (isinstance(in_event, ResipInfoSentFailedEvent)):
      toCallEvent = DialogToCallInfoSentFailedEvent(
        self._stateMachine.dialogId,
        self._stateMachine.callId,
        in_event.getSipMessageCopy())
      if (not self.enqueueEventForCall(toCallEvent)):
        return False

      # Stay in the same state
      return True
    
    elif (isinstance(in_event, ResipTerminatedEvent)):
      toCallEvent = DialogToCallTerminatedEvent(self._stateMachine.dialogId,
                                                self._stateMachine.callId)
      if (not self.enqueueEventForCall(toCallEvent)):
        return False
      
      out_pystrNewStateName.set("DISCONNECTED")
      return True
    
    else:
      PyLoggingFunctions.log_warn(self.getLogger(),
               "%s::processEvent received an invalid event %s",
               self.__class__.__name__, in_event.__class__.__name__)
    
    return True

#
# Disconnecting state
#

class SipCallEngineDisconnectingStateInOut(SipCallEngineStateInOut):
  def __init__(self, in_stateMachine):
    SipCallEngineStateInOut.__init__(self, in_stateMachine)
    
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "SipCallEngineDisconnectingStateInOut.__init__")
  
  def getName(self):
    return "DISCONNECTING"
  
  def processEvent(self, in_event, out_pystrNewStateName):
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::processEvent %s",
              self.__class__.__name__, in_event.__class__.__name__)
    
    out_pystrNewStateName.set(self.getName())
    
    if (isinstance(in_event, ResipTerminatedEvent)):
      toCallEvent = DialogToCallTerminatedEvent(self._stateMachine.dialogId,
                                                self._stateMachine.callId)
      if (not self.enqueueEventForCall(toCallEvent)):
        return False
      
      out_pystrNewStateName.set("DISCONNECTED")
      return True

    elif (isinstance(in_event, CallToDialogSendInfoMsgEvent)):
      # Too late to send the INFO as requested: ignore
      PyLoggingFunctions.log_info(self.getLogger(),
          "SipCallEngineDisconnectingStateInOut.processEvent: Discarding "\
              "request to send INFO message as connection is tearing down")
      # Stay in same state
      return True

    elif (isinstance(in_event, ResipInfoReceivedEvent)):
      # Too late to handle received INFO (cannot answer as dialog is tearing down):
      #   ignore and stay in same state (bug 1018)
      return True

    elif (isinstance(in_event, ResipInfoSentSucceededEvent)):
      # Too late ignore and stay in same state
      return True

    elif (isinstance(in_event, ResipSdpAnsweredEvent)):
      # Too late, ignore and stay in same state
      return True
    
    else:
      PyLoggingFunctions.log_warn(self.getLogger(),
               "%s::processEvent received an invalid event %s",
               self.__class__.__name__, in_event.__class__.__name__)
    
    return True

#
# Disconnected state
#

class SipCallEngineDisconnectedStateInOut(SipCallEngineStateInOut):
  def __init__(self, in_stateMachine):
    SipCallEngineStateInOut.__init__(self, in_stateMachine)
    
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(),
               "SipCallEngineDisconnectedStateInOut.__init__")
  
  def getName(self):
    return "DISCONNECTED"
  
  def processEvent(self, in_event, out_pystrNewStateName):
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::processEvent %s",
              self.__class__.__name__, in_event.__class__.__name__)
    
    out_pystrNewStateName.set(self.getName())
    
    if (isinstance(in_event, CallToDialogSendInfoMsgEvent)):
      # Too late to send the INFO as requested: ignore
      PyLoggingFunctions.log_info(self.getLogger(),
          "SipCallEngineDisconnectedStateInOut.processEvent: Discarding "\
              "request to send INFO message as connection is over")
      # Stay in same state
      return True

    elif (isinstance(in_event, CallToDialogRingEvent)):
      # Too late, UAC cancelled
      PyLoggingFunctions.log_info(self.getLogger(),
          "%s::processEvent: Discarding "\
          "request to send 180 response as UAC connection is over. Event: %s",\
          self.__class__.__name__, in_event.__class__.__name__)
      # Stay in same state
      return True

    elif (isinstance(in_event, CallToDialogAnswerEvent)):
      # Too late, UAC cancelled
      PyLoggingFunctions.log_info(self.getLogger(),
          "%s::processEvent: Discarding "\
          "request to send SDP answer as UAC connection is over. Event: %s",\
          self.__class__.__name__, in_event.__class__.__name__)
      # Stay in same state
      return True

    elif (isinstance(in_event, CallToDialogHangupEvent)):
      # Too late, UAC cancelled
      PyLoggingFunctions.log_info(self.getLogger(),
          "%s::processEvent: Discarding "\
          "request to hang up as UAC connection is over. Event: %s",\
          self.__class__.__name__, in_event.__class__.__name__)
      # Stay in same state
      return True

    elif (isinstance(in_event, CallToDialogRejectEvent)):
      # Too late, UAC cancelled
      PyLoggingFunctions.log_info(self.getLogger(),
          "%s::processEvent: Discarding "\
          "request to reject call as UAC connection is over. Event: %s",\
          self.__class__.__name__, in_event.__class__.__name__)
      # Stay in same state
      return True


    else:
      PyLoggingFunctions.log_warn(self.getLogger(),
               "%s::processEvent received an invalid event %s",
               self.__class__.__name__, in_event.__class__.__name__)
    
    return True

## !!MIRROR FILE, DO NOT MODIFY!!
