#!netborder-python

"""Utilities to load a config (dict) for a Call or App class from a file via global 
config param netborder.sip.ua.python.callsConfigURL
"""

import os

import netborderPyInfra
import paraxip.infra.PyLoggingFunctions as PyLoggingFunctions

from SipCallEngineExtending import CallConfigLoader

# Configuration cache, key is <appCls name>.<callCls name>
configCache = {}

# Public functions
""" 
getConfigURLForApp: figure URL of config file for a given App cls

1. global param netborder.sip.ua.python.callsConfigURL
2. if not defined use passed-in NetBorderApp class name to build the 
  config override URL (installer configuration)

Returns None on error
""" 
def getConfigURLForApp(appCls):

  configURL = netborderPyInfra.PyGlobalConfig.getParameter(
      CallConfigLoader.getCallsConfigURLGlobalParamName())

  if configURL is None:
    # Global param not set, look for file named:
    #   $PARAXIP_INSTALLDIR/config/<appCls name>.call-properties
    installDir = os.getenv('PARAXIP_INSTALLDIR')
    if installDir is None:
      PyLoggingFunctions.log_error(netborderPyInfra.Logger(\
          'netborder.internal.getCallConfigForAppAndCall'), 
          'Installation error: PARAXIP_INSTALLDIR environment variable not set')
      return None
    
    configURL = installDir + '/config/' + appCls.__name__ + \
        '.call-properties'

  return configURL

""" 
getCallConfigForAppAndCall: retrieve the Call config (dict) from the 
  App instance and Call cls.

1. If config is in the cache, re-use it, else
2.   Get default config for callCls
3.  Update the config from the App parameters that have aliases that match the Call 
      parameters
4.  If netborder.sip.ua.python.callsConfigURL global property is set,
      use it as config overrides URL
5.  If not, use passed-in NetBorderApp class name to build the config override URL
      (installer configuration)
6.  Using config override URL and class name (as prefix for key names), apply 
      overrides to default config
7.  Cache and return resulting config

@args

app:      NetBorderApp instance
callCls:  Call class that is going to be used for call

Returns None on error

"""
def getCallConfigForAppAndCall(app, callCls):
  
  configCacheKey = app.__class__.__name__ + '.' + callCls.__name__

  # Return cached config if entry exists
  if configCacheKey in configCache:
    return configCache[configCacheKey]

  config = callCls.getParameterInfoDB().getDefaultConfig()

  # Update the Call config from the App parameters that have aliases that match 
  #   the valid Call parameters
  config.update(\
      app.getParameterInfoDB().getAliasesConfig(\
        app.getConfig(),
        callCls.getParameterInfoDB()))

  callsConfigURL = getConfigURLForApp(app.__class__)
  
  # Update call config using call config URL
  if not CallConfigLoader.updateConfigForCall(callsConfigURL, callCls.__name__, 
      config):
    # Error was logged
    return None
  
  # cache resulting config
  configCache[configCacheKey] = config

  return config

"""
getAppConfig: retrieve the App config (dict) from the App cls.

returns None on error
"""
def getAppConfig(appCls):

  config = appCls.getParameterInfoDB().getDefaultConfig()
  configURL = getConfigURLForApp(appCls)
  
  # Update default config using config URL
  if not CallConfigLoader.updateConfigForApp(configURL, config):
    # Error was logged
    return None

  return config
## !!MIRROR FILE, DO NOT MODIFY!!
