import paraxip.infra.PyLoggingFunctions as PyLoggingFunctions
from paraxip.infra.Assert import isLogAssert

from paraxip.sipua.user.b2bua.B2buaStates import ProgrammableState
from JSR309APICommonPythonModules import SuccessMediaEvent, SDPEvent, \
    ErrorMediaEvent
from jsr309 import JSR309Event
from SipCallEngineExtending import StopEvent
from paraxip.sipua.internal.SipCallEngineEvent import DialogToCallConnectedEvent, \
    DialogToCallSdpAnsweredEvent
from paraxip.sipua.internal.SipCallEngineState import SipCallEngineState
import PyResipStack as resip

# NC: JSR309 NetworkConnection

class ProgrammableStateWithNC(ProgrammableState):
  def __init__(self, in_stateMachine, successStateName, 
      failureStateName, disconnectStateName):
    ProgrammableState.__init__(self, in_stateMachine, successStateName, 
      failureStateName, disconnectStateName)

    self._name = 'PROGRAMMABLE_STATE_WITH_NC'

    # JSR309 NetworkConnection
    self._networkConnection = None

  # Must be set before transitioning to state
  def setNetworkConnection(self, in_networkConnection):
    self._networkConnection = in_networkConnection  

  # SipCallEngineState override
  def warnUnexpectedEvent(self, in_strStateName, in_event):
    if isinstance(in_event, JSR309Event):
      PyLoggingFunctions.log_warn(self.getLogger(),
             "%s received an unexpected JSR309Event %s",
             in_strStateName, in_event.mediaEvent)
    else:
      SipCallEngineState.warnUnexpectedEvent(self, in_strStateName, in_event)

  def onStateExit(self, in_event, in_strDestStateName):
    __pychecker__ = 'unusednames=in_event,in_strDestStateName'
    # Release network connection upon exit
    # TODO revisit need for this once JSR309 shutdown is reworked
    self._networkConnection = None

# Base class for states also dealing with SIP messaging
# TODO Defined LegStatus dict in SipEngineCallSM to handle disconnects also

class ProgrammableStateWithNCAndDialogId(ProgrammableStateWithNC):
  def __init__(self, in_stateMachine, successStateName, 
      failureStateName, disconnectStateName):
    ProgrammableStateWithNC.__init__(self, in_stateMachine, successStateName, 
      failureStateName, disconnectStateName)

    self._name = 'PROGRAMMABLE_STATE_WITH_NC_AND_DIALOG_ID'

    # ID of dialog on each Offer/Answer is performed
    self._dialogId = None # INPUT

  def setDialogId(self, in_dialogId):
    self._dialogId = in_dialogId

  def entryAction(self, in_event, out_pystrNewStateName):
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::entryAction %s",
              self.__class__.__name__, in_event.__class__.__name__)
    
    # dialogId must be set
    if (isLogAssert(self.getLogger(), self._dialogId is not None)):
      return
    
    out_pystrNewStateName.set(self.getName())

# AnsweringSDPState state
# Gets an SDP answer for a given remote offer from a NetworkConnection,
#  performs provided onAnswerReady() action, and waits for 
# DialogToCallConnectedEvent coming from dialog with id 'dialogId'
# DialogToCallTerminatedEvent not handled see TODO above

class AnsweringSDPState(ProgrammableStateWithNCAndDialogId):
  def __init__(self, in_stateMachine, successStateName, 
      failureStateName, disconnectStateName):
    ProgrammableStateWithNCAndDialogId.__init__(self, in_stateMachine,
        successStateName, failureStateName, disconnectStateName)
    
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), 
      'AnsweringSDPState.__init__')

    self._name = 'ANSWERING_SDP'
    
    self._sdpOffer = None # INPUT

    self._setOfferAccepted = False
  
  # Must be set before transitioning to state
  def setSDPOffer(self, in_sdpOffer):
    self._sdpOffer = in_sdpOffer

  # I/F: override to handle SDP answer.
  # Override expected to trigger generation of Connected event, if not, set 
  #   out_pystrNewStateName to next state
  # Returning False triggers transition to failure state
  def onAnswerReady(self, in_sdpAnswer, out_pystrNewStateName):
    isLogAssert(self.getLogger(), false) # MUST override

  def entryAction(self, in_event, out_pystrNewStateName):
    ProgrammableStateWithNCAndDialogId.entryAction(self, in_event,
        out_pystrNewStateName)
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::entryAction %s",
              self.__class__.__name__, in_event.__class__.__name__)
    
    if (isLogAssert(self.getLogger(), self._sdpOffer is not None)):
      return
    
    self._setOfferAccepted = False
    self._networkConnection.setSDPOffer(self._sdpOffer)

  def processEvent(self, in_event, out_pystrNewStateName):
    __pychecker__ = 'maxreturns=0'
    
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::processEvent %s",
              self.__class__.__name__, in_event.__class__.__name__)
    
    out_pystrNewStateName.set(self.getName())
    
    if isinstance(in_event, JSR309Event):
      
      if isinstance(in_event.mediaEvent, SuccessMediaEvent):
        # setSDPOffer success: get the local SDP answer
        self._setOfferAccepted = True
        self._networkConnection.generateSDPAnswer()
        # Stay in same state
        return True
      
      elif isinstance(in_event.mediaEvent, SDPEvent) and \
          in_event.mediaEvent.getType() == "eSDP_ANSWER":
        # generateSDPAnswer success: use the answer and wait for Connected
        # (unless the handler set out_pystrNewStateName)
        if not self.onAnswerReady(in_event.mediaEvent.getSdpContentsCopy(),
            out_pystrNewStateName):
          out_pystrNewStateName.set(self._failureStateName)
        return True

      elif isinstance(in_event.mediaEvent, ErrorMediaEvent):
        if self._setOfferAccepted:
          PyLoggingFunctions.log_warn(self.getLogger(), 'Error generating local SDP answer: %s',
            in_event.mediaEvent) 
        else:
          PyLoggingFunctions.log_warn(self.getLogger(), 'Error: rejecting remote SDP offer %s',
            in_event.mediaEvent) 

        out_pystrNewStateName.set(self._failureStateName)
        return True

    elif isinstance(in_event, DialogToCallConnectedEvent) and \
          (in_event.getDialogId() == self._dialogId):
      # onAnswerReady() made peer to connect, transition to success state
      out_pystrNewStateName.set(self._successStateName)
      return True

    else:
      return ProgrammableState.processEvent(self, in_event, 
          out_pystrNewStateName)

  def onStateExit(self, in_event, in_strDestStateName):
    ProgrammableStateWithNC.onStateExit(self, in_event, in_strDestStateName)
    self._sdpOffer = None


# Generates SDP offer using NC, performs provided onOfferReady action, 
#   waits for DialogToCallConnectedEvent coming from dialog with id 'dialogId'
#   and set SDP answer on NC
# TODO handle ack not received timeout

class OfferingSDPState(ProgrammableStateWithNCAndDialogId):
  def __init__(self, in_stateMachine, successStateName, 
      failureStateName, disconnectStateName):
    ProgrammableStateWithNCAndDialogId.__init__(self, in_stateMachine,
        successStateName, failureStateName, disconnectStateName)
    
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), 
      'AnsweringSDPStateState.__init__')

    self._name = 'OFFERING_SDP'

    self._offerGenerated = False

  # I/F: override to handle SDP offer: override expected to trigger generation
  #   of DialogToCallConnectedEvent
  # Returning False triggers transition to failure state
  def onOfferReady(self, in_sdpOffer):
    isLogAssert(self.getLogger(), false) # MUST override

  def entryAction(self, in_event, out_pystrNewStateName):
    ProgrammableStateWithNCAndDialogId.entryAction(self, in_event, 
        out_pystrNewStateName)
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::entryAction %s",
              self.__class__.__name__, in_event.__class__.__name__)
    
    self._offerGenerated = False
    self._networkConnection.generateSDPOffer()

  def processEvent(self, in_event, out_pystrNewStateName):
    __pychecker__ = 'maxreturns=0'
    
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::processEvent %s",
              self.__class__.__name__, in_event.__class__.__name__)
    
    out_pystrNewStateName.set(self.getName())
    
    if isinstance(in_event, JSR309Event):
      
      if isinstance(in_event.mediaEvent, SDPEvent) and \
          in_event.mediaEvent.getType() == "eSDP_OFFER":
        # generateSDPOffer success: invoke onOfferReady() on subclass
        self._offerGenerated = True
        if not self.onOfferReady(in_event.mediaEvent.getSdpContentsCopy()):
          out_pystrNewStateName.set(self._failureStateName)
        return True

      elif isinstance(in_event.mediaEvent, SuccessMediaEvent):
        # setSDPAnswer success: we're done, move to success state
        out_pystrNewStateName.set(self._successStateName)
        return True

      elif isinstance(in_event.mediaEvent, ErrorMediaEvent):
        if self._offerGenerated:
          PyLoggingFunctions.log_warn(self.getLogger(), 'Error rejecting remote SDP answer: %s',
            in_event.mediaEvent) 
        else:
          PyLoggingFunctions.log_warn(self.getLogger(), 'Error generating local SDP offer: %s',
            in_event.mediaEvent) 

        out_pystrNewStateName.set(self._failureStateName)
        return True

    elif isinstance(in_event, DialogToCallSdpAnsweredEvent) and \
          (in_event.getDialogId() == self._dialogId):
      # onOfferReady() generated remote to answer with SDP, relay to NC
      self._networkConnection.setSDPAnswer(in_event.getSdpContents())
      return True # No state change

    elif isinstance(in_event, DialogToCallConnectedEvent) and \
          (in_event.getDialogId() == self._dialogId):
      # Offer/answer exchange done
      return True # No state change

    else:
      return ProgrammableState.processEvent(self, in_event, 
          out_pystrNewStateName)

###################################################################
# States that handle standard one-dialog SDP offer/answer scenarios

# UAS scenarios

# Handles incoming INVITE with SDP using an NC
class HandlingInviteWithSDPState(AnsweringSDPState):

  def __init__(self, in_stateMachine, successStateName, 
      failureStateName, disconnectStateName):
    AnsweringSDPState.__init__(self, in_stateMachine,
        successStateName, failureStateName, disconnectStateName)
    self._name = 'HANDLING_INVITE_WITH_SDP'

  # Override
  def onAnswerReady(self, in_sdpAnswer, out_pystrNewStateName):
    # Answer dialog and do not change state (Connected event expected)
    return self._stateMachine.answerInDialog(self._dialogId,
                                             in_sdpAnswer,
                                             False) #isSdpOffer

# Handles incoming INVITE without SDP using an NC
class HandlingInviteWithoutSDPState(OfferingSDPState):

  def __init__(self, in_stateMachine, successStateName, 
      failureStateName, disconnectStateName):
    OfferingSDPState.__init__(self, in_stateMachine,
        successStateName, failureStateName, disconnectStateName)
    self._name = 'HANDLING_INVITE_WITHOUT_SDP'

  # Override
  def onOfferReady(self, in_sdpOffer):
    return self._stateMachine.answerInDialog(self._dialogId,
                                             in_sdpOffer,
                                             True) #isSdpOffer

# Handles incoming re-INVITE with SDP using an NC
class HandlingReInviteWithSDPState(AnsweringSDPState):

  def __init__(self, in_stateMachine, successStateName, 
      failureStateName, disconnectStateName):
    AnsweringSDPState.__init__(self, in_stateMachine,
        successStateName, failureStateName, disconnectStateName)
    self._name = 'HANDLING_REINVITE_WITH_SDP'

  # Override
  def onAnswerReady(self, in_sdpAnswer, out_pystrNewStateName):
    if self._stateMachine.provideSdpAnswerToDialog(self._dialogId,
        in_sdpAnswer):
      # No event is generated by provideAnswer -> transition directly to success
      #   state
      out_pystrNewStateName.set(self._successStateName)
      return True
    else:
      return False


# Handles incoming re-INVITE without SDP using an NC
class HandlingReInviteWithoutSDPState(OfferingSDPState):

  def __init__(self, in_stateMachine, successStateName, 
      failureStateName, disconnectStateName):
    OfferingSDPState.__init__(self, in_stateMachine,
        successStateName, failureStateName, disconnectStateName)
    self._name = 'HANDLING_REINVITE_WITHOUT_SDP'

  # Override
  def onOfferReady(self, in_sdpOffer):
    return self._stateMachine.provideSdpOfferToDialog(self._dialogId,
                                                      in_sdpOffer)

# UAC scenarios

# Puts session on hold following RFC 3264 section 8.4
class PuttingSessionOnHoldRFC3264State(OfferingSDPState):

  def __init__(self, in_stateMachine, successStateName, 
      failureStateName, disconnectStateName):
    OfferingSDPState.__init__(self, in_stateMachine,
        successStateName, failureStateName, disconnectStateName)
    self._name = 'PUTTING_SESSION_ON_HOLD_RFC3264'

  # Override
  def onOfferReady(self, in_sdpOffer):
    # Modify offer from NC to be sendonly
    # NOTE Assuming only one medium (m= line)
    medium = in_sdpOffer.session().media()[0]

    # Verify that the offer medium is sendrecv or sendonly
    if isLogAssert(self.getLogger(), \
        medium.exists(resip.Data('sendrecv')) or \
        medium.exists(resip.Data('sendonly')) or \
        (not medium.exists(resip.Data('recvonly')) and \
          not medium.exists(resip.Data('inactive')))):
      return False

    # Set the medium to sendonly
    medium.clearAttribute(resip.Data('sendrecv'))
    medium.addAttribute(resip.Data('sendonly'))

    # Re-invite the remote peer with the modified sdp
    return self._stateMachine.reInviteDialog(self._dialogId, in_sdpOffer)

## !!MIRROR FILE, DO NOT MODIFY!!
