#
#  Copyright (C) 2010 Paraxip Technologies.
#  All Rights Reserved.
#

from netborderPyInfra import TraceScope, Logger, UUID, LogLevel
import paraxip.infra.PyLoggingFunctions as PyLoggingFunctions
from paraxip.infra.Assert import logAssert, isLogAssert

from SipCallEngineExtending import PyTaskObserver, PyApplicationEP, \
    PyGetCallIdResult, ProxyOnDialogCreatedEvent, ProxyOnCallCreatedEvent, \
    ProxyOnExistingEPCreatedEvent, ProxyOnDialogDestroyedEvent, \
    ProxyOnCallDestroyedEvent, ProxyOnExistingEPDestroyedEvent, StopEvent, \
    KillEvent, InitialEvent

from paraxip.sipua.internal.CallManager import CallManager
from paraxip.sipua.internal.EventProcessorManager import EventProcessorManager
from paraxip.sipua.internal.ObjectPool import ObjectPool
from paraxip.sipua.internal.BaseStateMachine import BaseStateMachine
from paraxip.sipua.internal.SipCallEngineSMInOutBaseClass import \
    SipCallEngineSMInOutBaseClass
from paraxip.sipua.internal.SipCallEngineSMIn import SipCallEngineSMIn
from paraxip.sipua.internal.SipCallEngineSMOut import SipCallEngineSMOut
from paraxip.sipua.internal.SipCallEngineSMCall import SipCallEngineSMCall
from paraxip.sipua.internal.SipCallEngineEvent import SipCallEngineEvent
import paraxip.sipua.internal.call_config_loader as call_config_loader

from paraxip.sipua.user.reject.RejectCall import RejectCall

# Force declaration of class attributes in __init__ method
__pychecker__ = "initattr"

fileScopeLogger = Logger("netborder.NetBorderApp.py")

class TaskWorkEvent(SipCallEngineEvent):
  def __init__(self, in_szTaskName):
    SipCallEngineEvent.__init__(self, None)

    __pychecker__ = 'unusednames=t'
    t = TraceScope(fileScopeLogger,
        "TaskWorkEvent.__init__")

    self._strTaskName = in_szTaskName

  def getTaskName(self):
    return self._strTaskName

class ReadyForWorkEvent(TaskWorkEvent):
  def __init__(self, in_szTaskName):
    TaskWorkEvent.__init__(self, in_szTaskName)

    __pychecker__ = 'unusednames=t'
    t = TraceScope(fileScopeLogger,
        "ReadyForWorkEvent.__init__")

class NotReadyForWorkEvent(TaskWorkEvent):
  def __init__(self, in_szTaskName, in_szReason):
    TaskWorkEvent.__init__(self, in_szTaskName)

    __pychecker__ = 'unusednames=t'
    t = TraceScope(fileScopeLogger,
        "NotReadyForWorkEvent.__init__")

    self._strReason = in_szReason

  def getReason(self):
    return self._strReason

class FlowControlTaskObserver(PyTaskObserver):
  def __init__(self, in_globalContext):
    PyTaskObserver.__init__(self)

    __pychecker__ = 'unusednames=t'
    t = TraceScope(fileScopeLogger,
        "FlowControlTaskObserver.__init__")

    self._globalContext = in_globalContext

  def readyForWork(self, in_szTaskName):
    __pychecker__ = 'unusednames=t'
    t = TraceScope(fileScopeLogger, 
        "FlowControlTaskObserver.readyForWork")
    
    # Enqueuing high priority event 
    # (default priority = 0 which is the lowest)
    logAssert(fileScopeLogger,
              self._globalContext.enqueuePythonEvent(
                None,
                ReadyForWorkEvent(in_szTaskName), 100))

  def notReadyForWork(self, in_szTaskName, in_szReason):
    __pychecker__ = 'unusednames=t'
    t = TraceScope(fileScopeLogger,
        "FlowControlTaskObserver.notReadyForWork")
 
    # Enqueuing high priority event 
    # (default priority = 0 which is the lowest)
    logAssert(fileScopeLogger,
              self._globalContext.enqueuePythonEvent(
                None,
                NotReadyForWorkEvent(in_szTaskName, in_szReason),100))

###############################################################################

class NetBorderApp(PyApplicationEP):
  def __init__(self):
    PyApplicationEP.__init__(self)
    
    self.setLogger(Logger("netborder.application.py"))

    __pychecker__ = 'unusednames=t'
    t = TraceScope(fileScopeLogger, "NetBorderApp.__init__")

    self._refGlobalContext = None

    # Pools of various available SM, indexed by Class name. 
    self._poolAvailableSM = ObjectPool() 

    # Contains active Calls SMs
    self._callManager = CallManager()    

    # Indexes of all active SMs (both Calls and Dialogs)
    self._eventProcessorManager = EventProcessorManager() 

    self._flowControlTaskObserver = None
    self._taskNotReadyForWorkCount = 0

    self._isShutingDown = False

  def onStartup(self, in_refGlobalContext):
    """
      Called on App startup: if overriden:
      * Call base class function
      * log error if returning False as process will exit
    """
    # Keep a reference on the GlobalContext
    self._refGlobalContext = in_refGlobalContext    

    # Configure the flow controller for tasks
    self._flowControlTaskObserver = FlowControlTaskObserver(
      self._refGlobalContext)
    self._refGlobalContext.addSipCallEngineTaskObserver(
      self._flowControlTaskObserver)

    return True

  def onShutdown(self):
    """
    Begin to shuting down the Application.
    This method can be overloaded by derived Application:
    class MyApp(NetBorderApp):
    def onShutdown(self):
      NetBorderApp.onShutdown(self) # Must be called first
      self.doMyWork()               # Do your stuff after
    """
    __pychecker__ = 'unusednames=t'
    t = TraceScope(fileScopeLogger, "NetBorderApp.onShutdown")

    self._isShutingDown = True

    return True

  def onShutdownCompleted(self):
    """
    The shutdown of NetBorderApp is completed.
    This method can be overloaded by derived Application:
    class MyApp(NetBorderApp):
    def onShutdownCompleted(self):
      self.doMyWork()               # Do your stuff first
      NetBorderApp.onShutdownCompleted(self) # Must be called at the end
    """
    __pychecker__ = 'unusednames=t'
    t = TraceScope(fileScopeLogger, 
                                    "NetBorderApp.onShutdownCompleted")

    # Get rid of all SM (will cause of release of all object e.g. proxy if 
    # any where left
    self._poolAvailableSM.clear()

    # Inform the infrastructure that the shutdown has been done successfully
    self._refGlobalContext.shutdownCompleted()

    # Bug 3925: making sure garbage collection is run before we exit from the 
    #  Python framework. Otherwise some PySipMessage objects could still be 
    #  alive in the Embedded Interpreter namespace when the CallDataRegistryImpl 
    #  singleton is detroyed due to circluar references (and CallLoggerS in 
    #  PySipMessage need access to it)
    import gc
    gc.collect()

    return True

  def isCallIdActive(self, in_szCallId):
    return self._callManager.isCallSMExist(in_szCallId)

  def onInviteRequest(self, out_pystrRejectReason):
    """
    Return 0 if accepting the request, otherwise return the SIP reject code
    with the reason in out_pystrRejectReason
    """
    __pychecker__ = 'unusednames=t'
    t = TraceScope(fileScopeLogger,
                                    "NetBorderApp.onInviteRequest")

    if (self._taskNotReadyForWorkCount > 0):
      out_pystrRejectReason.set("Service Unavailable (Server Congestion)")
      return 503 # Service Unavailable

    return 0

  def onInviteGetCallId(self, in_sipMessage):
    """
    Returns the Call Id to use as the Call of the new Incoming Dialog to be
    created. Giving the Call Id of an existing Call adds the new Incoming
    Dialog to it, giving a new Call Id creates a new Call, returning a SIP
    error code to rejects the Call
    """
    __pychecker__ = 'unusednames=t'
    t = TraceScope(fileScopeLogger,
                                    "NetBorderApp.onInviteGetCallId")

    # Check if call-id to use is provided by calling UAC
    strCallIDFromCaller = in_sipMessage.getHeaderValue('NetBorder-Call-Id')
    if(strCallIDFromCaller != ''):
      # Call Id provided by caller
      return PyGetCallIdResult(strCallIDFromCaller)
    else:
      # Create a new SMCall by giving out a new UUID
      return PyGetCallIdResult(UUID().getAsString())

  def onInviteGetCall(self, in_sipMessage):
    """
      Method to be overriden by derived App classes
      Must returns a tuple ([SM class:Class], [config:dict]) to tell which  \
      Call class to run, and with what configuration.
    """
    pass

  def onCallEnd(self, in_event):
    __pychecker__ = 'unusednames=in_event'
    pass

  # When not associated to a call
  def onClientRegistrationSuccess(self, in_event):
    __pychecker__ = 'unusednames=in_event'
    pass

  def onClientRegistrationFailure(self, in_event):
    __pychecker__ = 'unusednames=in_event'
    pass

  # Registrar functionality
  def onRegistrationRequest(self, in_event):
    __pychecker__ = 'unusednames=in_event'
    pass

  def onTimeout(self, in_event):
    __pychecker__ = 'unusednames=in_event'
    pass
  
  def onTimeoutCancelled(self, in_event):
    __pychecker__ = 'unusednames=in_event'
    pass

  # Ex SOAP command
  def onExternalCommand(self, in_event):
    __pychecker__ = 'unusednames=in_event'
    pass

  # Helper
  # @returns Same tupple than onInviteGetCall()
  def rejectCall(self, statusCode, reasonText):
    config = {RejectCall.PARAM_REJECT_STATUS_CODE: statusCode,
              RejectCall.PARAM_REJECT_REASON_TEXT: reasonText}
    return RejectCall, config


  # Helper
  #
  # app:      NetBorderApp instance
  # callCls:  Call class that is going to be used for call
  # @returns Same tupple than onInviteGetCall()

  def updateAndValidateConfigForAppAndCall(self, config, app, callCls):

    # Update call config following App and Call classes
    configForAppAndCall = call_config_loader.getCallConfigForAppAndCall(\
        app, callCls)
    if configForAppAndCall is None:
      return self.rejectCall(500, 'Cannot retrieve configuration'\
              ' for Call \'%s\' of NetBorder Application \'%s\'' % \
              (callCls.__name__, app.__class__.__name__))

    config.update(configForAppAndCall)

    # Verify that current config is valid for call
    fileName = call_config_loader.getConfigURLForApp(app.__class__)
    if not callCls.getParameterInfoDB().validateConfig(config, fileName):
      return self.rejectCall(500, 'Invalid configuration'\
              ' for Call \'%s\' of NetBorder Application \'%s\'' % \
              (callCls.__name__, app.__class__.__name__))

    # All is OK
    return (callCls, config)
  
  def processEvent(self, in_event):
    # TODO cleanup this function
    __pychecker__ = 'maxlines=0 maxreturns=0'

    PyLoggingFunctions.log_debug(self.getLogger(), "%s::processEvent %s",
              self.__class__.__name__, in_event.__class__.__name__)  

    if (isinstance(in_event, ReadyForWorkEvent)):
      if (self._taskNotReadyForWorkCount > 0):        
        self._taskNotReadyForWorkCount = self._taskNotReadyForWorkCount - 1
        if(self._taskNotReadyForWorkCount == 0):
          PyLoggingFunctions.log_warn(self.getLogger(), "System no longer in overload condition: "
                   "will start accepting new calls")                       
      return True

    elif (isinstance(in_event, NotReadyForWorkEvent)):
      self._taskNotReadyForWorkCount = self._taskNotReadyForWorkCount + 1
      PyLoggingFunctions.log_warn(self.getLogger(), "System overload (some thread(s) are not ready"
                 " for work): will be refusing new calls in an attempt to "
                 "reduce system load")
      return True
    
    elif (isinstance(in_event, ProxyOnDialogCreatedEvent)):
      strCallId = in_event.getCallId()
      strDialogId = in_event.getDialogId()

      if (isLogAssert(self.getLogger(), (strCallId is not None) and \
                                        (strDialogId is not None))):
        return False

      # Get the Call SM associated to strCallId, the Call SM could not have
      # been created yet
      callSM = self._callManager.getCallSM(strCallId)
      if (callSM is None):
        # Get a Call SM from the pool
        callSM = self.getCallSM(in_event.getInitialInviteMsgRef())
        if (isLogAssert(self.getLogger(), callSM is not None)):
          return False

        if (not callSM.initialize(self, self._refGlobalContext, strCallId)):
          PyLoggingFunctions.log_warn(self.getLogger(), "Failed to initialize Call State " \
                                     "Machine with id %s", strCallId)
          return False
        callSM.goToInitialState(InitialEvent(callSM.getId()))

        # Save this Call SM
        self._callManager.addCallSM(callSM.getId(), callSM)
        self._eventProcessorManager.addEP(callSM.getId(), callSM)

      # Increment the proxy count on the Call SM
      callSM.incProxyCount()     

      dialogSM = None
      if (in_event.isIncomingDialog()):
        # Get an Incoming Dialog SM from the pool
        dialogSM = self._poolAvailableSM.get(SipCallEngineSMIn)
        if (isLogAssert(self.getLogger(), dialogSM is not None)):
          return False
        callSM.addInDialog(strDialogId, dialogSM)

      else:
        # IMPORTANT: This event only happens upon the reception of the first
        #            SIP message from the callee, this means if the callee is an
        #            unreachable destination then this event will happen
        #            32 seconds after the request to makeOutCall function.

        # Get an Outgoing Dialog SM from the pool
        dialogSM = self._poolAvailableSM.get(SipCallEngineSMOut)
        if (isLogAssert(self.getLogger(), dialogSM is not None)):
          return False
        callSM.addOutDialog(strDialogId, dialogSM)

      if (not dialogSM.initialize(self, self._refGlobalContext,
                                  strCallId, strDialogId)):
        PyLoggingFunctions.log_warn(self.getLogger(), "Failed to initialize Dialog State " \
                                   "Machine with id %s", strDialogId)
        return False
      dialogSM.goToInitialState(InitialEvent(dialogSM.getId()))

      # Save this Event Processor
      self._eventProcessorManager.addEP(dialogSM.getId(), dialogSM)

      # Increment the proxy count on the Dialog SM
      dialogSM.incProxyCount()

      return True

    elif (isinstance(in_event, ProxyOnCallCreatedEvent)):
      strCallId = in_event.getCallId()
      if (isLogAssert(self.getLogger(), strCallId is not None)):
        return False

      # Get the Call SM associated to strCallId, the Call SM could have not
      # been created yet
      callSM = self._callManager.getCallSM(strCallId)
      if (callSM is None):
        # Get a Call SM from the pool
        callSM = self.getCallSM(in_event.getInitialInviteMsgRef())
        if (isLogAssert(self.getLogger(), callSM is not None)):
          return False

        if (not callSM.initialize(self, self._refGlobalContext, strCallId)):
          PyLoggingFunctions.log_warn(self.getLogger(), "Failed to initialize Call State " \
                                     "Machine with id %s", strCallId)
          return False
        callSM.goToInitialState(InitialEvent(callSM.getId()))

        # Save this Call SM
        self._callManager.addCallSM(callSM.getId(), callSM)
        self._eventProcessorManager.addEP(callSM.getId(), callSM)

      # Increment the proxy count on this State Machine
      callSM.incProxyCount()

      return True

    elif (isinstance(in_event, ProxyOnExistingEPCreatedEvent)):
      strDestEPId = in_event.getDestObjectId()
      if (isLogAssert(self.getLogger(), strDestEPId is not None)):
        return False

      # Get the Call SM associated to strCallId
      EP = self._eventProcessorManager.getEP(strDestEPId)
      if (isLogAssert(self.getLogger(), EP is not None)):
        return False

      # Increment the proxy count on this Event Processor
      EP.incProxyCount()

      return True

    elif (isinstance(in_event, ProxyOnDialogDestroyedEvent)):
      strCallId = in_event.getCallId()
      strDialogId = in_event.getDialogId()

      if (isLogAssert(self.getLogger(), (strCallId is not None) and \
                                        (strDialogId is not None))):
        return False

      dialogSM = self._eventProcessorManager.getEP(strDialogId)
      if (isLogAssert(self.getLogger(), dialogSM is not None)):
        return False

      # A Call lives always as long as the Dialog lives
      callSM = self._callManager.getCallSM(strCallId)
      if (isLogAssert(self.getLogger(), callSM is not None)):
        return False

      # There is only one proxy that should point on a Dialog
      if (isLogAssert(self.getLogger(), dialogSM.isUniqueProxy())):
        return False
      dialogSM.decProxyCount()

      # Remove this Dialog from the Event Processor Manager
      self._eventProcessorManager.removeEP(dialogSM.getId())

      # Remove this Dialog from its Call
      if (issubclass(dialogSM.__class__, SipCallEngineSMIn)):
        callSM.removeInDialog(dialogSM.getId())
      elif (issubclass(dialogSM.__class__, SipCallEngineSMOut)):
        callSM.removeOutDialog(dialogSM.getId())
      else:
        logAssert(self.getLogger(), False)

      # Check that the Dialog is in final state
      if (not dialogSM.isInFinalState()):
        PyLoggingFunctions.log_warn(self.getLogger(), "Dialog State Machine with Id %s has been " \
                                   "released without being in final state. " \
                                   "The current state is %s",
                                   dialogSM.getId(),
                                   dialogSM.getCurrentStateName())

      # Reset the Dialog SM
      dialogSM.reset()

      # Put back the Dialog SM into the pool of available State Machine
      self._poolAvailableSM.takeBack(dialogSM)

      # Enqueue a ProxyOnCallDestroyedEvent since a Dialog always contains
      # a virtual proxy on its Call
      self._refGlobalContext.enqueuePythonEvent(
        None,
        ProxyOnCallDestroyedEvent(strCallId))

      return True

    elif (isinstance(in_event, ProxyOnCallDestroyedEvent)):
      strCallId = in_event.getCallId()
      if (isLogAssert(self.getLogger(), strCallId is not None)):
        return False

      callSM = self._callManager.getCallSM(strCallId)
      if (isLogAssert(self.getLogger(), callSM is not None)):
        return False

      callSM.decProxyCount()
      # Check if the SM has reached a final state and that is no proxy 
      # still using it. If it is the case, then we can safely release
      # the state machine (i.e. put it back in the object pool)
      # See releaseCallSM() method for more details on why we need 
      # to do this.
      if (callSM.getProxyCount() == 0 and 
          callSM.isInFinalState()):
        PyLoggingFunctions.log_debug(self.getLogger(), "Call %s SM has no proxy and is in a "
                  "final state. Releasing it", strCallId)
        self.releaseCallSM(callSM)      

      return True

    elif (isinstance(in_event, ProxyOnExistingEPDestroyedEvent)):
      strDestEPId = in_event.getDestObjectId()
      if (isLogAssert(self.getLogger(), strDestEPId is not None)):
        return False

      # Get the Call SM associated to strCallId
      EP = self._eventProcessorManager.getEP(strDestEPId)
      if (isLogAssert(self.getLogger(), EP is not None)):
        return False

      if (issubclass(EP.__class__, SipCallEngineSMCall)):
        # Enqueue a ProxyOnCallDestroyedEvent
        self._refGlobalContext.enqueuePythonEvent(
          None,
          ProxyOnCallDestroyedEvent(strDestEPId))

      elif (issubclass(EP.__class__, BaseStateMachine)):
        # Check if the proxy was the only one pointing on this Call SM
        if (EP.isUniqueProxy()):
          EP.decProxyCount()

          # Remove this Call SM as active one
          self._eventProcessorManager.removeEP(EP.getId())

          # Check that the State Machine is in final state
          if (not EP.isInFinalState()):
            PyLoggingFunctions.log_warn(self.getLogger(), "State Machine with Id %s has been " \
                                       "released without being in final state. " \
                                       "The current state is %s",
                                       EP.getId(),
                                       EP.getCurrentStateName())

          # Reset the Call SM
          EP.reset()

          # Put back the Call SM into the pool of available State Machine
          self._poolAvailableSM.takeBack(EP)

        else:
          EP.decProxyCount()

      else:
        # Check if the proxy was the only one pointing on this Call SM
        if (EP.isUniqueProxy()):
          EP.decProxyCount()

          # Remove this Call SM as active one
          self._eventProcessorManager.removeEP(EP.getId())

          # Put back the Call SM into the pool of available State Machine
          self._poolAvailableSM.takeBack(EP)

        else:
          EP.decProxyCount()

      return True

    elif (isinstance(in_event, StopEvent)):
      # Log at warning the list of State Machine still active on the reception
      # of Stop event
      self.logActiveStateMachineOnStopEvent()

      # Process StopEvent on all active Calls
      itCall = self._callManager.getIterator()
      for call in itCall:
        call.processPythonEvent(in_event)
      itCall.close()

      # Begin to shuting down 
      bShutdownResult = self.onShutdown()

      if (self._callManager.countCallSM() < 1):
        # There is no active Calls to wait for
        bShutdownCompletedResult = self.onShutdownCompleted()
        return (bShutdownResult and bShutdownCompletedResult)

    elif (isinstance(in_event, KillEvent)):
      # Log at error the list of State Machine still active on the reception
      # of Kill event
      self.logActiveStateMachineOnKillEvent()

      # Process KillEvent on all active Calls
      it = self._callManager.getIterator()
      for call in it:
        call.processPythonEvent(in_event)
      it.close()

      return True

    else:      
      strDestEPId = in_event.getDestObjectId()
      if (isLogAssert(self.getLogger(), strDestEPId is not None)):
        return False

      # Get the Event Processor associated to unique Id
      EP = self._eventProcessorManager.getEP(strDestEPId)
      if (isLogAssert(self.getLogger(), EP is not None)):
        return False

      # Ask the Event Processor to process this event
      return EP.processPythonEvent(in_event)

  # Returns an instance of a Call SM, based on an incoming INVITE
  # Returns 'None' on error
  def getCallSM(self, in_sipMessage):
    __pychecker__ = 'unusednames=t,in_sipMessage'
    t = TraceScope(fileScopeLogger, "NetBorderApp.getCallSM")
    
    # Get Call Class Name and Config from user-provided function
    (strSMClass, dictConfig) = self.onInviteGetCall(in_sipMessage)
    
    # Retrieve one such call from the pool, and configure it
    callSM = self._poolAvailableSM.get(strSMClass)

    if callSM.configure(dictConfig):
      return callSM
    else:
      # Problem configuring Call instance: reject SIP INVITE Call using built-in 
      # RejectCall Call class
      rejectCallSM = self._poolAvailableSM.get(RejectCall)
      rejectConfig = {RejectCall.PARAM_REJECT_STATUS_CODE: 500,
          RejectCall.PARAM_REJECT_REASON_TEXT: 'Error configuring Call ' +
          strSMClass}
      rejectCallSM.configure(rejectConfig)
      return rejectCallSM

  def getFlowControlTaskObserver(self):
    return self._flowControlTaskObserver

  def logActiveStateMachineOnStopEvent(self):
    uiActiveCalls = self._callManager.countCallSM()
    if (uiActiveCalls > 0):
      PyLoggingFunctions.log_warn(self.getLogger(), "%d Calls was still active"
        " while stopping the Application", uiActiveCalls)
    self.logActiveStateMachine(LogLevel.info)

  def logActiveStateMachineOnKillEvent(self):
    self.logActiveStateMachine(LogLevel.error)

  def logActiveStateMachine(self, in_logLevel):
    itEP = self._callManager.getIterator()
    for ep in itEP:
      if (issubclass(ep.__class__, SipCallEngineSMIn)):
        PyLoggingFunctions.log_warn(self.getLogger(), in_logLevel,
            "An incoming Dialog State Machine is still active. "
            "Call Id = %s, Dialog Id = %s, and current state is %s",
            ep.getCallId(), ep.getDialogId(), ep.getCurrentStateName())
      elif (issubclass(ep.__class__, SipCallEngineSMOut)):
        PyLoggingFunctions.log_warn(self.getLogger(), in_logLevel,
            "An outgoing Dialog State Machine is still active. "
            "Call Id = %s, Dialog Id = %s, and current state is %s",
            ep.getCallId(), ep.getDialogId(), ep.getCurrentStateName())
      elif (issubclass(ep.__class__, SipCallEngineSMCall)):
        PyLoggingFunctions.log_warn(self.getLogger(), in_logLevel,
            "A Call State Machine is still active. "
            "Call Id = %s, and current state is %s",
            ep.getCallId(), ep.getCurrentStateName())
    itEP.close()

  def releaseCallSM (self, in_callSM):
    """ This method releases a call SM by putting it back in its pool.
        The call SM must only be released when NO proxy points to this object 
        AND when the SM is in FINAL state. Here is an example of why this needs
        to be done:
        1. An inbound call is received causing an inbould dialog and its proxy
           to be created.
        2. The SM makes an out call (tru makeOutCall)
        3. The inbound dialog hangs-up causing the inbound proxy to be 
           destroyed. However, this is not a good time to release the SM
           even if the proxy count is null, because a new proxy will be created
           in the soon future (could take up to 32 sec if the remote 
           destination is not reachable). The application knows about this,
           that's why it is not in a final state.
        4. The out dialog is created (following step 2) with its associated
           proxy. The application is notified.
        5. The application will hangup this call causing it to reach a final 
           state and reducing the proxy count to 0 at some point. The SM can
           then be put back into the pool.
    """
    __pychecker__ = 'unusednames=t'
    t = TraceScope(fileScopeLogger,
                                     "NetBorderApp.releaseCallSM")
    # Remove this Call SM as active one
    self._callManager.removeCallSM(in_callSM.getId())
    self._eventProcessorManager.removeEP(in_callSM.getId())

    # Check that the Call is in final state
    if (isLogAssert(self.getLogger(), in_callSM.isInFinalState())):
        return False   

    self._refGlobalContext.releasingCall(in_callSM.getId())

    # Reset the Call SM
    in_callSM.reset()

    # Put back the Call SM into the pool of available State Machines
    self._poolAvailableSM.takeBack(in_callSM)

    if (self._isShutingDown):
      # If there is no more active Call State Machines
      if (self._callManager.countCallSM() < 1):
        # The shutdown is completed
        self.onShutdownCompleted()
## !!MIRROR FILE, DO NOT MODIFY!!
