#
#  @file   B2buaStates.py
#  @author Francis Moreau
#  date    Mon May 22 2007
#  
#  Description:
#
#  Type your description here
#
#  Copyright (C) 2007 Paraxip Technologies.
#  All Rights Reserved.
#

import paraxip.infra.PyLoggingFunctions as PyLoggingFunctions
from netborderPyInfra import TraceScope, PyString

from SipCallEngineExtending import InitialEvent, StopEvent, KillEvent
import PyResipStack
from paraxip.sipua.internal.SipCallEngineState import SipCallEngineState
from paraxip.sipua.internal.SipCallEngineEvent import \
    InDialogToCallCalledEvent, DialogToCallSdpOfferedEvent, \
    DialogToCallSdpOfferRequiredEvent, DialogToCallTerminatedEvent, \
    OutDialogToCallCallingEvent, DialogToCallFailedEvent, \
    OutDialogToCallProvisionalEvent, OutDialogToCallEarlyMediaEvent, \
    DialogToCallSdpAnsweredEvent, DialogToCallInfoEvent, \
    DialogToCallConnectedEvent, DialogToCallAckNotReceivedEvent 

from paraxip.sipua.internal.ConnectionStatus import ConnectionStatus

# Force declaration of class attributes in __init__ method
__pychecker__ = "initattr"

#
# Base class
#

class B2buaState(SipCallEngineState):
  def __init__(self, in_stateMachine):
    SipCallEngineState.__init__(self, in_stateMachine)
    
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "B2buaState.__init__")

# Programmable State (settable name, success, failure, disconnect exit states)
# Handle STOP events (transition to disconnecting state)
# TODO not specific to b2buas, move elsewhere

class ProgrammableState(B2buaState):
  def __init__(self, in_stateMachine, successStateName=None, 
      failureStateName=None, disconnectStateName=None):
    B2buaState.__init__(self, in_stateMachine)

    self._name = None
    self._successStateName = successStateName 
    self._failureStateName = failureStateName 
    self._disconnectStateName = disconnectStateName 
  
  def getName(self):
    return self._name

  def setName(self, in_stateName):
    self._name = in_stateName 

  def setSuccessStateName(self, in_stateName):
    self._successStateName = in_stateName 

  def setFailureStateName(self, in_stateName):
    self._failureStateName = in_stateName 

  def setDisconnectStateName(self, in_stateName):
    self._disconnectStateName = in_stateName 

  def processEvent(self, in_event, out_pystrNewStateName):
    __pychecker__ = 'maxreturns=0'
    
    out_pystrNewStateName.set(self.getName())
    
    if isinstance(in_event, StopEvent):
      out_pystrNewStateName.set(self._disconnectStateName)
      return True
    else:
      self.warnUnexpectedEvent(self.getName(), in_event)
      return True

#
# Idle state
#

class IdleState(B2buaState):
  def __init__(self, in_stateMachine, in_strNextState):
    B2buaState.__init__(self, in_stateMachine)
    
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "IdleState.__init__")
    
    self._strNextState = in_strNextState
  
  def getName(self):
    return "IDLE"
  
  def processEvent(self, in_event, out_pystrNewStateName):
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::processEvent %s",
              self.__class__.__name__, in_event.__class__.__name__)
    
    out_pystrNewStateName.set(self.getName())
    
    if (isinstance(in_event, InitialEvent)):
      return True
    
    elif (isinstance(in_event, InDialogToCallCalledEvent)):
      # Store initial Caller invite msg
      self._stateMachine.callerInitialInvite = in_event.getSipMessage()
      self._stateMachine.callerLegStatus = ConnectionStatus.Connecting

      # By default 'local contact' is set to 'request line'
      # which is good when used as a proxy but not if used as a b2bua
      # (b2bua wants to stay in the loop)
      inDialogSM = self._stateMachine.getInDialog(in_event.getDialogId())
      if (inDialogSM is not None):

        # Use default From header URI as contact for this dialog
        strLocalContact = inDialogSM.sipLeg.getDefaultFromHeaderUri()
        
        if ((strLocalContact is None) or (strLocalContact == "")):
          PyLoggingFunctions.log_error(self.getLogger(), 
                                       "getDefaultFromHeaderUri failed to "
                                       "return LocalContact(s).")
          return False        
        
        # Bug 1729: make sure the transport parameter is provided in the 
        # contact URI of our responses if not the default (UDP)
        strTransport = in_event.getSipMessage().getReceivedTransportName()
        assert strTransport
        if strTransport != 'UDP':
          # We received the request on a transport different than UDP
          strLocalContact += ';transport=%s' % (strTransport) 

        inDialogSM.sipLeg.setLocalContact(strLocalContact)
      
      # Stay in the same state, sdp offer or need for one coming up
      return True
    
    elif (isinstance(in_event, DialogToCallSdpOfferedEvent) and
          (in_event.getDialogId() == self._stateMachine.callerDialogId)): #IN
      self._stateMachine.bInvitedWithSdp = True
      
      # Save the caller SDP offered then dial up the callee
      self._stateMachine.callerSdp = in_event.getSdpContents()
      
      out_pystrNewStateName.set(self._strNextState)
      return True
    
    elif (isinstance(in_event, DialogToCallSdpOfferRequiredEvent)):
      self._stateMachine.bInvitedWithSdp = False
      
      # Received an INVITE without an SDP: dial up the callee
      self._stateMachine.callerSdp = None
      
      out_pystrNewStateName.set(self._strNextState)
      return True

    elif (isinstance(in_event, DialogToCallTerminatedEvent)):
      self._stateMachine.dialogIdInitiatingDisconnect = \
        in_event.getDialogId()

      out_pystrNewStateName.set("DISCONNECTING")
      return True
    
    elif (isinstance(in_event, StopEvent)):
      out_pystrNewStateName.set("DISCONNECTING")
      return True

    else:
      PyLoggingFunctions.log_warn(self.getLogger(),
               "%s::processEvent received an invalid event %s",
               self.__class__.__name__, in_event.__class__.__name__)
    
    return True

#
# DialingCallee state
#

class DialingCalleeState(B2buaState):
  def __init__(self, in_stateMachine):
    B2buaState.__init__(self, in_stateMachine)
    
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "DialingCalleeState.__init__")
  
  def getName(self):
    return "DIALING_CALLEE"
  
  def entryAction(self, in_event, out_pystrNewStateName):
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::entryAction %s",
              self.__class__.__name__, in_event.__class__.__name__)
    
    # Inviting the callee with the caller SDP if present, otherwise invite
    # without SDP (then here callerSdp could be None) 
    if not self._stateMachine.callCallee(self._stateMachine.callerSdp):
      # Error: details already logged
      out_pystrNewStateName.set("DISCONNECTING")
      return
    
    out_pystrNewStateName.set(self.getName())
  
  def processEvent(self, in_event, out_pystrNewStateName):
    __pychecker__ = 'maxreturns=0 maxlines=0'
    
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::processEvent %s",
              self.__class__.__name__, in_event.__class__.__name__)
    
    out_pystrNewStateName.set(self.getName())
    
    if (isinstance(in_event, OutDialogToCallCallingEvent)):
      # Stay in the same state
      return True
    
    elif (isinstance(in_event, DialogToCallFailedEvent) and
          (in_event.getDialogId() == self._stateMachine.calleeDialogId)): #OUT
      self._stateMachine.calleeLegStatus = ConnectionStatus.Disconnecting
      
      self._stateMachine.uiCalleeSipErrorCode = \
        in_event.getSipMessage().getStatusCode()
      
      out_pystrNewStateName.set("DISCONNECTING")
      return True
    
    elif (isinstance(in_event, OutDialogToCallProvisionalEvent)):
      statusCode = in_event.getSipMessage().getStatusCode()
      if (statusCode == 180):
        # Send 180 ringing to the caller
        if (not self._stateMachine.ringInDialog(
              self._stateMachine.callerDialogId)):
          return False
      
      # Stay in the same state
      return True
    
    elif (isinstance(in_event, OutDialogToCallEarlyMediaEvent)):
      # Early SDP from the callee, check if we got an SDP offer from the 
      #   caller yet (ie INVITE with SDP)
      if(self._stateMachine.callerSdp is None):
        # No, callee UAS is not compliant, warn and disconnect
        PyLoggingFunctions.log_warn(self.getLogger(), "Received 183 Session Progress SIP response "\
            "following an INVITE request without an SDP offer, which is not "\
            "compliant to RFC 3261. Disconnecting the call.")
        out_pystrNewStateName.set("DISCONNECTING")
        return True

      # Save the callee SDP answered
      self._stateMachine.calleeSdp = in_event.getSdpContents()
      
      # Send 183 early media to the caller
      if (not self._stateMachine.sendEarlyMediaToInDialog(
            self._stateMachine.callerDialogId,
            self._stateMachine.calleeSdp)):
        return False
      
      out_pystrNewStateName.set("EARLY_MEDIA")
      return True
    
    elif (isinstance(in_event, DialogToCallSdpAnsweredEvent) and
          (in_event.getDialogId() == self._stateMachine.calleeDialogId)): #OUT
      # Callee has been invited with SDP
      # Save the callee SDP answered
      self._stateMachine.calleeSdp = in_event.getSdpContents()
      
      # Stay in the same state since DialogToCallConnectedEvent is coming
      return True
    
    elif (isinstance(in_event, DialogToCallSdpOfferedEvent) and
          (in_event.getDialogId() == self._stateMachine.calleeDialogId)): #OUT
      # Callee has been invited without SDP
      # Save the callee SDP offered
      self._stateMachine.calleeSdp = in_event.getSdpContents()
      
      # Stay in the same state since DialogToCallConnectedEvent is coming
      return True
    
    elif (isinstance(in_event, DialogToCallConnectedEvent)):
      self._stateMachine.calleeLegStatus = ConnectionStatus.Connected
      self._stateMachine.bCalleeAnswered = True
      
      # Answer the caller with the callee SDP
      if (not self._stateMachine.answerInDialog(
            self._stateMachine.callerDialogId,
            self._stateMachine.calleeSdp,
            not self._stateMachine.bInvitedWithSdp, # SDP is an offer is we were
                                                    # invited without SDP
            self._stateMachine.getAnswerMessage())):
        return False
      self._stateMachine.callerLegStatus = ConnectionStatus.Answering
      
      out_pystrNewStateName.set("ANSWERING_CALLER")
      return True
    
    elif (isinstance(in_event, DialogToCallTerminatedEvent)):
      self._stateMachine.dialogIdInitiatingDisconnect = \
        in_event.getDialogId()
      
      out_pystrNewStateName.set("DISCONNECTING")
      return True
    
    elif (isinstance(in_event, StopEvent)):
      out_pystrNewStateName.set("DISCONNECTING")
      return True

    else:
      PyLoggingFunctions.log_warn(self.getLogger(),
               "%s::processEvent received an invalid event %s",
               self.__class__.__name__, in_event.__class__.__name__)
    
    return True

#
# EarlyMedia state
#

class EarlyMediaState(B2buaState):
  def __init__(self, in_stateMachine):
    B2buaState.__init__(self, in_stateMachine)
    
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "EarlyMediaState.__init__")
  
  def getName(self):
    return "EARLY_MEDIA"
  
  def processEvent(self, in_event, out_pystrNewStateName):
    __pychecker__ = 'maxreturns=0 maxlines=0'
    
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::processEvent %s",
              self.__class__.__name__, in_event.__class__.__name__)
    
    out_pystrNewStateName.set(self.getName())
    
    if (isinstance(in_event, DialogToCallFailedEvent) and
        (in_event.getDialogId() == self._stateMachine.calleeDialogId)): #OUT
      self._stateMachine.calleeLegStatus = ConnectionStatus.Disconnecting
      
      self._stateMachine.uiCalleeSipErrorCode = \
        in_event.getSipMessage().getStatusCode()
      
      out_pystrNewStateName.set("DISCONNECTING")
      return True
    
    elif (isinstance(in_event, DialogToCallSdpAnsweredEvent) and
          (in_event.getDialogId() == self._stateMachine.calleeDialogId)): #OUT
      # 200 OK contains an SDP even if 180 or 183 was containing one too
      # Save the callee SDP offered
      self._stateMachine.calleeSdp = in_event.getSdpContents()
      
      # Stay in the same state since DialogToCallConnectedEvent is coming
      return True
    
    elif (isinstance(in_event, DialogToCallConnectedEvent)):
      self._stateMachine.calleeLegStatus = ConnectionStatus.Connected
      self._stateMachine.bCalleeAnswered = True
      
      # Answer the caller with the callee SDP
      if (not self._stateMachine.answerInDialog(
            self._stateMachine.callerDialogId,
            None,
            False, # SDP is answer
            self._stateMachine.getAnswerMessage())):
        return False
      self._stateMachine.callerLegStatus = ConnectionStatus.Answering
      
      out_pystrNewStateName.set("ANSWERING_CALLER")
      return True
    
    elif (isinstance(in_event, DialogToCallTerminatedEvent)):
      self._stateMachine.dialogIdInitiatingDisconnect = \
        in_event.getDialogId()
      
      out_pystrNewStateName.set("DISCONNECTING")
      return True
    
    elif (isinstance(in_event, StopEvent)):
      out_pystrNewStateName.set("DISCONNECTING")
      return True

    else:
      PyLoggingFunctions.log_warn(self.getLogger(),
               "%s::processEvent received an invalid event %s",
               self.__class__.__name__, in_event.__class__.__name__)
    
    return True

#
# AnsweringCaller state
#

class AnsweringCallerState(B2buaState):
  def __init__(self, in_stateMachine):
    B2buaState.__init__(self, in_stateMachine)
    
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "AnsweringCallerState.__init__")
  
  def getName(self):
    return "ANSWERING_CALLER"
  
  def processEvent(self, in_event, out_pystrNewStateName):
    __pychecker__ = 'maxreturns=0 maxlines=0'
    
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::processEvent %s",
              self.__class__.__name__, in_event.__class__.__name__)
    
    out_pystrNewStateName.set(self.getName())
    
    if (isinstance(in_event, DialogToCallConnectedEvent)):
      self._stateMachine.callerLegStatus = ConnectionStatus.Connected
      
      out_pystrNewStateName.set("CALLER_CALLEE_CONNECTED")
      return True
    
    elif (isinstance(in_event, DialogToCallSdpAnsweredEvent) and
          in_event.getDialogId() == self._stateMachine.callerDialogId): #IN
      # Save the caller SDP answered
      self._stateMachine.callerSdp = in_event.getSdpContents()
      
      # Send the caller SDP answer (in ACK) as the callee SDP answer (in ACK)
      # (caller sent INVITE w/o SDP)
      if (not self._stateMachine.provideSdpAnswerToDialog(
            self._stateMachine.calleeDialogId,
            self._stateMachine.callerSdp)):
        return False
      
      # Do not change state: DialogToCallConnectedEvent coming just after
      return True
    
    elif (isinstance(in_event, DialogToCallAckNotReceivedEvent)): #IN
      # Can happen here after answering 200 Ok to an INVITE without SDP, 
      #   see bug 989
      # TODO use mutable logger for this warning
      PyLoggingFunctions.log_warn(self.getLogger(), "The SIP UAC did not answer our " \
        "SIP 200 response with a ACK request as specified by RFC 3261. " \
        "Terminating the call. (State: %s)",self.getName())
      
      out_pystrNewStateName.set("DISCONNECTING")
      return True
    
    elif (isinstance(in_event, DialogToCallTerminatedEvent)):
      self._stateMachine.dialogIdInitiatingDisconnect = \
        in_event.getDialogId()
      
      out_pystrNewStateName.set("DISCONNECTING")
      return True
    
    elif (isinstance(in_event, StopEvent)):
      out_pystrNewStateName.set("DISCONNECTING")
      return True

    else:
      PyLoggingFunctions.log_warn(self.getLogger(),
               "%s::processEvent received an invalid event %s",
               self.__class__.__name__, in_event.__class__.__name__)
    
    return True

#
# CallerCalleeConnected state
#

class CallerCalleeConnectedState(B2buaState):
  def __init__(self, in_stateMachine):
    B2buaState.__init__(self, in_stateMachine)
    
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "CallerCalleeConnectedState.__init__")

  def getName(self):
    return "CALLER_CALLEE_CONNECTED"
  
  def processEvent(self, in_event, out_pystrNewStateName):
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::processEvent %s",
              self.__class__.__name__, in_event.__class__.__name__)
    
    out_pystrNewStateName.set(self.getName())
    
    if (isinstance(in_event, DialogToCallSdpOfferedEvent) or \
        isinstance(in_event, DialogToCallSdpOfferRequiredEvent)):

      # re-INVITE received (with SDP on DialogToCallSdpOfferedEvent, without
      # SDP on DialogToCallSdpOfferRequiredEvent)
      reInvitedState = self._stateMachine.getReInvitedState()

      # The REINVITED state is generic; tell the state what is the peer dialog
      # where the re-INVITE must be relayed
      if (in_event.getDialogId() == self._stateMachine.callerDialogId):
        reInvitedState.setPeerDialogId(self._stateMachine.calleeDialogId)
      else:
        reInvitedState.setPeerDialogId(self._stateMachine.callerDialogId)

      out_pystrNewStateName.set(reInvitedState.getName())
      return True
    
    elif (isinstance(in_event, DialogToCallAckNotReceivedEvent)): #IN
      # Can happen here after answering 200 Ok to an INVITE with SDP and receiving 
      #   DialogToCallConnectedEvent, see bug 989
      # TODO use mutable logger for this warning
      PyLoggingFunctions.log_warn(self.getLogger(), "The SIP UAC did not answer our " \
        "SIP 200 response with a ACK request as specified by RFC 3261. " \
        "Terminating the call. (State: %s)",self.getName())
      
      out_pystrNewStateName.set("DISCONNECTING") 
      return True
    
    elif (isinstance(in_event, DialogToCallInfoEvent)):
      # Stay in the same state
      return True

    elif (isinstance(in_event, DialogToCallTerminatedEvent)):
      self._stateMachine.dialogIdInitiatingDisconnect = \
        in_event.getDialogId()
      
      out_pystrNewStateName.set("DISCONNECTING")
      return True
    
    elif (isinstance(in_event, StopEvent)):
      out_pystrNewStateName.set("DISCONNECTING")
      return True

    else:
      PyLoggingFunctions.log_warn(self.getLogger(),
               "%s::processEvent received an invalid event %s",
               self.__class__.__name__, in_event.__class__.__name__)
    
    return True

#
# Reinvited state
# Re-invite received while streams are connected (SDP were exchanged)

class ReinvitedState(ProgrammableState):
  
  def __init__(self, in_stateMachine, successStateName=None, 
      failureStateName=None, disconnectStateName=None):
    ProgrammableState.__init__(self, in_stateMachine, successStateName, 
      failureStateName, disconnectStateName)
    
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "ReinvitedState.__init__")

    self._name = "REINVITED"

    self._reInvitedDialogId = None
    self._peerDialogId = None

    self._reinvitedSdp = None
    self._peerSdp = None

  def getReinvitedSdp(self):
    return self._reinvitedSdp

  def getPeerSdp(self):
    return self._peerSdp

  def setPeerDialogId(self, in_peerDialogId):
    self._peerDialogId = in_peerDialogId

  # Assumes transtion from DialogToCallSdpOfferedEvent or SDP_OFFER_REQUESTED
  def entryAction(self, in_event, out_pystrNewStateName):
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::entryAction %s",
              self.__class__.__name__, in_event.__class__.__name__)
    
    assert (isinstance(in_event, DialogToCallSdpOfferedEvent) or
            isinstance(in_event, DialogToCallSdpOfferRequiredEvent))

    self._reInvitedDialogId = in_event.getDialogId()

    if (isinstance(in_event, DialogToCallSdpOfferedEvent)):
      self._reinvitedSdp = in_event.getSdpContents()

    # Re-invite peer dialog with SDP received via re-INVITE
    # in_event.getSdpContents() can be None (re-INVITE w/o SDP; handled 
    #  differently by reInvite)
    if (not self._stateMachine.reInviteDialog(
          self._peerDialogId,
          in_event.getSdpContents())):
      return
    
    out_pystrNewStateName.set(self.getName())

  def processEvent(self, in_event, out_pystrNewStateName): #!
    __pychecker__ = 'maxreturns=0'
    
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::processEvent %s",
              self.__class__.__name__, in_event.__class__.__name__)

    if self.getName() == None or \
       self._successStateName == None or \
       self._failureStateName == None or \
       self._disconnectStateName == None:
      PyLoggingFunctions.log_error(self.getLogger(), 
                                  "ReinvitedState not configured properly: "
                                  "self.getName(): %s, _successStateName: %s" 
                                  "_failureStateName: %s, _disconnectStateName: %s" %
                                  (self.getName(),
                                  self._successStateName,
                                  self._failureStateName,
                                  self._disconnectStateName))
      return False
    
    out_pystrNewStateName.set(self.getName())
    
    if (isinstance(in_event, DialogToCallSdpAnsweredEvent) and
        (in_event.getDialogId() == self._peerDialogId)):
      # 200 OK with SDP answer (follows re-INVITE with SDP)
      # Relay the peer SDP answer to re-INVITEd leg
      if (not self._stateMachine.provideSdpAnswerToDialog(
            self._reInvitedDialogId,
            in_event.getSdpContents())):
        return False

      self._peerSdp = in_event.getSdpContents()
      # We're done with re-INVITE w/SDP sequence
      out_pystrNewStateName.set(self._successStateName)
      return True
    
    elif (isinstance(in_event, DialogToCallSdpOfferedEvent) and
          (in_event.getDialogId() == self._peerDialogId)):
      if (not self._stateMachine.provideSdpOfferToDialog(
            self._reInvitedDialogId,
            in_event.getSdpContents())):
        return False

      self._peerSdp = in_event.getSdpContents()
      # Not done yet with re-INVITE w/o SDP sequence
      return True

    elif (isinstance(in_event, DialogToCallSdpAnsweredEvent) and
          (in_event.getDialogId() == self._reInvitedDialogId)):
      # ACK with SDP answer from dialog re-INVITEd w/o SDP
      # Relay the SDP answer to the peer dialog (in ACK)
      if(not self._stateMachine.provideSdpAnswerToDialog(
           self._peerDialogId,
           in_event.getSdpContents())):
        return False

      self._reinvitedSdp = in_event.getSdpContents()
      # We're done with re-INVITE w/o SDP sequence
      out_pystrNewStateName.set(self._successStateName)
      return True

    elif (isinstance(in_event, DialogToCallFailedEvent) and
          (in_event.getDialogId() == self._peerDialogId)):
      # The re-INVITE with the new SDP was refused

      # Rejecting the re-INVITE request
      if (not self._stateMachine.rejectDialog(
            self._reInvitedDialogId,
            in_event.getSipMessage().getStatusCode())):
        return False

      out_pystrNewStateName.set(self._failureStateName)
      return True

    elif (isinstance(in_event, DialogToCallInfoEvent)):
      # Stay in the same state
      return True

    elif isinstance(in_event, DialogToCallAckNotReceivedEvent):
      # Didn't get expected ACK with SDP answer (bug 1398)
      # TODO use mutable logger for this warning
      PyLoggingFunctions.log_warn(self.getLogger(), "The SIP UAC did not answer our " \
        "SIP 200 response with an SDP offer with a ACK request with " \
        "an SDP answer as specified by RFC 3261. Terminating the call. " \
        "(State: %s)", self.getName())

      out_pystrNewStateName.set(self._disconnectStateName)
      return True

    elif (isinstance(in_event, DialogToCallTerminatedEvent)):
      self._stateMachine.dialogIdInitiatingDisconnect = \
        in_event.getDialogId()
      
      out_pystrNewStateName.set(self._disconnectStateName)
      return True
    
    elif (isinstance(in_event, StopEvent)):
      out_pystrNewStateName.set(self._disconnectStateName)
      return True

    else:
      PyLoggingFunctions.log_warn(self.getLogger(),
               "%s::processEvent received an invalid event %s",
               self.__class__.__name__, in_event.__class__.__name__)
    
    return True

#
# Disconnecting state
#

class DisconnectingState(B2buaState):
  def __init__(self, in_stateMachine):
    B2buaState.__init__(self, in_stateMachine)
    
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "DisconnectingState.__init__")
  
  def getName(self):
    return "DISCONNECTING"
  
  def entryAction(self, in_event, out_pystrNewStateName):
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::entryAction %s",
              self.__class__.__name__, in_event.__class__.__name__)
    
    if (self._stateMachine.dialogIdInitiatingDisconnect is not None):
      if (self._stateMachine.dialogIdInitiatingDisconnect == \
          self._stateMachine.callerDialogId):
        self._stateMachine.callerLegStatus = ConnectionStatus.Disconnected
      else:
        self._stateMachine.calleeLegStatus = ConnectionStatus.Disconnected
    
    if (self._stateMachine.callerLegStatus == ConnectionStatus.Disconnected and \
        (self._stateMachine.calleeLegStatus == ConnectionStatus.Unknown or \
         self._stateMachine.calleeLegStatus == ConnectionStatus.Disconnected)):
      out_pystrNewStateName.set("DISCONNECTED")
      return
    
    # Reject the caller if it's connecting
    if (self._stateMachine.callerLegStatus == ConnectionStatus.Connecting):

      PyLoggingFunctions.log_info(self.getLogger(), 'Rejecting caller b2bua')

      rejectResponseCode, rejectSipMessage = self._stateMachine.getRejectMessage()
      if (not self._stateMachine.rejectDialog(
            self._stateMachine.callerDialogId,
            rejectResponseCode,
            rejectSipMessage)):
        return
      self._stateMachine.callerLegStatus = ConnectionStatus.Disconnecting
    
    # Hanging up the caller if it's connected
    elif (self._stateMachine.callerLegStatus == ConnectionStatus.Answering or
          self._stateMachine.callerLegStatus == ConnectionStatus.Connected):
      if (not self._stateMachine.hangupDialog(
            self._stateMachine.callerDialogId)):
        return
      self._stateMachine.callerLegStatus = ConnectionStatus.Disconnecting
    
    # Hanging up the callee if it's connecting or connected
    if (self._stateMachine.calleeLegStatus == ConnectionStatus.Connecting or
        self._stateMachine.calleeLegStatus == ConnectionStatus.Connected):
      calleeSM = self._stateMachine.getOutDialog(self._stateMachine.calleeDialogId)
      if (calleeSM is not None):
        if (not self._stateMachine.hangupDialog(
              self._stateMachine.calleeDialogId)):
          return
        self._stateMachine.calleeLegStatus = ConnectionStatus.Disconnecting
      else:
        # Between a new outgoing call request and the call to addOutDialog
        # function it is possible to have events on another leg, and then
        # here it is possible that getOutDialog fails because addOutDialog
        # function has not been called yet
        self._stateMachine.calleeLegStatus = ConnectionStatus.MustDisconnect

    out_pystrNewStateName.set(self.getName())
  
  def processEvent(self, in_event, out_pystrNewStateName):
    __pychecker__ = 'maxreturns=0'
    
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::processEvent %s",
              self.__class__.__name__, in_event.__class__.__name__)
    
    out_pystrNewStateName.set(self.getName())
    
    if (isinstance(in_event, OutDialogToCallCallingEvent)):
      if (self._stateMachine.calleeLegStatus == ConnectionStatus.MustDisconnect):
        # Hanging up the callee
        if (not self._stateMachine.hangupDialog(
              self._stateMachine.calleeDialogId)):
          return False
        self._stateMachine.calleeLegStatus = ConnectionStatus.Disconnecting
        
      # Stay in the same state; DialogToCallTerminatedEvent will follow
      return True
    
    elif (isinstance(in_event, DialogToCallFailedEvent) and
          (in_event.getDialogId() == self._stateMachine.calleeDialogId)): #OUT
      self._stateMachine.calleeLegStatus = ConnectionStatus.Disconnecting
      
      # Stay in the same state; DialogToCallTerminatedEvent will follow
      return True
    
    elif (isinstance(in_event, DialogToCallSdpAnsweredEvent) and
          (in_event.getDialogId() == self._stateMachine.calleeDialogId)): #OUT
      # Already been cancelled in entryAction function. This event
      # was probably already pending in the queue. Then stay in the same state
      # since DialogToCallTerminatedEvent will follow.
      return True
    
    elif (isinstance(in_event, DialogToCallSdpOfferedEvent) and
          (in_event.getDialogId() == self._stateMachine.calleeDialogId)): #OUT
      # Already been cancelled in entryAction function. This event
      # was probably already pending in the queue. Then stay in the same state
      # since DialogToCallTerminatedEvent will follow.
      return True
    
    elif (isinstance(in_event, OutDialogToCallProvisionalEvent) or \
          isinstance(in_event, DialogToCallConnectedEvent) or \
          isinstance(in_event, OutDialogToCallEarlyMediaEvent)):
      # Out call was already cancelled in entryAction function. This event
      # was probably already pending in the queue. Then stay in the same state
      # since DialogToCallTerminatedEvent will follow.
      return True
    
    elif (isinstance(in_event, DialogToCallInfoEvent)):
      # Stay in the same state
      return True

    elif (isinstance(in_event, DialogToCallTerminatedEvent)):
      if (in_event.getDialogId() == self._stateMachine.callerDialogId):
        self._stateMachine.callerLegStatus = ConnectionStatus.Disconnected
      else:
        self._stateMachine.calleeLegStatus = ConnectionStatus.Disconnected

      if (self._stateMachine.callerLegStatus == ConnectionStatus.Disconnected and \
          (self._stateMachine.calleeLegStatus == ConnectionStatus.Unknown or \
           self._stateMachine.calleeLegStatus == ConnectionStatus.Disconnected)):
        out_pystrNewStateName.set("DISCONNECTED")
      
      return True
    
    elif (isinstance(in_event, StopEvent)):
      # Stay in the same state
      return True

    elif (isinstance(in_event, KillEvent)):
      # Go directly in final state
      out_pystrNewStateName.set("DISCONNECTED")
      return True

    else:
      PyLoggingFunctions.log_warn(self.getLogger(),
               "%s::processEvent received an invalid event %s",
               self.__class__.__name__, in_event.__class__.__name__)
    
    return True

#
# Disconnected state
#

class DisconnectedState(B2buaState):
  def __init__(self, in_stateMachine):
    B2buaState.__init__(self, in_stateMachine)
    
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "DisconnectedState.__init__")
  
  def getName(self):
    return "DISCONNECTED"
  
  def processEvent(self, in_event, out_pystrNewStateName):
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::processEvent %s",
              self.__class__.__name__, in_event.__class__.__name__)
    
    out_pystrNewStateName.set(self.getName())
    
    # No event should be received in this state
    PyLoggingFunctions.log_warn(self.getLogger(),
             "%s::processEvent received an invalid event %s",
             self.__class__.__name__, in_event.__class__.__name__)
    
    return True

#
# PuttingSessionOnHoldRFC2543 state
# Putting session on hold following section B.5 of RFC 2543 
# Used by IBM agent bridge

class PuttingSessionOnHoldRFC2543(ProgrammableState):
  def __init__(self, in_stateMachine):
    ProgrammableState.__init__(self, in_stateMachine)
    
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "PuttingSessionOnHoldRFC2543.__init__")

    self._name = "PUTTING_SESSION_ON_HOLD_RFC2543"

    self._dialogId = None

  def setDialogId(self, in_dialogId):
    self._dialogId = in_dialogId

  def entryAction(self, in_event, out_pystrNewStateName):
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::entryAction %s",
              self.__class__.__name__, in_event.__class__.__name__)

    strSDP = "v=0\r\n" \
             "o=- 1182439127 1182439127 IN IP4 127.0.0.1\r\n" \
             "s=Paraxip RFC2543 Held Session\r\n" \
             "c=IN IP4 0.0.0.0\r\n" \
             "t=0 0\r\n" \
             "m=audio 9000 RTP/AVP 0\r\n" \
             "a=rtpmap:0 PCMU/8000\r\n"
    errStr = PyString()
    rfc2543holdSdpOffer = PyResipStack.SDP()
    if not rfc2543holdSdpOffer.parse(strSDP, errStr):
      PyLoggingFunctions.log_error(self.getLogger(), "failed to parse sdp :\n%s", errStr)
      return

    # Re-invite session with held SDP
    if (not self._stateMachine.reInviteDialog(
          self._dialogId,
          rfc2543holdSdpOffer)):
      return
    
    out_pystrNewStateName.set(self.getName())

  def processEvent(self, in_event, out_pystrNewStateName):
    __pychecker__ = 'maxreturns=0'
    
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::processEvent %s",
              self.__class__.__name__, in_event.__class__.__name__)
    
    out_pystrNewStateName.set(self.getName())
    
    if (isinstance(in_event, DialogToCallSdpAnsweredEvent)):
      # We're done with re-INVITE w/SDP sequence
      out_pystrNewStateName.set(self._successStateName)
      return True    

    elif (isinstance(in_event, DialogToCallFailedEvent)):
      # The re-INVITE with the new SDP was refused
      PyLoggingFunctions.log_warn(self.getLogger(), 
          "PuttingSessionOnHoldRFC2543 RFC2543 hold SDP refused")

      out_pystrNewStateName.set(self._failureStateName)
      return True

    # TODO base class that handles other possible SIP requests (ex. NOTIFY)
    elif (isinstance(in_event, DialogToCallInfoEvent)):
      # Stay in the same state
      return True

    elif (isinstance(in_event, DialogToCallTerminatedEvent)):
      self._stateMachine.dialogIdInitiatingDisconnect = \
        in_event.getDialogId()
      
      out_pystrNewStateName.set(self._disconnectStateName)
      return True
    
    elif (isinstance(in_event, StopEvent)):
      out_pystrNewStateName.set(self._disconnectStateName)
      return True

    else:
      PyLoggingFunctions.log_warn(self.getLogger(),
               "%s::processEvent received an invalid event %s",
               self.__class__.__name__, in_event.__class__.__name__)
    
    return True

def test_module():
  import os
  this_file = os.path.split(__file__)[1]
  print this_file + ' test start'
  
  from paraxip.sipua.user.cpa import CpaCall
  
  rvs = ReinvitedState(CpaCall.CpaCall())
  # Expect a 'ReinvitedState not configured properly' error
  pe_rv = rvs.processEvent(DialogToCallTerminatedEvent("Dialog_ID", "DontKnow"), PyString("NewStateName"))
  assert not pe_rv
  
  print this_file + ' test end'
  
if __name__ == '__main__':
  test_module()
## !!MIRROR FILE, DO NOT MODIFY!!
