
from paraxip.sipua.user.cpa import CpaResult

class CallEventException(Exception):
  pass

class CallEvent:
  """
  A CallEvent is any event that happens in a Call which will be needed in
  the future to determine what will be logged and the call flow. 
  """
  
  def __init__(self):
    raise CallEventException("CallEvent should not be instantiated (ABS).")
  
  """
  This is a way to implement isinstance(...): Python isinstance() will
  not work with different import path.
  (see http://mail.python.org/pipermail/python-list/2002-June/781964.html)
  
  One can use:
    hasattr(class, '_Trait_isCallEvent')
  instead of:
    isinstance(class, CallEvent)
  """
  _Trait_isCallEvent = None
  
  def __eq__(self, other):
    if type(self) == type(other):
      if self.__dict__ == other.__dict__:
        return True
    return False

class CallCpaResult(CallEvent):
  """
  A 'typed' CpaResult. Contains one of the string defined in CpaResult.py.
  Be careful not to confuse this class with the _strings_ in CpaResult.py.
  """
  
  "Use: hasattr(class, '_Trait_isCallCpaResult')"
  _Trait_isCallCpaResult = None
  
  def __init__(self, in_strCpaResult):
    raise CallEventException("CallCpaResult should not be instantiated (ABS).")
  
  def __str__(self):
    return "%s: %s" % (self.__class__.__name__, self.getStr())


class Cpa_Generic(CallCpaResult):
  """Can be used for the Cpa results that have no specific behavior.
  Is fine if the user is only expected to use getStr()."""
  
  def __init__(self, in_strCpaResult):
    """
    @input in_strCpaResult The CpaResult string, as defined in CpaResult.py
    """    
    assert CpaResult.getCpaResultType(in_strCpaResult) != CpaResult.CpaResult.Type_Unknown
    self.__strCpaResult = in_strCpaResult
  
  def getStr(self):
    return self.__strCpaResult

#
# @tag: Human, Answering-Machine and Fax return 603 on error, should getSipErrorCodeAsInteger
#       be defined for those results?

class Cpa_Busy(Cpa_Generic):
  def getSipErrorCodeAsInteger(self):
    return 486
  
  def getErrorCodeOriginator(self):
    return SipErrorOriginator.Cpa

class Cpa_ReOrder(Cpa_Generic):
  def getSipErrorCodeAsInteger(self):
    return 480

class Cpa_SitTemporary(Cpa_Generic):
  def getSipErrorCodeAsInteger(self):
    return 480

class Cpa_SitPermanent(Cpa_Generic):  
  def getSipErrorCodeAsInteger(self):
    return 404

class _Factory_CallCpaResult:
  translationDict = {
  CpaResult.CpaResult.Busy: Cpa_Busy,
  CpaResult.CpaResult.ReOrder: Cpa_ReOrder,
  CpaResult.CpaResult.SIT_Temporary: Cpa_SitTemporary,
  CpaResult.CpaResult.SIT_Permanent: Cpa_SitPermanent
  }

def Factory_CallCpaResult(in_strCpaResult):
  """Constructs an object of type CallCpaResult using one of the string
  used in the interface with the C++ CpaEngineImpl."""
  
  assert CpaResult.getCpaResultType(in_strCpaResult) != CpaResult.CpaResult.Type_Unknown
  
  rv = None
  if in_strCpaResult in _Factory_CallCpaResult.translationDict:
    rv = _Factory_CallCpaResult.translationDict[in_strCpaResult](in_strCpaResult)
  
  if rv == None:
    rv = Cpa_Generic(in_strCpaResult) 
  
  return rv

class SipError(CallEvent):
  """
  Creation example:
  CallEvent.SipError( CallEvent.SipErrorOriginator.Cpa, 
                      CallEvent.SipErrorCode(400) )
  """
  
  _Trait_isSipError = None
  
  def __init__(self, in_Originator, in_ErrorCode, in_context = None):
    assert isinstance(in_ErrorCode, SipErrorCode)
    assert isinstance(in_Originator, SipErrorOriginator)
    
    self.__sipErrorOriginator = in_Originator
    self.__sipErrorCode = in_ErrorCode
    self.__context = in_context

  def __str__(self):
    rv = "SipError: (%s, %s)" % (self.__sipErrorOriginator, self.__sipErrorCode)
    if self.__context is not None:
      rv += " context:" + self.__context
    return rv
  
  def __eq__(self, other):
    if other.getSipErrorCodeAsInteger() == self.getSipErrorCodeAsInteger() and \
       other.getErrorCodeOriginator() == self.getErrorCodeOriginator() and \
       other.getContextStr() == self.getContextStr():
      return True
    else:
      return False
  
  def __hash__(self):
    """
    Generate an integer that is unique for every instance of this class that
    we want to be distinguishable.
    
    This is necessary if one wants to use this class as a dictionary key.
    """
    
    sipErrorCode = self.getSipErrorCodeAsInteger()
    errorCodeOriginator = self.getErrorCodeOriginator().getErrorOriginatorAsInt()
    assert errorCodeOriginator < 10
    hash_int = 10 * sipErrorCode + errorCodeOriginator
    return hash_int
    
  def getSipErrorCodeAsInteger(self):
    return self.__sipErrorCode.getErrorCodeAsInteger()
  
  def getErrorCodeOriginator(self):
    return self.__sipErrorOriginator
  
  def getContextStr(self):
    return self.__context
  
class SipErrorCode:
  """
  As member of SipError.
  @tag Become a nested class
  """
  
  def __init__(self, in_errorCode):
    assert isinstance(in_errorCode, int)
    self.__errorCode = in_errorCode
  
  def __str__(self):
    return "%s" % self.__errorCode
  
  def getErrorCodeAsInteger(self):
    return self.__errorCode

class SipErrorOriginator:
  """
  As member of SipError.
  @tag Become a nested class
  """
  
  #Type Enum
  _Type_Cpa = 1
  _Type_Caller = 2
  _Type_Callee = 3
  
  def __init__(self, in_Originator):
    if not self.__isValidOriginator(in_Originator):
      raise CallEventException("SipErrorOriginator type is invalid")    
    self.__sipErrorOriginator = in_Originator
  
  def __isValidOriginator(self, in_Originator):
    """Checks if in_Originator's value is contained in the set of this 
    class' static variables beginning by '_Type'."""
    
    lstValid = [in_Originator == self.__class__.__dict__[item] for item in self.__class__.__dict__ if item.count('_Type') >= 1]
    return (True in lstValid)
  
  def __str__(self):
    """Do a reverse look on the class' static variables to find the type
    name from the type value."""
    
    nameFound = []
    for (key, value) in self.__class__.__dict__.iteritems():
      if self.__sipErrorOriginator == value:
        nameFound.append(key)
    
    if len(nameFound)!=1:
      raise CallEventException("Confusion with SipErrorOriginator's type string representation")
    
    SipErrorOriginator_Type = nameFound[0]
    return "SipErrorOriginator: %s" % SipErrorOriginator_Type
  
  def __eq__(self, other):
    if type(other) == type(self):
      if self.__dict__ == other.__dict__:
        return True
    return False
  
  def getErrorOriginatorAsInt(self):
    return self.__sipErrorOriginator

#Add pre-built classes for convenience
SipErrorOriginator.Cpa = SipErrorOriginator(SipErrorOriginator._Type_Cpa)
SipErrorOriginator.Caller = SipErrorOriginator(SipErrorOriginator._Type_Caller)
SipErrorOriginator.Callee = SipErrorOriginator(SipErrorOriginator._Type_Callee)   

class Timeout(CallEvent):
  """
  Creation example:
    Timeout( Timeout.Postconnect, "While waiting for Cpa result" )
    
  Philosophical debate: A Timeout is not _always_ a CallProblem, it is
                        sometimes expected in a 'healthy' call. In the
                        future, it may be decided that a Timeout is really
                        a CallEvent and not a CallProblem.
  """
  
  _Trait_isTimeout = None
  
  #Type Enum
  Preconnect = 'Type_Preconnect'
  Postconnect = 'Type_Postconnect'
  Postconnect_Eamd = 'Type_PostconnectEamd'
  
  def __init__(self, in_TimeoutType, in_context = None):
    assert self.__isTimeoutTypeLegit(in_TimeoutType)
    self.__TimeoutType = in_TimeoutType
    
    if in_context is not None:
      assert isinstance(in_context, basestring)
    "Decorative string used mainly for debug purposes."
    self.__context = in_context
  
  def __str__(self):
    rv = __name__ + " : " + self.getTimeoutType()
    if self.__context is not None:
      rv += " context: " + self.__context
    return rv 
    
  def __isTimeoutTypeLegit(self, in_TimeoutType):
    # A bit of a trick here: it checks all the 'static'
    # variables' values contained in this class.
    if in_TimeoutType in self.__class__.__dict__.values():
      return True
    else:
      return False
  
  def getTimeoutType(self):
    return self.__TimeoutType
  
  def getSipErrorCodeAsInteger(self):
    """A timeout is not necessarily a Sip error, but if it is interpreted
    as a Sip error, it should be able to return an error code."""
    return 408

def test_module():
  Factory_CallCpaResult(CpaResult.CpaResult.Human)
  cpaTimeout = Timeout( Timeout.Postconnect, 'Something_Verbose')
  assert hasattr(cpaTimeout, 'getSipErrorCodeAsInteger')
  
  SipError( SipErrorOriginator.Callee, 
            SipErrorCode(1234567890))   #Notice that there is no check
                                        #on the sip error code value...yet
  
  #Test validation of SipErrorOriginator type
  SipErrorOriginator(SipErrorOriginator._Type_Cpa)
  SipErrorOriginator(SipErrorOriginator._Type_Caller)
  SipErrorOriginator(SipErrorOriginator._Type_Callee)
  gotExcpt = False
  try:
    SipErrorOriginator(1234)
  except CallEventException:
    gotExcpt = True
  assert gotExcpt
  
  #Test string representation of SipErrorOriginator
  so_1 = SipErrorOriginator(SipErrorOriginator._Type_Cpa)
  so_2 = SipErrorOriginator(SipErrorOriginator._Type_Caller)
  so_3 = SipErrorOriginator(SipErrorOriginator._Type_Callee)
  assert str(so_1).count('_Type_Cpa') == 1
  assert str(so_2).count('_Type_Caller') == 1
  assert str(so_3).count('_Type_Callee') == 1
  
  #Check that the shortcuts are equivalent
  assert SipErrorOriginator(SipErrorOriginator._Type_Cpa) == SipErrorOriginator.Cpa
  assert SipErrorOriginator(SipErrorOriginator._Type_Caller) == SipErrorOriginator.Caller
  assert SipErrorOriginator(SipErrorOriginator._Type_Callee) == SipErrorOriginator.Callee
  assert SipErrorOriginator(SipErrorOriginator._Type_Cpa) != SipErrorOriginator.Caller
  assert SipErrorOriginator(SipErrorOriginator._Type_Caller) != SipErrorOriginator.Cpa
  
  #Factory
  Factory_CallCpaResult(CpaResult.CpaResult.Human)
  Factory_CallCpaResult(CpaResult.CpaResult.Busy)

  
if __name__ == '__main__':
  test_module()  

## !!MIRROR FILE, DO NOT MODIFY!!
