#
#  @file   ComputeCpaResult.py
#  @author Francis Moreau
#  date    Mon July 30 2007
#
#  Copyright (C) 2007 Paraxip Technologies.
#  All Rights Reserved.
#
"""
Contains the classes that allows to filter the raw results from the CPA.
"""

import paraxip.infra.PyLoggingFunctions as PyLoggingFunctions
from paraxip.sipua.user.cpa import CpaResult

#Python imports
import re # Regular Expressions

class MinThresholdsContainer:
  def __init__(self, in_pre, in_post, in_eamd, in_global):
    self.min_pre = in_pre
    self.min_post = in_post
    self.min_eamd = in_eamd
    self.unique_global = in_global

  def setMissingThresholdsToGlobal(self):
    """Since the CPA does not have a concept of a global threshold,
    the values are modified to reflect the influence of the global threshold."""
    assert self.unique_global is not None
    if self.min_pre is None:
      self.min_pre = self.unique_global
    if self.min_post is None:
      self.min_post = self.unique_global
    if self.min_eamd is None:
      self.min_eamd = self.unique_global

class fake_cpaResult:
  """It is necessary to create this 'spoof class' in order to test the module.
  
  We cannot access the interface for this object without importing from
  C++ code, which would render the class untestable from Python.
  """
  
  def __init__(self):
    self.m_strOutputName = None
    self.m_dOutputValue = None

class CpaResultFilter:
  """
  This class is a container for the CpaThresholds and contains the logic that
  determines which CpaResult is selected given a set of thresholds and
  state variables.
  """
  
  def __init__(self, in_logger):
    self.__Logger = in_logger
    #There should be only one global threshold... use dict for coherence
    self.__Thresholds_Global = {}
    self.__Thresholds_PreConnect = {}
    self.__Thresholds_PostConnect = {}
    self.__Thresholds_Eamd = {}
    
  ###
  ### Public Functions
  ###
  
  def addThreshold(self, in_strThresholdName, in_ThresholdValue):
    """
    Classify and add threshold to the appropritate group.
    e.g.: addThreshold("EAMD_SILENCE", 0.8)
    """
    assert in_ThresholdValue >= 0
    assert in_ThresholdValue <= 1

    if in_strThresholdName.lower() == 'global':
      assert( len(self.__Thresholds_Global)==0 )
      self.__Thresholds_Global[in_strThresholdName] = in_ThresholdValue
    elif CpaResult.isPreConnect(in_strThresholdName):
      self.__Thresholds_PreConnect[in_strThresholdName] = in_ThresholdValue
    elif CpaResult.isPostConnect(in_strThresholdName):
      self.__Thresholds_PostConnect[in_strThresholdName] = in_ThresholdValue
    elif CpaResult.isEamd(in_strThresholdName):
      self.__Thresholds_Eamd[in_strThresholdName] = in_ThresholdValue
    else:
      PyLoggingFunctions.log_error(self.__Logger,
                'Invalid threshold value in addThreshold: %s at %s.' %
                (in_strThresholdName, in_ThresholdValue))

  def configureCpaThresholdsFromString(self, in_strThreshold):
    """e.g.: global=0.9;human=0.8"""
    
    assert isinstance(in_strThreshold, basestring)

    PyLoggingFunctions.log_debug(self.__Logger, 'configureCpaThresholdsFromString: ' + in_strThreshold)

    vThreshold = in_strThreshold.split(";")
    if (len(vThreshold) == 0):
      PyLoggingFunctions.log_error(self.__Logger, "No valid threshold found in string %s",
                in_strThreshold)

    # break the sting into key,value
    for entries in vThreshold:
      vKeyValue = entries.split("=")
      if(len(vKeyValue) != 2):
        # Can happen at the end, let's see if anything good left
        pass
      else:
        PyLoggingFunctions.log_trace(self.__Logger,
                 "adding key %s to value %s ",
                  vKeyValue[0] , vKeyValue[1])
        self.addThreshold(vKeyValue[0], float(vKeyValue[1]))

    if self.getNumAcceptedThresholds()==0:
      PyLoggingFunctions.log_error(self.__Logger,
                "Could not find any valid threshold in string.")

  def configureCpaThresholdFromDict(self, in_dictThreshold):
    # Can be constructed out of the config dictionary
    assert isinstance(in_dictThreshold, dict)
    
    for k, v in in_dictThreshold.iteritems():
      reThresholdKey = re.compile("cpa.result.threshold\.(.*)")
      thresholdMatch = reThresholdKey.match(k)
      if(thresholdMatch):
        strKey = thresholdMatch.group(1);
        self.addThreshold(strKey,v)

    if self.getNumAcceptedThresholds()==0:
      PyLoggingFunctions.log_error(self.__Logger,
                "Could not find any valid threshold in dictionary.")
  
  def getNumAcceptedThresholds(self):
    total = 0
    total += len(self.__Thresholds_Global)
    total += len(self.__Thresholds_PreConnect)
    total += len(self.__Thresholds_PostConnect)
    total += len(self.__Thresholds_Eamd)
    return total
  
  def getAllThresholdsStr(self):
    str_out = ""
    for item in self.__Thresholds_Global.iteritems():
      str_out += item[0] + '=' + str(item[1]) + ';'
    for item in self.__Thresholds_PreConnect.iteritems():
      str_out += item[0] + '=' + str(item[1]) + ';'
    for item in self.__Thresholds_PostConnect.iteritems():
      str_out += item[0] + '=' + str(item[1]) + ';'
    for item in self.__Thresholds_Eamd.iteritems():
      str_out += item[0] + '=' + str(item[1]) + ';'
    str_out = str_out.strip(';')#Do not want a trailing ';'.
    PyLoggingFunctions.log_trace(self.__Logger, 'getAllThresholdsStr-> ' + str_out)
    return str_out

  def getMinThresholds(self):
    min_t = MinThresholdsContainer(
                                  self.__getMinThreshold_PreConnect(),
                                  self.__getMinThreshold_PostConnect(),
                                  self.__getMinThreshold_Eamd(),
                                  self.__getThreshold_Global())
    return min_t

  def getHighestCpaResultAboveThreshold(self,
                in_cpaResultsListSorted,
                in_bCalleeIsConnected):
    """
    @return The STRING representing the result or None if no valid result found.
    """
    
    __pychecker__ = 'unusednames=t'
    t = PyLoggingFunctions.log_trace(self.__Logger,
                  "CpaThresholds.getResult")

    candidateCpaResult = None
    if (len(in_cpaResultsListSorted) == 0):
      return None

    #
    # Check specifics
    #
    
    candidateCpaResult = \
      self.__returnHighestValidCpaResult(in_cpaResultsListSorted)

    finalCpaResult = None
    if candidateCpaResult is not None:
      if self.__isCpaResultCandidateValid(
                                candidateCpaResult,
                                in_bCalleeIsConnected):
        finalCpaResult = candidateCpaResult
    
    if finalCpaResult is not None:
      return finalCpaResult.m_strOutputName
    
    strThresholds = 'Thresholds_PreConnect: %s\nThresholds_PostConnect:%s\nThresholds_Eamd:%s' %\
                     (self.__Thresholds_PreConnect, self.__Thresholds_PostConnect, self.__Thresholds_Eamd)
    PyLoggingFunctions.log_debug(self.__Logger, "No CPA result above thresholds:\n%s, "
                             "...looking at global threshold." % strThresholds)
    
    
    #
    # Check globals
    #

    finalCpaResult = \
      self.__returnCpaResultGreaterThanGlobalThreshold(in_cpaResultsListSorted)
    
    if finalCpaResult is None:
      strLog = 'ComputeCpaResult could not find a value > thresholds ' +\
               '(including global) in:'
      for entry in in_cpaResultsListSorted:
        strLog += '\n' + entry.m_strOutputName + ' at: ' \
                  + str(entry.m_dOutputValue)
      strLog += '\nGlobal Threshold: %s.' % str(self.__Thresholds_Global.values()[0])
      PyLoggingFunctions.log_debug(self.__Logger, strLog)
    else:
      PyLoggingFunctions.log_debug(self.__Logger, "CPA result matched against global threshold.")
    
    if finalCpaResult is not None:
      return finalCpaResult.m_strOutputName
    else:
      return None

  ###
  ### Tests
  ###

  def test_StringConfig(self):
    constructStr = 'global=0.9;CPA_HUMAN=0.8'
    self.configureCpaThresholdsFromString(constructStr)
    assert self.getAllThresholdsStr() == constructStr

  def test_ThresholdBehavior(self):
    constructStr = 'global=0.9;CPA_HUMAN=0.7;CPA_MACHINE=0.8;CPA_FAX=0.3'
    self.configureCpaThresholdsFromString(constructStr)
    assert self.getAllThresholdsStr() == constructStr

    fake_cpaR_1 = fake_cpaResult()
    fake_cpaR_1.m_strOutputName = 'CPA_HUMAN'
    fake_cpaR_1.m_dOutputValue = 0.69
    fake_cpaR_2 = fake_cpaResult()
    fake_cpaR_2.m_strOutputName = 'CPA_MACHINE'
    fake_cpaR_2.m_dOutputValue = 0.79
    fake_cpaR_3 = fake_cpaResult()
    fake_cpaR_3.m_strOutputName = 'CPA_FAX'
    fake_cpaR_3.m_dOutputValue = 0.29
    cpaResultListSorted = [fake_cpaR_1, fake_cpaR_2, fake_cpaR_3]

    #Test 'np winner'.
    winner_name = self.getHighestCpaResultAboveThreshold(cpaResultListSorted, True)
    PyLoggingFunctions.log_debug(self.__Logger, 'CPA Result: %s.' % winner_name)
    assert winner_name is None
    #Test 'one winner'.
    fake_cpaR_1.m_dOutputValue = 0.70
    winner_name = self.getHighestCpaResultAboveThreshold(cpaResultListSorted, True)
    PyLoggingFunctions.log_debug(self.__Logger, 'CPA Result: %s.' % winner_name)
    assert winner_name is CpaResult.CpaResult.Human
    #Test 'two winners'.
    fake_cpaR_3.m_dOutputValue = 0.75
    winner_name = self.getHighestCpaResultAboveThreshold(cpaResultListSorted, True)
    PyLoggingFunctions.log_debug(self.__Logger, 'CPA Result: %s.' % winner_name)
    assert winner_name is CpaResult.CpaResult.Fax

  def test_getMinThreshold(self):
    constructStr = \
      'CPA_HUMAN=0.7;CPA_SIT_TEMPORARY=0.2;CPA_SIT_PERMANENT=0.1;CPA_MACHINE=0.8;CPA_FAX=0.3'
    self.configureCpaThresholdsFromString(constructStr)
    min_t = self.__getMinThreshold_PreConnect()
    PyLoggingFunctions.log_debug(self.__Logger, 'min_t returned: %s.' % min_t)
    assert min_t == 0.1
    min_t = self.__getMinThreshold_Eamd()
    PyLoggingFunctions.log_debug(self.__Logger, 'min_t returned: %s.' % min_t)
    assert min_t == None
    min_t = self.__getMinThreshold_PostConnect()
    PyLoggingFunctions.log_debug(self.__Logger, 'min_t returned: %s.' % min_t)
    assert min_t == 0.3

  def test_globals(self):
    # strottier: shouldn't it be cpa.result.threshold.CPA_HUMAN ?
    constructDict = {'CPA_HUMAN':0.99, 'cpa.result.threshold.global':0.8}
    self.configureCpaThresholdFromDict(constructDict)

    fake_cpaR_1 = fake_cpaResult()
    fake_cpaR_1.m_strOutputName = 'CPA_HUMAN'
    fake_cpaR_1.m_dOutputValue = 0.69
    fake_cpaR_2 = fake_cpaResult()
    fake_cpaR_2.m_strOutputName = 'CPA_MACHINE'
    fake_cpaR_2.m_dOutputValue = 0.79
    fake_cpaR_3 = fake_cpaResult()
    fake_cpaR_3.m_strOutputName = 'CPA_FAX'
    fake_cpaR_3.m_dOutputValue = 0.29
    cpaResultListSorted = [fake_cpaR_1, fake_cpaR_2, fake_cpaR_3]

    #Test 'np winner'.
    winner_name = self.getHighestCpaResultAboveThreshold(cpaResultListSorted, True)
    PyLoggingFunctions.log_debug(self.__Logger, 'CPA Result: %s.' % winner_name)
    assert winner_name is None
    fake_cpaR_2.m_dOutputValue = 0.81
    winner_name = self.getHighestCpaResultAboveThreshold(cpaResultListSorted, True)
    PyLoggingFunctions.log_debug(self.__Logger, 'CPA Result: %s.' % winner_name)
    assert winner_name is CpaResult.CpaResult.AnsweringMachine

  ###
  ### Private Functions
  ###
  
  def __returnHighestValidCpaResult(self, in_cpaResultsListSorted):
    """Returns the highest valid CpaResult.
    
    Gathers all the results above threshold with matching names. If there is
    more than one match, returns the CpaResult with the highest absolute value
    amongst valid results.

    @return A CpaResult class
    """
    
    PyLoggingFunctions.log_trace(self.__Logger, '__returnHighestValidCpaResult with list: ' + \
              str(in_cpaResultsListSorted))

    #Dict of: {threshold:result_str}; e.g.: {0.75:'CPA_HUMAN'}
    #Stored as such to be able to use the unmodified max function
    CpaResultsAboveThreshold = {} 

    for cpaResult in in_cpaResultsListSorted:
      if cpaResult.m_strOutputName in self.__Thresholds_PreConnect:
        if cpaResult.m_dOutputValue >= \
                    self.__Thresholds_PreConnect[cpaResult.m_strOutputName]:
          CpaResultsAboveThreshold[cpaResult.m_dOutputValue] = cpaResult

      if cpaResult.m_strOutputName in self.__Thresholds_PostConnect:
        if cpaResult.m_dOutputValue >= \
                    self.__Thresholds_PostConnect[cpaResult.m_strOutputName]:
          CpaResultsAboveThreshold[cpaResult.m_dOutputValue] = cpaResult

      if cpaResult.m_strOutputName in self.__Thresholds_Eamd:
        if cpaResult.m_dOutputValue >= \
                    self.__Thresholds_Eamd[cpaResult.m_strOutputName]:
          CpaResultsAboveThreshold[cpaResult.m_dOutputValue] = cpaResult
          
    #
    # More than one result possible, return highest one.
    #
    if len(CpaResultsAboveThreshold)==0:
      return None
    elif len(CpaResultsAboveThreshold)==1:
      return CpaResultsAboveThreshold.items()[0][1]
    elif len(CpaResultsAboveThreshold)>1:
      max_threshold = max( CpaResultsAboveThreshold )
      max_CpaResult = \
        CpaResultsAboveThreshold[max_threshold]
      return max_CpaResult

  def __returnCpaResultGreaterThanGlobalThreshold(self, in_cpaResultsListSorted):
    """
    @return A CpaResult class
    """
    assert len(self.__Thresholds_Global)==1
    globalThreshold = self.__Thresholds_Global.values()[0]

    candidateCpaResult = None
    for cpaResult in in_cpaResultsListSorted:
      if cpaResult.m_dOutputValue > globalThreshold:
        if candidateCpaResult is None:
          candidateCpaResult = cpaResult
        elif cpaResult.m_dOutputValue > candidateCpaResult.m_dOutputValue:
          candidateCpaResult = cpaResult
    return candidateCpaResult

  def __getMinValueInDict(self, in_dict):
    """@return Minimum value in dict; None if no value in dict.

    It will return None quite often (e.g.: in the legacy configuration files
    there is a global threshold but no specific thresholds (pre, post, eamd)).
    In that case, {pre, post, eamd} will be equal to None.

    It is NOT the responsibility of this class to judge what None should be
    interpreted as (resist the temptation to return 0.0 in the case of None).
    """

    min_item = None
    for item in in_dict.values():
      if min_item is None: min_item = item
      if item < min_item:
        min_item = item
    return min_item

  def __getMinThreshold_PreConnect(self):
    rv = self.__getMinValueInDict(self.__Thresholds_PreConnect)
    PyLoggingFunctions.log_debug(self.__Logger, '__getMinThreshold_PreConnect returning ' +
                              str(rv))
    return rv

  def __getMinThreshold_PostConnect(self):
    rv = self.__getMinValueInDict(self.__Thresholds_PostConnect)
    PyLoggingFunctions.log_debug(self.__Logger, '__getMinThreshold_PostConnect returning ' +
                              str(rv))
    return rv

  def __getMinThreshold_Eamd(self):
    rv = self.__getMinValueInDict(self.__Thresholds_Eamd)
    PyLoggingFunctions.log_debug(self.__Logger, '__getMinThreshold_Eamd returning ' +
                              str(rv))
    return rv

  def __getThreshold_Global(self):
    rv = self.__getMinValueInDict(self.__Thresholds_Global)
    PyLoggingFunctions.log_debug(self.__Logger, '__getThreshold_Global returning ' +
                              str(rv))
    return rv
  
  def __isCpaResultCandidateValid(self, candidateCpaResult,
                                        in_bCalleeIsConnected):
    in_strResult = candidateCpaResult.m_strOutputName
    in_fValue = candidateCpaResult.m_dOutputValue

    if in_bCalleeIsConnected:
      if(CpaResult.isPostConnect(in_strResult) or
         CpaResult.isEamd(in_strResult)):
        return True
      else:
        PyLoggingFunctions.log_info(self.__Logger,
                 "Wrong type of CPA result above threshold in a connected state, "
                 "result = %s, value = %f",
                 in_strResult,
                 in_fValue)

    else:
      if CpaResult.isPostConnect(in_strResult):
        return True
      else:
        PyLoggingFunctions.log_info(self.__Logger,
                 "Wrong type of CPA result above threshold in a pre-connect state, "
                 "result = %s, value = %f",
                 in_strResult,
                 in_fValue)

    return False

class FakeLogger:
  """
  A Fakelogger that outputs to stdout; used for test-cases.
  """
  
  isEnabled = True
  
  def isEnabledFor(self, in_logLevel):
    __pychecker__ = 'unusednames=in_logLevel'
    return self.isEnabled
  
  def forcedLog(self, in_LogLevel, in_Msg, in_file, in_unknown):
    __pychecker__ = 'unusednames=in_file,in_unknown'
    print 'FakeLog[%s]: %s' % (in_LogLevel, in_Msg)

def test_module():
  fakeLogger = FakeLogger()
  fakeLogger.isEnabled = True #Quiesce?
  PyLoggingFunctions.log_debug(fakeLogger, 'Test of %s begin' % __file__)
  CpaResultFilter(fakeLogger).test_StringConfig()
  CpaResultFilter(fakeLogger).test_ThresholdBehavior()
  CpaResultFilter(fakeLogger).test_getMinThreshold()
  CpaResultFilter(fakeLogger).test_globals()
  PyLoggingFunctions.log_debug(fakeLogger, 'Test of %s end' % __file__)

if __name__ == '__main__':
  test_module()
## !!MIRROR FILE, DO NOT MODIFY!!
