#
#  @file   CpaCall.py
#  @author Francis Moreau
#  date    Mon May 22 2007
#  
#  Description:
#
#  Session Controller's state-machine when it is communicating with the CPA.
#
#  Copyright (C) 2007 Paraxip Technologies.
#  All Rights Reserved.
#

#Python Core Libraries
import re
import os.path

#Paraxip Core libraries
import paraxip.infra.PyLoggingFunctions as PyLoggingFunctions
from netborderPyInfra import TraceScope, CallLogger, Logger  

#Machine Learning
from netborderPyML import PyMLClassifierResult

#Sip-Related
from SipCallEngineExtending import PySipMessage
from paraxip.sipua.internal.SipCallEngineSMCall import SipCallEngineSMCall
from paraxip.sipua.internal.ConnectionStatus import ConnectionStatus
from paraxip.sipua.user.cpa import VendorCustomSIPHeaders

#Cpa Class 'Parts'
from paraxip.sipua.user.b2bua.B2buaStates import IdleState
from paraxip.sipua.user.b2bua.B2buaCall import B2buaCall

#Cpa-Data Related
from paraxip.sipua.user.cpa import CpaStates
from paraxip.sipua.user.cpa import CpdResult
from paraxip.sipua.user.cpa import CpaResult
from paraxip.sipua.user.cpa import ComputeCpaResult
from paraxip.sipua.user.cpa import DetailedCpdResult
from paraxip.sipua.user.cpa import CallEventStack
from paraxip.sipua.user.cpa import CallEvent

#Cpa-Data Conversion
from paraxip.sipua.user.cpa.calldata_converter import SmartConverter

## CPASTATS
from paraxip.sipua.ui.model.cpaOM import CpaOM 
from paraxip.sipua.user.cpa import CpaStatsCsv

## Misc
sdpFromStreamManagerRegEx = re.compile('s=StreamManager')

# Force declaration of class attributes in __init__ method
__pychecker__ = "initattr"
    
# The Cpa scenario derives from b2bua scenario
class CpaCall(B2buaCall):
  """
  Class logged as: netborder.cpa.CpaCall.pyc
  """
  
  # Valid parameter names for this Call
  PARAM_CPA_PRECONNECT_TIMEOUT_MS = 'cpa.preconnect.timeout.ms'
  PARAM_CPA_POSTCONNECT_TIMEOUT_MS = 'cpa.postconnect.timeout.ms'
  PARAM_CPA_RESULT_THRESHOLD_GLOBAL = 'cpa.result.threshold.global'
  PARAM_CPA_RESULT_THRESHOLD_HUMAN = 'cpa.result.threshold.CPA_HUMAN'
  PARAM_CPA_RESULT_THRESHOLD_ANSWERING_MACHINE = \
    'cpa.result.threshold.CPA_MACHINE'
  PARAM_CPA_RESULT_THRESHOLD_FAX = 'cpa.result.threshold.CPA_FAX'
  PARAM_CPA_ENGINE_HOST = 'cpa.engine.host'
  PARAM_CPA_ENGINE_PORT = 'cpa.engine.port'
  PARAM_CPA_STATS_CUSTOM_HEADERS = 'cpa.stats.customHeaders'
  
  # Provide parameter information for this Call 
  @staticmethod
  def getParameterInfoDB():

    # Start with inherited params
    parameterInfoDB = B2buaCall.getParameterInfoDB()

    # Parameters that must be set explicitly (mandatory, no defaults)
    parameterInfoDB.addParameterInfoForName(
        CpaCall.PARAM_CPA_PRECONNECT_TIMEOUT_MS,
        type=int)
    parameterInfoDB.addParameterInfoForName(
        CpaCall.PARAM_CPA_ENGINE_PORT,
        type=int)

    # Parameter with default values
    parameterInfoDB.addParameterInfoForName(
        CpaCall.PARAM_CPA_POSTCONNECT_TIMEOUT_MS,
        type=int,
        defaultValue=7000)
    parameterInfoDB.addParameterInfoForName(
        CpaCall.PARAM_CPA_RESULT_THRESHOLD_GLOBAL,
        type=float,
        defaultValue=0.9)  
    parameterInfoDB.addParameterInfoForName(
        CpaCall.PARAM_CPA_ENGINE_HOST,
        type=str,
        defaultValue='127.0.0.1')

    # Optional parameters (no defaults)
    parameterInfoDB.addParameterInfoForName(
        CpaCall.PARAM_CPA_RESULT_THRESHOLD_HUMAN,
        type=float,
        mandatory=False)
    parameterInfoDB.addParameterInfoForName(
        CpaCall.PARAM_CPA_RESULT_THRESHOLD_ANSWERING_MACHINE,
        type=float,
        mandatory=False)
    parameterInfoDB.addParameterInfoForName(
        CpaCall.PARAM_CPA_RESULT_THRESHOLD_FAX,
        type=float,
        mandatory=False)

  # Internal params that cannot be read from file (for now)
  # Custom headers that are stored in CPA stats
  # Defaults to all custom headers present in SIP INVITE
    parameterInfoDB.addParameterInfoForName(
        CpaCall.PARAM_CPA_STATS_CUSTOM_HEADERS,
        type=dict)

    return parameterInfoDB

  def __init__(self):
    __pychecker__ = 'unusednames=t'

    B2buaCall.__init__(self)
    
    t = TraceScope(self.getLogger(), "CpaCall.__init__")
    
    self._strCpdResultCustomHeader = "CPD-Result"
    self._strCallIdCustomHeader = "X-Netborder-Call-ID"
    
    ###
    ### PER CALL MEMBERS
    ###
    
    self.cpaDialogId = None
    self.declarePerCallMembers('cpaDialogId')
    
    ###
    ### Call State Variables
    ###
    
    self.cpaLegStatus = ConnectionStatus.Unknown
    self.declarePerCallMembers('cpaLegStatus') 
    
    self.cpaSdpAnswered = None
    self.declarePerCallMembers('cpaSdpAnswered')
    
    self.__callEventStack = CallEventStack.CallEventStack()
    self.declarePerCallMembers('_' + 'CpaCall' + '__callEventStack')
    
    ###
    ### Utility classes
    ###
    self._cpaResultFilter = None #Is a class that will be instantiated
    self.declarePerCallMembers('_cpaResultFilter')
    self._cpaSipResponse = None  #Is a class that will be instantiated 
    self.declarePerCallMembers('_cpaSipResponse')
    ##
    
    # CPASTATS
    self.cpaStats = None
    self.declarePerCallMembers('cpaStats')
    ### PER CALL MEMBERS END
  
  def getDescriptionOfDialogAndLegStatus(self):
    strDescr = "\n";
    
    strDescr += "dialogIdInitiatingDisconnect: %s\n" % (self.dialogIdInitiatingDisconnect)
    
    strDescr += "callerLegStatus: %s\n" % (self.callerLegStatus)
    strDescr += "calleeLegStatus: %s\n" % (self.calleeLegStatus)
    strDescr += "cpaLegStatus: %s\n" % (self.cpaLegStatus)
    
    strDescr += "callerDialogId: %s\n" % (self.callerDialogId)
    strDescr += "calleeDialogId: %s\n" % (self.calleeDialogId)
    strDescr += "cpaDialogId: %s\n" % (self.cpaDialogId)
    
    return strDescr
  
  def getCallStack(self):
    return self.__callEventStack
  
  def notifyCallEvent(self, in_callEvent):
    self.__callEventStack.notifyCallEvent(in_callEvent)
 
  def setSMCallLogger(self):
    self.setLogger(CallLogger("netborder.cpa.%s" % os.path.split(__file__)[1]))
  
  def addStates(self):
    dialingCpaState = CpaStates.DialingCpaState(self)
    
    # IDLE state (from B2BUA scenario)
    idleState = IdleState(self, dialingCpaState.getName())
    self.setInitialStateName(idleState.getName())
    self.setInitialState(idleState)
    
    # DIALING_CPA state
    self.addState(dialingCpaState.getName(), dialingCpaState)
    
    # DIALING_CALLEE state
    dialingCalleeState = CpaStates.DialingCalleeState(self)
    self.addState(dialingCalleeState.getName(), dialingCalleeState)
    
    # WAITING_CPA_RESULT state
    waitingForCpaPreOrPostResultState = CpaStates.WaitingForCpaPreOrPostResultState(self)
    self.addState(waitingForCpaPreOrPostResultState.getName(), waitingForCpaPreOrPostResultState)
    
    # REINVITING_CALLEE state
    reinvitingCalleeState = CpaStates.ReinvitingCalleeState(self)
    self.addState(reinvitingCalleeState.getName(), reinvitingCalleeState)
    
    # ANSWERING_CALLER state
    answeringCallerState = CpaStates.AnsweringCallerState(self)
    self.addState(answeringCallerState.getName(), answeringCallerState)
    
    # CALLER_CALLEE_CONNECTED state
    callerCalleeConnectedState = CpaStates.CallerCalleeConnectedState(self)
    self.addState(callerCalleeConnectedState.getName(),
                  callerCalleeConnectedState)
    
    # DISCONNECTING state
    disconnectingState = CpaStates.DisconnectingState(self)
    self.addState(disconnectingState.getName(), disconnectingState)
    
    # DISCONNECTED state
    disconnectedState = CpaStates.DisconnectedState(self)
    self.addFinalState(disconnectedState.getName(), disconnectedState)
                       
    # REINVITED state
    self._reinvitedState = CpaStates.ReinvitedState(self)
    self._reinvitedState.setSuccessStateName(
      callerCalleeConnectedState.getName())
    self._reinvitedState.setFailureStateName(
      callerCalleeConnectedState.getName())
    self._reinvitedState.setDisconnectStateName(
      disconnectingState.getName())
    self.addState(self._reinvitedState.getName(), self._reinvitedState)
  
  def initialize(self, in_refApplication, in_refGlobalContext, in_szCallId):
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "CpaCall.initialize")
    
    B2buaCall.initialize(self,
                                         in_refApplication,
                                         in_refGlobalContext,
                                         in_szCallId)
    
    
    PyLoggingFunctions.log_debug(self.getLogger(), "CallCpa.py initialized with config: " + str(self.config))
    
    self._cpaResultFilter = ComputeCpaResult.CpaResultFilter(self.getLogger())
    self._cpaResultFilter.configureCpaThresholdFromDict(self.config)
    
    ## CPASTATS
    cpaOM = CpaOM()
    if CpaCall.PARAM_CPA_STATS_CUSTOM_HEADERS in self.config:
      self.addCustomHeaderInformationToCpaStats(
          self.config[CpaCall.PARAM_CPA_STATS_CUSTOM_HEADERS],
          cpaOM)
    else:
      PyLoggingFunctions.log_debug(self.getLogger(),
          CpaCall.PARAM_CPA_STATS_CUSTOM_HEADERS + " is not in the "
          "configuration, can not use custom headers.")
    self.cpaStats = cpaOM
    
    return True
  
  #@override
  def notifyLoopDetected(self):
    super(CpaCall, self).notifyLoopDetected()
    
    ce = CallEvent.SipError( CallEvent.SipErrorOriginator.Caller, 
                             CallEvent.SipErrorCode(482),
                             "Loop Detected (a Request-Uri to a B2BUA must be different than "
                             "its address/port).")
    self.notifyCallEvent(ce)
    
  def addCustomHeaderInformationToCpaStats(self, in_dictOfCustomHeaders, in_cpaOM):
    """
    Adds what is recognized as a valid custom header to the related fields
    in the cpa-stats object.
    """
    
    PyLoggingFunctions.log_debug(self.getLogger(),
                        "CustomHeaders: %s.", in_dictOfCustomHeaders)
    
    vendorHelper = VendorCustomSIPHeaders.Helper()
    vendorCampaignID = \
      vendorHelper.GetCustomVendorCampaignID_FromDict(in_dictOfCustomHeaders)
    vendorCampaignName = \
      vendorHelper.GetCustomVendorCampaignName_FromDict(in_dictOfCustomHeaders)
    
    ## 'Commit' the values to self.cpaStats
    
    if vendorCampaignID!=vendorHelper.GetNotFoundMarker():
      in_cpaOM.addOM(CpaOM.OM_CPA_CAMPAIGN_ID, vendorCampaignID)
    
    if vendorCampaignName!=vendorHelper.GetNotFoundMarker():
      in_cpaOM.addOM(CpaOM.OM_CPA_CAMPAIGN_NAME, vendorCampaignName)
    
  def addOutDialog(self, in_szDialogId, in_sipCallEngineSMOut):
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "CpaCall.addOutDialog")
    SipCallEngineSMCall.addOutDialog(self, in_szDialogId, in_sipCallEngineSMOut)
    
    # First outgoing call is with the cpa UA
    if (self.cpaDialogId == None):
      # Save cpa dialog id
      self.cpaDialogId = in_szDialogId
    else:
      # Save callee dialog id
      self.calleeDialogId = in_szDialogId
  
  def scenarioName(self):
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(),
               "CpaCall.scenarioName")
    return "Netborder CPA"
  
  def prepareCpaEndpointMsg(self):
    # Inviting cpa endpoint without sdp
    cpaEndpointUrl = "sip:%s:%d" % (self.config[CpaCall.PARAM_CPA_ENGINE_HOST],
                                    self.config[CpaCall.PARAM_CPA_ENGINE_PORT])
    inviteMsg = PySipMessage()
    
    if (not self.globalContext.constructInviteMessage(inviteMsg,
                                                      self.getCallId(),
                                                      cpaEndpointUrl)):
      # Error logged by constructInviteMessage
      return
    
    # Add the custom headers
    inviteMsg.addCustomHeader("NetBorder-Call-Id", self._strCallId + ".analyzer-engine")
    inviteMsg.addCustomHeader("CPA-Threshold",
                              self._cpaResultFilter.getAllThresholdsStr())
    
    return inviteMsg
  
  def callCpaEndpoint(self):
    
    inviteMsg = self.prepareCpaEndpointMsg()
    
    # Send the SIP INVITE message
    self.globalContext.makeOutCall(inviteMsg)
    self.cpaLegStatus = ConnectionStatus.Connecting
  
  def getFinalCPAOutcomeFromInfoMsg(self, in_sipMsg, in_bCalleeIsConnected):
    """
    Returns the CPA outcome as a string, None if criteria for a final decision 
    wasn't reached. The sip message is sent directly from the CPA.
    """
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "CpaCall.getFinalCPAOutcomeFromInfoMsg")
    
    if (in_sipMsg.hasContents()):
      if (in_sipMsg.getContentType() == "application" and \
          in_sipMsg.getContentSubType() == "cpa-update"):
        
        # OK, the body of this INFO contains cpa intermediate results
        strCPAResults = in_sipMsg.getBodyText();
        
        cpaResult = PyMLClassifierResult()
        bRv = cpaResult.parseResultsFromString(strCPAResults)
        if(not bRv):
          PyLoggingFunctions.log_error(self.getLogger(),
                    "Cannot parse CPA result = %s", strCPAResults)
          return None
        
        # No need to sort, already sorted. Leave the CPU to other more 
        # meaningful task.
        #cpaResult.sort()
        
        cpaResultsListSorted = cpaResult.getResultListCopy()
        if (len(cpaResultsListSorted) == 0):
          PyLoggingFunctions.log_error(self.getLogger(),
                    "No valid CPA result in = %s", strCPAResults)
          return None
        
        #
        # getHighestCpaResultAboveThreshold with check threshold; returns None if < thresholds
        #
        cpaResult = self._cpaResultFilter.getHighestCpaResultAboveThreshold(
            cpaResultsListSorted,
            in_bCalleeIsConnected)
        return cpaResult
    return None
  
  def getAnswerMessageWithCpaResult(self):
    """
    This is where the '200 OK' containing the CPA result to be sent
    to the caller is constructed.
    """
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "CpaCall.getAnswerMessageWithCpaResult")
    
    sipMessage = PySipMessage()

    if (not self.constructAnswerMessageInDialog(self.callerDialogId, sipMessage)):
      return None
    
    strCpdResult = self.getCpdResultFromStateVariables()

    assert sipMessage.addCustomHeader(self._strCpdResultCustomHeader,
                                      strCpdResult)

    strCpdDetailedResult = self.getDetailedCpdResultFromStateVariables()

    assert sipMessage.addCustomHeader(DetailedCpdResult.DetailedCpdResult.getSipHeaderName(),
                                      strCpdDetailedResult)  
    
    # Notify UAC of our internal call-id
    assert sipMessage.addCustomHeader(self._strCallIdCustomHeader,
                                      self._strCallId)

    return sipMessage
  
  def getRejectMessageAndSipErrorCodeFromStateVariables(self):
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "CpaCall.getRejectMessageAndSipErrorCodeFromStateVariables")
    
    sipMessage = PySipMessage()
    
    #
    # Note that the message will not contain ANYTHING useful.
    # SipCallEngineSMInOutBaseClass.py -> 
    #     PySipLeg.cpp -> 
    #       InviteSession.cxx::constructRejectMessage:
    #         return SharedPtr<SipMessage> pSipMessage(new SipMessage);
    #
    self.constructRejectMessageInDialog(self.callerDialogId, sipMessage)
    
    strCpdResult = self.getCpdResultFromStateVariables()
    
    # CPASTATS
    self.cpaStats.addCurrentTimeOM(CpaOM.OM_CPA_TIME_CPA_COMPLETED)
    
    assert sipMessage.addCustomHeader(self._strCpdResultCustomHeader,
                                      strCpdResult)
    
    strCpdDetailedResult = self.getDetailedCpdResultFromStateVariables()
    assert sipMessage.addCustomHeader(DetailedCpdResult.DetailedCpdResult.getSipHeaderName(),
                                      strCpdDetailedResult)
    
    assert sipMessage.addCustomHeader(self._strCallIdCustomHeader,
                                      self._strCallId)
    
    uiRejectResponseCode = self.getIntSipRejectResponseCodeFromStateVariables()
    
    return uiRejectResponseCode, sipMessage

  def getIntSipRejectResponseCodeFromStateVariables(self):
    """
    @output: An integer representing the SIP error code OR
             None if there was no 'CallProblem' signaled.
    """
    callEventStack = self.getCallStack()
    callProblem = callEventStack.getLastCallProblem()
    
    if callProblem is None:
      #
      # If there is no CallProblem, check if the CpaResult can be
      # interpreted as a Sip Error (e.g.: Cpa_Busy, Cpa_ReOrder, ...)
      #
      cpaResult = callEventStack.getCpaResult()
      intSipErrorCode = None
      if hasattr(cpaResult, 'getSipErrorCodeAsInteger'):
        intSipErrorCode = cpaResult.getSipErrorCodeAsInteger()
      return intSipErrorCode
    else:
      return callProblem.getSipErrorCodeAsInteger()
      
  def __getCallAnalysisFromStateVariables(self):
    callEventStack = self.getCallStack()
    strCpdResult = SmartConverter.do(callEventStack, CpdResult.CpdResult)
    detailedCpdResult = SmartConverter.do(callEventStack, DetailedCpdResult.DetailedCpdResult)
    
    return strCpdResult, detailedCpdResult
  
  def getCpdResultFromStateVariables(self):    
    callEventStack = self.getCallStack()
    strCpdResult = SmartConverter.do(callEventStack, CpdResult.CpdResult)  
    return strCpdResult
  
  def getDetailedCpdResultFromStateVariables(self):
    callEventStack = self.getCallStack()
    detailedCpdResult = SmartConverter.do(callEventStack, DetailedCpdResult.DetailedCpdResult)
    return detailedCpdResult
  
  def dumpCpaStats(self):
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(),
               "CpaCall.dumpCpaStats")
    
    #Protect against a problem observed under load
    requestUri = None
    if self.callerInitialInvite is not None:
      requestUri = self.callerInitialInvite.getRequestUri()
    else:
      requestUri = 'PyErr'
      PyLoggingFunctions.log_error(self.getLogger(),
                                   "Logging a call with no initial invite.")  
    
    CpaStatsCsv.WriteToCsv(self.cpaStats, self._strCallId, requestUri)
    
  def isSdpFromStreamManager(self, in_sdp):
    """
      SDP is from Genesys Stream Manager  if it contains the following stream:
      - s=StreamManager-
      This hack is used by some customer (Telus) to evaluate if the call was place 
       in a queue waiting to be transferred to an agent. The logic is that call put in queue
       are "sent" to StreamManager
    """
    strSDP = in_sdp.getBodyDataStringCopy()
    # TODO: Improve. This is a bit a crude approximation for a check.
    if(sdpFromStreamManagerRegEx.search(strSDP)):
      PyLoggingFunctions.log_trace(self.getLogger(),"SDP is from Stream Manager")
      return True
    else:
      PyLoggingFunctions.log_trace(self.getLogger(),"SDP is not from StreamManager")
      return False

  def transitionFromPreOrPostDetectionDone(self, out_pystrNewStateName):
    """
    'Virtual' method to be overridden by subclasses to allow customization 
    of the state transition that should happen after the {Pre, Port} detection
    has terminated.
    
    @return {True:success, False: failure}
    
    E.g. in child: 
      out_pystrNewStateName.set('NAME_OF_NEXT_STATE_TO_TRANSITION_TO')
      return True
    """
    return self._reintegrateCallFlowAfterCPADone(out_pystrNewStateName)

  # Helper
  def _reintegrateCallFlowAfterCPADone(self, out_pystrNewStateName):
    
    # Hangup the cpa endpoint
    if (not self.hangupDialog(self.cpaDialogId)):
      return False
    
    # Disconnect the CPA, will not need updates anymore
    self.cpaLegStatus = ConnectionStatus.Disconnecting
    
    # Only human, answering machine or fax can be detected
    # when the Callee is Connected
    if (self.calleeLegStatus == ConnectionStatus.Connected):
      # Re-invite the callee
      out_pystrNewStateName.set("REINVITING_CALLEE")
      return True
    
    # Only busy, re-order or any SIT can be detected in pre-connect mode
    else:
      out_pystrNewStateName.set("DISCONNECTING")
      return True

## !!MIRROR FILE, DO NOT MODIFY!!
