#
#  @file   CpaUASCall.py
#  @author Francis Moreau
#  date    Mon Jun 11 2007
#  
#  Copyright (C) 2007 Paraxip Technologies.
#  All Rights Reserved.
#

from paraxip.sipua.internal.SipCallEngineSMCall import SipCallEngineSMCall
from paraxip.sipua.internal.ConnectionStatus import ConnectionStatus
import paraxip.infra.PyLoggingFunctions as PyLoggingFunctions

from netborderPyInfra import TraceScope, CallLogger
from netborderPyML import PyMLClassifierResult  
from paraxip.sipua.user.cpa import CpaUASStates
from paraxip.sipua.user.cpa import ComputeCpaResult
from SipCallEngineExtending import PySipMessage, PyCPAClassifierType

#Python imports
import re
import os.path

#
# Call state machine
#

# Force declaration of class attributes in __init__ method
__pychecker__ = "initattr"

class CpaUASCall(SipCallEngineSMCall):
  def __init__(self):
    SipCallEngineSMCall.__init__(self)
    
    self.setLogger(CallLogger("netborder.cpa.engine-service.%s" % \
        os.path.split(__file__)[1]))
    
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "CpaUASCall.__init__")
    
    ### PER CALL MEMBERS
    self.callerDialogId = None
    self.callerLegStatus = ConnectionStatus.Unknown
    self.declarePerCallMembers('callerDialogId', 'callerLegStatus')

    self.dialogIdInitiatingDisconnect = None
    self.declarePerCallMembers('dialogIdInitiatingDisconnect')
    
    self.uiRejectCallerWithInternalSipErrorCode = None
    self.strHangupWithInternalReasonHeader = None
    self.declarePerCallMembers('uiRejectCallerWithInternalSipErrorCode', \
        'strHangupWithInternalReasonHeader')
    
    #Will be created by state 'InDialogToCallCalledEvent'.
    #Is of type PySipCallMedia
    self.media = None
    self.bMediaComputeCPAResultsCalled = False
    self.declarePerCallMembers('media', 'bMediaComputeCPAResultsCalled')
    
    self.cpaCalleeAnswered = False
    self._CallRejectReason = None
    self.declarePerCallMembers('cpaCalleeAnswered', '_CallRejectReason')
    ### PER CALL MEMBERS END
    
    self._cpaMsgsSentMemory = None

    # IDLE state
    idleState = CpaUASStates.IdleState(self)
    self.setInitialStateName(idleState.getName())
    self.setInitialState(idleState)

    # GENERATING_SDP_OFFER state
    generatingSdpOfferState = CpaUASStates.GeneratingSdpOfferState(self)
    self.addState(generatingSdpOfferState.getName(), generatingSdpOfferState)

    
    # ANSWERING state
    answeringState = CpaUASStates.AnsweringState(self)
    self.addState(answeringState.getName(), answeringState)
    
    # CONNECTED state
    connectedState = CpaUASStates.ConnectedState(self)
    self.addState(connectedState.getName(), connectedState)
    
    # SENDING_CPA_UPDATE state
    sendingCPAUpdateState = CpaUASStates.SendingCPAUpdateState(self)
    self.addState(sendingCPAUpdateState.getName(), sendingCPAUpdateState)
    
    # DISCONNECTING state
    disconnectingState = CpaUASStates.DisconnectingState(self)
    self.addState(disconnectingState.getName(), disconnectingState)
    
    # DISCONNECTED state
    _disconnectedState = CpaUASStates.DisconnectedState(self)
    self.addFinalState(_disconnectedState.getName(), _disconnectedState)   
    
  def initialize(self, in_refApplication, in_refGlobalContext, in_szCallId):
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "CpaUASCall.initialize")
    
    SipCallEngineSMCall.initialize(self,
                                   in_refApplication,
                                   in_refGlobalContext,
                                   in_szCallId)
    
    PyLoggingFunctions.log_debug(self.getLogger(), "CpaUASCall.py initialized with config: " + str(self.config))
    self._CallRejectReason = None
    
    return True
      
  def addInDialog(self, in_szDialogId, in_sipCallEngineSMIn):
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "CpaUASCall.addInDialog")
    SipCallEngineSMCall.addInDialog(self, in_szDialogId, in_sipCallEngineSMIn)
    
    # Save the caller dialog id
    self.callerDialogId = in_szDialogId
  
  def getAnswerMessage(self):
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(),
               "CpaUASCall.getAnswerMessage")
    
    sipMessage = PySipMessage()
    
    if (not self.constructAnswerMessageInDialog(self.callerDialogId, sipMessage)):
      return None
    
    # Add custom header here
    # assert sipMessage.addCustomHeader("header name", "header value")
    
    return sipMessage
  
  def getRejectMessage(self):
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(),
               "CpaUASCall.getRejectMessage")
    
    sipMessage = PySipMessage()
    
    self.constructRejectMessageInDialog(self.callerDialogId, sipMessage)
    
    # Get the sip reject response code
    uiRejectResponseCode = 603 # Decline
    if (self.uiRejectCallerWithInternalSipErrorCode is not None):
      uiRejectResponseCode = self.uiRejectCallerWithInternalSipErrorCode
    
    # Add custom header here
    if self._CallRejectReason is not None:
      assert sipMessage.addCustomHeader("RejectReason", self._CallRejectReason)
    
    return uiRejectResponseCode, sipMessage

  def handleIncomingInfoMsg(self, in_infoReceivedEvent):
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(),
               "CpaUASCall.handleIncomingInfoMsg")
    
    # SIP INFO coming in from the controller: check if it is reporting messages
    # from the callee
    sipMsg = in_infoReceivedEvent.getSipMessage()
    if sipMsg.hasContents():
      if sipMsg.getContentType() == "application" and \
          sipMsg.getContentSubType() == "sipfrag":
        # The body of this INFO contains a SIP fragment relayed from the callee
        # If it is the first line of a SIP response, relay the status code to
        # the Call Analyzer
        p = re.compile('SIP/2.0 (\d+) ')
        m = p.match(sipMsg.getBodyText())
        if not m is None:
          # TODO: Now we're only receiving INFO message to be informed that the
          #       callee answered the call. But it's not impossible in a near
          #       future to receive INFO message for other reason, then before
          #       setting the flag cpaCalleeAnswered we will have to ensure it's
          #       after the right INFO message.
          self.cpaCalleeAnswered = True
          
          self.media.handleOutOfBandEvent(int(m.group(1)))
          PyLoggingFunctions.log_debug(self.getLogger(), "CpaUASCall.py relayed CONNECT event to engine")

          self.media.switchToClassifier(PyCPAClassifierType(PyCPAClassifierType.Postconnect))
          PyLoggingFunctions.log_debug(self.getLogger(), "CpaUASCall.py requested switch to PostConnect classifier")

def test_module():
  cpaUASCall = CpaUASCall()
  cpaUASCall.getAnswerMessage()
  from paraxip.sipua.internal import SipCallEngineEvent   
  event = SipCallEngineEvent.CallToDialogAnswerEvent(None, None, None, PySipMessage())
  cpaUASCall.handleIncomingInfoMsg(event)
  PyCPAClassifierType(PyCPAClassifierType.Postconnect)

if __name__ == '__main__':
  test_module()
## !!MIRROR FILE, DO NOT MODIFY!!
