
## Logging
import paraxip.infra.PyLoggingFunctions as PyLoggingFunctions

## Our dependencies
import paraxip.sipua.internal.SipCallEngineEvent as SipCallEngineEvent
import paraxip.sipua.internal.SMTimeoutProcessor as SMTimeoutProcessor

## Python stdlib
import time

class TimerInfo:  
  """
  Information kept by the SMTimerManager on the different timers.
  Since the timers are global objects, this information is used as a way
  to link to the global objects and to make sure the timers are not forgotten.
  """
  
  def __init__(self, in_strTimerTag, in_uiTimerId, in_fTimeRef):
    #The name of the timer, (ex.: 'PreConnect')
    self.strTimerTag = in_strTimerTag
    
    #Unique? timer ID given by the global timer class
    self.uiTimerID = in_uiTimerId
    
    #Time at which the timer was started
    self.fTimeRef = in_fTimeRef
  
  def __str__(self):
    return "%s (TimerID:%s) started at: %s" % \
                  (self.strTimerTag, self.uiTimerID, self.fTimeRef)

class SMTimerManager:
  """
  State Machine Timer Manager
  Used by the states to keep track of the different timers used.
  """
  
  def __init__(self, in_stateMachine):
    self.__TimersInfo = []
    self.__LinkToSM = in_stateMachine
  
  def __del__(self):
    assert self.getNumTimers()==0
  
  def getNumTimers(self):
    return len(self.__TimersInfo)
  
  def getLogger(self):
    return self.__LinkToSM.getLogger()
  
  def startTimer(self, in_strTimerTag, in_TimeoutMs):
    """
    Adds a timer. The timer will be cancelled automatically (when this object 
    gets deleted).
    """
    PyLoggingFunctions.log_trace(self.getLogger(), "Starting timer: " + \
              in_strTimerTag + " with timeout:" + str(in_TimeoutMs))
    
    timeoutEvent = SipCallEngineEvent.TimeoutEvent()
    timeoutCancelledEvent = SipCallEngineEvent.TimeoutCancelledEvent()
    
    timeoutProcessor = SMTimeoutProcessor.SMTimeoutProcessor(
      self.__LinkToSM,
      timeoutEvent,
      timeoutCancelledEvent)
    
    uiTimerId = self.__LinkToSM.globalContext.scheduleTimer(
      timeoutProcessor, in_TimeoutMs, 0)
    
    timeoutEvent.timerId = uiTimerId
    timeoutCancelledEvent.timerId = uiTimerId
    
    # Save the time at which we start waiting for a cpa result in 
    # pre-connect
    fTimeRef = time.clock()
    
    thisTimerInfo = TimerInfo(in_strTimerTag, uiTimerId, fTimeRef)
    PyLoggingFunctions.log_debug(self.getLogger(), "Adding TimerInfo: %s, %s, %s." % \
                                          (in_strTimerTag, uiTimerId, fTimeRef))
    self.__TimersInfo.append(thisTimerInfo)
    PyLoggingFunctions.log_debug(self.getLogger(), "Started timer: " + \
              in_strTimerTag + " with timeout:" + str(in_TimeoutMs) + \
              "ms and TimerId:"+str(uiTimerId)
              )
  
  def getTimerInfoByName(self, in_strTimerTag):
    timerInfoToReturn = None
    for timerInfo in self.__TimersInfo:
      if timerInfo.strTimerTag == in_strTimerTag:
        assert timerInfoToReturn == None #Make sure no duplicate
        timerInfoToReturn = timerInfo
    return timerInfoToReturn    
  
  def getTimerInfoByID(self, in_uiID):
    timerInfoToReturn = None
    for timerInfo in self.__TimersInfo:
      if timerInfo.uiTimerID == in_uiID:
        assert timerInfoToReturn == None #Make sure no duplicate
        timerInfoToReturn = timerInfo
    return timerInfoToReturn   
  
  def removeTimerByID(self, in_uiTimerIDToRemove):
    objToRemove = None
    for timerInfo in self.__TimersInfo:
      if timerInfo.uiTimerID == in_uiTimerIDToRemove:
        assert objToRemove == None #Make sure no duplicate
        objToRemove = timerInfo
    
    self.__removeTimer(objToRemove)

  def removeTimerByName(self, in_strTimerTag):
    objToRemove = None
    for timerInfo in self.__TimersInfo:
      if timerInfo.strTimerTag == in_strTimerTag:
        assert objToRemove == None #Make sure no duplicate
        objToRemove = timerInfo
    
    self.__removeTimer(objToRemove, in_strTimerTag)

  def __removeTimer(self, in_timerInfo, in_str_CorrespondingTimerTag):
    """
    in_str_CorrespondingTimerTag is just there to debug in case no
    timer was found.
    """
    if in_timerInfo is not None:
      PyLoggingFunctions.log_debug(self.getLogger(), "Removing timer: " + \
                str(in_timerInfo) )
      self.__LinkToSM.globalContext.cancelTimer(in_timerInfo.uiTimerID)
      self.__TimersInfo.remove( in_timerInfo )
    else:
      PyLoggingFunctions.log_error(self.getLogger(), '%s cannot be removed, it does not exist.' %\
                                  in_str_CorrespondingTimerTag)

  def removeAllTimers(self):
    while len(self.__TimersInfo)> 0:
      timerInfo = self.__TimersInfo.pop()
      PyLoggingFunctions.log_debug(self.getLogger(), "Removing timer: " + str(timerInfo) )
      self.__LinkToSM.globalContext.cancelTimer(timerInfo.uiTimerID)

## !!MIRROR FILE, DO NOT MODIFY!!
