"""
Is tested in sip/hub/python/runtest.pl
"""

import random
import time

from paraxip.infra.PyLoggingFunctions import log_warn, Logger

class Helper:
  """This class allows one to:
    - Add predefined custom headers to a config dictionary.
    - Parse a SIP message and extract the custom values into
      a dictionary.
  """
  
  ##
  ## Private Data
  ##
  
  # The headers searched in the dictionary, listed in order of priority.
  # The first match is returned. If there is more than 1 match, a warning
  # is issued.
  # Paraxip's header NEEDS to be the first one in the list, the class expects
  # to retrieve it using __lststrTagCampaignID[0].
  
  __lststrTagCampaignID = ["X-Netborder-Cpa-Reference-ID", "X-Genesys-CallUUID"]
  __lststrTagCampaignName = ["X-Netborder-Cpa-Campaign-Name"]
  
  def __init__(self):
    self.__bHasWarned = False
    self.__CpaStatsLogger = Logger("cpa-stats")
  
  ##
  ## Public Functions
  ##
  
  def GetNotFoundMarker(self):
    return ""
  
  def GetCustomVendorCampaignID_FromDict(self, in_dict):
    """@param in_dict: A dictionary containing SIP custom fields and maybe a
                       custom vendor campaign ID.
     
    @return String of the ID or empty marker if not found.
    """
    
    return self.__PrioritySearch(self.__lststrTagCampaignID, in_dict)
  
  def GetCustomVendorCampaignName_FromDict(self, in_dict):
    """@param in_dict: A dictionary containing SIP custom fields and maybe a
                       custom vendor campaign name.
     
    @return String of the ID or empty marker if not found.
    """
    return self.__PrioritySearch(self.__lststrTagCampaignName, in_dict)
  
  def AddCampaignID_ToDict(self, in_strCampaignID, out_Dict):
    out_Dict[self.__GetParaxipTagCampaignID()] = in_strCampaignID
  
  def AddCampaignName_ToDict(self, in_strCampaignName, out_Dict):
    out_Dict[self.__GetParaxipTagCampaignName()] = in_strCampaignName
  
  #@tag: Extract from SIP message? ---> check if there is not a better way, seems backwards...
  #def
  
  ##
  ## Private Functions
  ##
  
  def __GetParaxipTagCampaignID(self):
    return self.__lststrTagCampaignID[0]
  
  def __GetParaxipTagCampaignName(self):
    return self.__lststrTagCampaignName[0]
  
  def __SearchDict(self, in_strKeyToSearch, in_dict):
    if in_dict.has_key(in_strKeyToSearch):
      return in_dict[in_strKeyToSearch]
    else:
      return self.GetNotFoundMarker()
  
  def __PrioritySearch(self, in_lststrHeaderToSearch, in_dictConfig):
    """Search all items from in_lststrHeaderToSearch in in_dictConfig.
    @return The first match found, self.GetNotFoundMarker() 
            if nothing found.
    Also has the responsibility to issue a warning if more 
    than one header is found.
    """
    
    matchesFound = []
    for in_lststrHeaderToSearch in in_lststrHeaderToSearch:
      tempMatch = self.__SearchDict(in_lststrHeaderToSearch, in_dictConfig)
      if tempMatch != self.GetNotFoundMarker():
        matchesFound.append(tempMatch)
    
    if len(matchesFound) == 0:
      return self.GetNotFoundMarker()
    else:
      if len(matchesFound) > 1:
        #Issue a warning, should not be more than 1 custom header
        self.__bHasWarned = True
        log_warn(self.__CpaStatsLogger,
                "More than one header found: " + str(matchesFound) +
                ", will return the first match.")
      
      #The FIRST match is returned (to respect priority order).
      return matchesFound[0]
  
  ##
  ## Test (public, then private)
  ##
  
  def Test(self):
    
    ## Tests with normal 
    
    headerDict = self.__GetExampleDict("Pixasam")
    assert self.GetCustomVendorCampaignName_FromDict(headerDict) != \
            self.GetNotFoundMarker()
    assert self.GetCustomVendorCampaignID_FromDict(headerDict) != \
            self.GetNotFoundMarker()
    
    headerDict = self.__GetExampleDict("Moxipagma", "abba1234321")
    assert self.GetCustomVendorCampaignName_FromDict(headerDict) == "Moxipagma"
    assert self.GetCustomVendorCampaignID_FromDict(headerDict) == "abba1234321"
    
    ## Tests with a custom header
    
    customDict = {}
    customDict['CPA-Threshold'] = 'CPA_HUMAN=0.7;global=0.9'
    customDict['NetBorder-Call-Id'] = '1256242418-582000-30249-0.cpa'
    customDict[self.__lststrTagCampaignID[1]] = 'OhACustomMerHeader'
    customDict['CustomParameter_Answer'] = '42'
    assert self.GetCustomVendorCampaignID_FromDict(customDict)\
      == 'OhACustomMerHeader'
    customDict = {}
    customDict['InvalidHeader'] = 'OhACustomMerHeaderThatIsNotValid'
    assert self.GetCustomVendorCampaignID_FromDict(customDict)\
      == self.GetNotFoundMarker()
    
    ## Test that the class issues a warning when more than 1 header matches
    
    customDict = {}
    customDict['CPA-Threshold'] = 'CPA_HUMAN=0.7;global=0.9'
    customDict['NetBorder-Call-Id'] = '1256242418-582000-30249-0.cpa'
    customDict[self.__lststrTagCampaignID[1]] = 'OhACustomMerHeader'
    customDict[self.__lststrTagCampaignID[0]] = 'ParaxipHeader'
    customDict['CustomParameter_Answer'] = '42'
    #Paraxip header should have priority
    assert self.GetCustomVendorCampaignID_FromDict(customDict)\
      == 'ParaxipHeader'
    assert self.__bHasWarned
  
  def __GetRandomCallID(self):
    random.seed(time.time())
    return random.randint(1000000,9999999)
  
  def __GetExampleDict(self, in_CampaignName, in_CampaignID = __GetRandomCallID):
    sampleDict = {}
    sampleDict['CPA-Threshold'] = 'CPA_HUMAN=0.7;global=0.9'
    sampleDict['NetBorder-Call-Id'] = '1256242418-582000-30249-0.cpa'
    sampleDict[self.__GetParaxipTagCampaignID()] = in_CampaignID
    sampleDict[self.__GetParaxipTagCampaignName()] = in_CampaignName
    sampleDict['CustomParameter_Answer'] = '42'
    return sampleDict

if __name__ == "__main__":
  vendorHelper = Helper()
  vendorHelper.Test()
  









## !!MIRROR FILE, DO NOT MODIFY!!
