#
#  @file   RoutingConfig.py
#  @author Francis Moreau
#  date    Mon June 23 2008
#  
#  Description:
#
#  Type your description here
#
#  Copyright (C) 2008 Paraxip Technologies.
#  All Rights Reserved.
#

import paraxip.infra.PyLoggingFunctions as PyLoggingFunctions
from paraxip.infra.Assert import *
import netborderPyInfra
from SipCallEngineExtending import PySipMessage
from cStringIO import StringIO
import re # Regular Expressions
from RoutingConfigEntry import *

fileScopeLogger = netborderPyInfra.Logger(
  "netborder.sip-router.RoutingConfig.py")

# Force declaration of class attributes in __init__ method
__pychecker__ = "initattr"

"""
The purpose of this configuration file is to give the end-user the ability to
configure NetBorder SIP Router load-balancing and failover features. This file
must be modified by the end-user according to is preferences and network
topology.

The user's specific parameters to change are included in RoutingConfig class. A
unique instance of the RoutingConfig class is instanciated by the RoutingEngine
class, that it's instanciated by the Application class. The RoutingConfig class
contains two main methods; initialize and routeInviteRequest. The
RoutingConfig's initialize method is called once and creates the default
configuration class member named defaultConfigEntries. This default
configuration is created according to the preferences and the network topology
of the end-user. The RoutingConfig's routeInviteRequest method is called every
time a new SIP INVITE message is received, the default behaviour of this
method is to return the defaultConfigEntries (the routing strategy to use by
the RoutingEngine). In very specific routing conditions where the default
routing is not enough it's always possible to create and return a new list of
RoutingConfigEntry entries to be used by the RoutingEngine from the
RoutingConfig's routeInviteRequest method. These specific routing conditions
could be based on the time of day, on a specific header included in the SIP
INVITE message, and so on.

The definition of RoutingConfigEntry is in the code section below, but here
is an enumeration with a brief description of the required and optional
parameters of this class :

RoutingConfigEntry(in_szUniqueName,
                   in_listRouteToServers,
                   in_uiWeight = 0,
                   in_uiPriority = 0,
                   in_inviteMatcher = None,
                   in_bRouteInParallel = False,
                   in_uiTimeoutMs = 5000)

The highest priority for in_uiPriority is 0. It's also important here to
understand how and when the RoutingEngine passes from a priority to another.
The RoutingEngine first looks at all the RoutingConfigEntry entries with the
lowest in_uiPriority and where their in_inviteMatcher parameter matches
the SIP INVITE message's header (e.g. request-uri, From header, To header),
and if there is no match then it passes to the next priority. Otherwise, if
there is at least one match then the RoutingEngine tries to establish a SIP
dialog with this one, if the dialog establishment fails or the in_uiTimeoutMs
is reached then the RoutingEngine passes to the next priority.

As for the priority the weight works on multiple entries of RoutingConfigEntry.
The weight is based on all the RoutingConfigEntry entries where the
in_uiPriority is the same and where their in_inviteMatcher parameter matchs
the SIP INVITE message.

Please visit this Web site for more details about Python regular
expression : http://docs.python.org/lib/module-re.html

Following are a few examples illustrating the possibilities and the flexibility
of NetBorder SIP Router regarding the load-balancing and the failover.

Example 1 => Distribute equally the traffic on 2 gateways. If "Server A" is
             selected and the call establishment fails then the "Server B" will
             also be tried, and vice versa :

    self.defaultConfigEntries = \
    [
      RoutingConfigEntry("Server A", [RouteTo("a1.domain.com")], 50),
      RoutingConfigEntry("Server B", [RouteTo("b1.domain.com")], 50)
    ]

Example 2 => Distribute the traffic on 2 gateways of different capacity, gw1 has
             1 T1 and gw2 has 4 T1 (different weights are used for this) :

    self.defaultConfigEntries = \
    [
      RoutingConfigEntry("GW 1", [RouteTo("gw1.domain.com")], 20),
      RoutingConfigEntry("GW 2", [RouteTo("gw2.domain.com")], 80)
    ]

Example 3 => Distribute the traffic on 2 gateways of different capacity, gw1 has
             1 T1 and gw2 has 4 T1, and use a backup gateway only if gw1 and gw2
             are not responding (different weights with different priorities
             are used for this) :

    self.defaultConfigEntries = \
    [
      RoutingConfigEntry("GW 1", [RouteTo("gw1.domain.com")], 20, 0),
      RoutingConfigEntry("GW 2", [RouteTo("gw2.domain.com")], 80, 0),
      RoutingConfigEntry("Backup GW", [RouteTo("backup.domain.com")], 0, 1)
    ]

Example 4 => Distribute the traffic on 2 groups of 2 gateways based on the
             request-uri, the first group handles the SIP INVITE requests where
             the request-uri's user part starts with 0, 1, 2, 3 or 4, while the
             second group handles the requests 5 to 9. There is another entry,
             with a priority of 1, that handles all the SIP INVITE requests where
             the request-uri's user part does not start with a digit [0-9].
             Finally, there is still another entry with a priority of 2 that will
             only be selected if all the others fail. For example, if the
             request-uri's user part starts with a 5 and gw3 and gw4 are not
             responding, the entry with the next priority that has a
             request-uri pattern matching will be tried ("All GWs" in this
             example). Since "All GWs" corresponds to gw4 and this one is not
             responding then the next entry will be tried, the "Backup GW" :

    self.defaultConfigEntries = \
    [
      RoutingConfigEntry("GW 1",
                         [RouteTo("gw1")],
                         20,
                         0,
                         RequestUriPattern("user", "^[0-4]")),
      RoutingConfigEntry("GW 2",
                         [RouteTo("gw2")],
                         80,
                         0,
                         RequestUriPattern("user", "^[0-4]")),
      RoutingConfigEntry("GW 3",
                         [RouteTo("gw3")],
                         20,
                         0,
                         RequestUriPattern("user", "^[5-9]")),
      RoutingConfigEntry("GW 4",
                         [RouteTo("gw4")],
                         80,
                         0,
                         RequestUriPattern("user", "^[5-9]")),
      RoutingConfigEntry("Default GW",
                         [RouteTo("gw4")],
                         0,
                         1),
      RoutingConfigEntry("Backup GW",
                         [RouteTo("backup.domain.com")],
                         0,
                         2)
    ]

Example 5 => Reject external calls starting with 9, otherwise distribute the
             traffic equally on gw1 and gw2 :

    self.defaultConfigEntries = \
    [
      RejectConfigEntry(603,
                        "Reject External Calls",
                        0,
                        RequestUriPattern("user", "^9")),
      RoutingConfigEntry("GW 1", [RouteTo("gw1")], 50, 1),
      RoutingConfigEntry("GW 2", [RouteTo("gw2")], 50, 1)
    ]
"""

class RoutingConfig:

  def __init__(self):
    __pychecker__ = 'unusednames=t'
    t = TraceScope(fileScopeLogger, "RoutingConfig.__init__")

    self.defaultConfigEntries = None

  @staticmethod
  def isConfigValid(in_routingConfigEntries):
    """
    Static method used by the RoutingEngine to validate an instance of
    RoutingConfig before using it.
    """
    if ((in_routingConfigEntries is None) or
        (not isinstance(in_routingConfigEntries, list)) or
        (len(in_routingConfigEntries) < 1)):
      PyLoggingFunctions.log_error(fileScopeLogger,
                "The list of Routing Configuration entries is empty")
      return False

    for entry in in_routingConfigEntries:
      if (not entry.isValid()):
        return False

    return True

  def initialize(self):
    __pychecker__ = 'unusednames=t'
    t = TraceScope(fileScopeLogger, "RoutingConfig.initialize")

    """
    USER CONFIGURATION: This configuration is a sample and must be modified to
                        represent the end-user preferences and network topology.
    """
    self.defaultConfigEntries = \
    [
      RejectConfigEntry(603, \
        'No routing configuration provided, see RoutingConfig.py')
    ]   

    return RoutingConfig.isConfigValid(self.defaultConfigEntries)

  def routeInviteRequest(self, in_inviteRequest):
    __pychecker__ = 'unusednames=t'
    t = TraceScope(fileScopeLogger, "RoutingConfig.routeInviteRequest")

    """
    USER CONFIGURATION: This method is called for every new SIP INVITE
                        message to give a chance to specify the routing
                        configuration for each new INVITE. By default, the
                        configuration used is the defaultConfigEntries
                        configured above.

                        The contain of the list defaultConfigEntries must never
                        be modified upon creation since it's used by all the
                        active RoutingSession(s) for which this method has
                        returned it.

    Example 1 => Reject all calls after 18h (getTime is sample function name):

      time = getTime() 
      if (time >= 18):
        return [RejectConfigEntry(603, "Reject Calls after 18h")]
      else:
        return self.defaultConfigEntries

    Example 2 => Load balance the traffic following a round robin pattern between 
                  2 different configurations :

      self._countCallRequested++
      if ((self._countCallRequested % 2) == 0):
        # Configuration to use when self._countCallRequested is even
        return self.defaultConfigEntries_Even
      else:
        # Configuration to use when self._countCallRequested is odd
        return self.defaultConfigEntries_Odd
    """
    return self.defaultConfigEntries

## !!MIRROR FILE, DO NOT MODIFY!!
