#
#  @file   RoutingConfigEntry.py
#  @author Dominic Lavoie
#  date    Mon July 7 2008
#  
#  Description:
#
#  Type your description here
#
#  Copyright (C) 2008 Paraxip Technologies.
#  All Rights Reserved.
#

import paraxip.infra.PyLoggingFunctions as PyLoggingFunctions
from paraxip.infra.Assert import *
import netborderPyInfra
from SipCallEngineExtending import PySipMessage
from cStringIO import StringIO
import re # Regular Expressions

fileScopeLogger = netborderPyInfra.Logger(
    "netborder.sip-router.RoutingConfigEntry.py")

# Force declaration of class attributes in __init__ method
__pychecker__ = "initattr"

class RouteTo:
  """
  This class contains the details about a specific routing destination used
  by RoutingConfigEntry.
  
  in_szLocalOutboundProxyUrl must be a complete SIP URL (ex. 'sip:127.0.0.1:5062')
  """
  def __init__(self,
               in_szServerAddress,
               in_uiServerPort = 5060,
               in_szRealmCredential = None,
               in_szUserNameCredential = None,
               in_szPasswordCredential = None,
               in_szLocalOutboundProxyUrl = None):
    # The address of the server to contact
    self.serverAddress = in_szServerAddress

    # The port of the server to contact
    self.serverPort = in_uiServerPort

    # The realm field when credentials are required on the server to contact
    self.realmCredential = in_szRealmCredential

    # The user name field when credentials are required on the server to contact
    self.userNameCredential = in_szUserNameCredential

    # The password field when credentials are required on the server to contact
    self.passwordCredential = in_szPasswordCredential

    # The Url (address:port) of the Local Outbound Proxy to use when making out 
    #   calls (default: None)
    self.localOutboundProxy = in_szLocalOutboundProxyUrl

    self._bIsValid = None

  def isValid(self):
    if (self._bIsValid is not None):
      return self._bIsValid

    self._bIsValid = True

    if (not isinstance(self.serverAddress, str)):
      PyLoggingFunctions.log_error(fileScopeLogger,
                "Setting of Server Address in RouteTo must be a string")
      self._bIsValid = False
    elif ((not isinstance(self.serverPort, int)) or (self.serverPort <= 0)):
      PyLoggingFunctions.log_error(fileScopeLogger,
                "Setting of Server Port in RouteTo must be greater than 0")
      self._bIsValid = False
    elif ((self.realmCredential is not None) and
          (not isinstance(self.realmCredential, str))):
      PyLoggingFunctions.log_error(fileScopeLogger,
                "Setting of Realm in RouteTo must be a string")
      self._bIsValid = False
    elif ((self.userNameCredential is not None) and
          (not isinstance(self.userNameCredential, str))):
      PyLoggingFunctions.log_error(fileScopeLogger,
                "Setting of User Name in RouteTo must be a string")
      self._bIsValid = False
    elif ((self.passwordCredential is not None) and
          (not isinstance(self.passwordCredential, str))):
      PyLoggingFunctions.log_error(fileScopeLogger,
                "Setting of Password in RouteTo must be a string")
      self._bIsValid = False
    elif ((self.localOutboundProxy is not None) and
          (not isinstance(self.localOutboundProxy, str))):
      PyLoggingFunctions.log_error(fileScopeLogger,
                "Setting of Local Outbound Proxy in RouteTo must be a string")
      self._bIsValid = False

    if (self._bIsValid):
      if (self.realmCredential is not None):
        # TODO: Bug 1393: not implemented yet
        PyLoggingFunctions.log_warn(fileScopeLogger, 'Setting of a Realm credential in RouteTo '
          'not implemented: provided Realm will be ignored')
  
      if (self.userNameCredential is not None):
        # TODO: Bug 1393: not implemented yet
        PyLoggingFunctions.log_warn(fileScopeLogger, 'Setting of a User Name credential in RouteTo '
          'not implemented: provided User Name will be ignored')
  
      if (self.passwordCredential is not None):
        # TODO: Bug 1393: not implemented yet
        PyLoggingFunctions.log_warn(fileScopeLogger, 'Setting of a Password credential in RouteTo '
          'not implemented: provided Password will be ignored')

    return self._bIsValid

  def __str__(self):
    strIOResult = StringIO()

    strIOResult.write("RouteTo members:\n")

    strIOResult.write("Server Address = %s\n" % (self.serverAddress))
    strIOResult.write("Server Port = %s\n" % (self.serverPort))
    strIOResult.write("Realm = %s\n" % (self.realmCredential))
    strIOResult.write("User Name = %s\n" % (self.userNameCredential))
    strIOResult.write("Password = %s\n" % (self.passwordCredential))
    strIOResult.write("Local Outbound Proxy = %s\n" % (self.localOutboundProxy))

    strResult = strIOResult.getvalue()
    strIOResult.close()
    return strResult

# Thin interface for objects that determine if an INVITE request matches or not
class InviteMsgMatcher:

  # To implement in the derived class. Returns True if the message matches
  def matchInvite(self, in_inviteMsg):
    return False
    
class PatternInviteMsgMatcher(InviteMsgMatcher):
  def __init__(self,
               in_rePatternToMatch = ".*",
               in_bIgnoreCase = True):
    # A regular expression matching the request-uri or a specific header of the
    # initial SIP INVITE, ".*" means everything is matching.
    self.rePatternToMatch = in_rePatternToMatch

    # True for ignoring case sensitive.
    self.ignoreCase = in_bIgnoreCase

    self.compiledRe = None
    if (self.ignoreCase):
      self.compiledRe = re.compile(in_rePatternToMatch, re.IGNORECASE)
    else:
      self.compiledRe = re.compile(in_rePatternToMatch)

    self._bIsValid = None

  def isValid(self):
    if (self._bIsValid is not None):
      return self._bIsValid

    self._bIsValid = True

    if (not isinstance(self.rePatternToMatch, str)):
      PyLoggingFunctions.log_error(fileScopeLogger,
                "Setting of Pattern to Match in Config Entry must be a string")
      self._bIsValid = False
    elif (not isinstance(self.ignoreCase, bool)):
      PyLoggingFunctions.log_error(fileScopeLogger,
                "Setting of Ignore Case in Config Entry must be a boolean")
      self._bIsValid = False

    return self._bIsValid

  def matchString(self, in_strToLook):
    szFcn = "PatternInviteMsgMatcher.matchString"
    t = TraceScope(fileScopeLogger, szFcn)
    reSearched = self.compiledRe.search(in_strToLook)
    PyLoggingFunctions.log_debug(fileScopeLogger, "%s : rePatternToMatch=%s, in_strToLook= %s, "
              "reSearched=%s, reSearched.group()=%s",
              szFcn, self.rePatternToMatch,
              in_strToLook,
              reSearched)
    return (reSearched is not None)

  # To implement in the derived class, probably using matchString
  def matchInvite(self, in_inviteMsg):
    return False

  def __str__(self):
    strIOResult = StringIO()

    strIOResult.write("Pattern To Match = %s\n" % (self.rePatternToMatch))
    strIOResult.write("Ignore Case = %s\n" % (self.ignoreCase))

    strResult = strIOResult.getvalue()
    strIOResult.close()
    return strResult

class RequestUriPattern(PatternInviteMsgMatcher):
  """
  An instance of this class is created and passed to RoutingConfigEntry when
  a pattern must be matched from the initial SIP INVITE's request-uri.

  Please visit this Web site for more details about Python regular
  expression : http://docs.python.org/lib/module-re.html
  """
  def __init__(self,
               in_szRequestUriSection = "all",
               in_rePatternToMatch = ".*",
               in_bIgnoreCase = True):
    PatternInviteMsgMatcher.__init__(self, in_rePatternToMatch, in_bIgnoreCase)

    # Specify the section of the request-uri where in_rePatternToMatch must be
    # applied. The section must be one of these keywords :
    # 
    # "all" => search in all the request-uri
    # "scheme" => search in the scheme section of the request-uri
    # "user" => search in the user part of the request-uri
    # "host" => search in the host section of the request-uri
    # "port" => search in the port section of the request-uri
    # "protocol" => search in the protocol section of the request-uri
    # "user-params" => search in the user defined parameters section of the
    #                  request-uri
    self.requestUriSection = in_szRequestUriSection

  def isValid(self):
    if (self._bIsValid is not None):
      return self._bIsValid

    if (not PatternInviteMsgMatcher.isValid(self)):
      self._bIsValid = False
    elif (not isinstance(self.requestUriSection, str)):
      PyLoggingFunctions.log_error(fileScopeLogger,
                "Setting of Request Uri Section of the Pattern to Match in "
                "Config Entry must be a string")
      self._bIsValid = False

    return self._bIsValid

  def __str__(self):
    strIOResult = StringIO()

    strIOResult.write("RequestUriPattern members:\n")

    strIOResult.write(PatternInviteMsgMatcher.__str__(self))
    strIOResult.write("Request Uri Section = %s\n" % (self.requestUriSection))

    strResult = strIOResult.getvalue()
    strIOResult.close()
    return strResult

  def matchInvite(self, in_inviteRequest):
    if (self.requestUriSection == "all"):
      return self.matchString(in_inviteRequest.getRequestLine())
    elif (self.requestUriSection == "user"):
      return self.matchString(in_inviteRequest.getUserPart("INVITE"))
    elif (self.requestUriSection == "scheme" or
          self.requestUriSection == "host" or
          self.requestUriSection == "port" or
          self.requestUriSection == "protocol" or
          self.requestUriSection == "user-params"):
      # TODO: Not supported yet
      PyLoggingFunctions.log_warn(fileScopeLogger,
               "Only \"all\" and \"user\" are supported on "
               "RequestUriPattern is supported for now")
      return False
    else:
      PyLoggingFunctions.log_warn(fileScopeLogger,
               "Unsupported URI section \"%s\" in RequestUriPattern",
               self.requestUriSection)
      return False
    return False


class FromHeaderPattern(PatternInviteMsgMatcher):
  """
  An instance of this class is created and passed to RoutingConfigEntry when
  a pattern must be matched from the initial SIP INVITE's From header.

  Please visit this Web site for more details about Python regular
  expression : http://docs.python.org/lib/module-re.html
  """
  def __init__(self,
               in_rePatternToMatch = ".*",
               in_bIgnoreCase = True):
    PatternInviteMsgMatcher.__init__(self, in_rePatternToMatch, in_bIgnoreCase)

  def isValid(self):
    if (self._bIsValid is not None):
      return self._bIsValid

    if (not PatternInviteMsgMatcher.isValid(self)):
      self._bIsValid = False

    return self._bIsValid

  def matchInvite(self, in_inviteRequest):
    return self.matchString(in_inviteRequest.getFromHeaderUri())

  def __str__(self):
    strIOResult = StringIO()

    strIOResult.write("FromHeaderPattern members:\n")

    strIOResult.write(PatternInviteMsgMatcher.__str__(self))

    strResult = strIOResult.getvalue()
    strIOResult.close()
    return strResult

class ToHeaderPattern(PatternInviteMsgMatcher):
  """
  An instance of this class is created and passed to RoutingConfigEntry when
  a pattern must be matched from the initial SIP INVITE's To header.

  Please visit this Web site for more details about Python regular
  expression : http://docs.python.org/lib/module-re.html
  """
  def __init__(self,
               in_rePatternToMatch = ".*",
               in_bIgnoreCase = True):
    PatternInviteMsgMatcher.__init__(self, in_rePatternToMatch, in_bIgnoreCase)

  def isValid(self):
    if (self._bIsValid is not None):
      return self._bIsValid

    if (not PatternInviteMsgMatcher.isValid(self)):
      self._bIsValid = False

    return self._bIsValid

  def matchInvite(in_inviteRequest):
    return self.matchString(in_inviteRequest.getToHeaderUri())

  def __str__(self):
    strIOResult = StringIO()

    strIOResult.write("ToHeaderPattern members:\n")

    strIOResult.write(PatternInviteMsgMatcher.__str__(self))
    
    strResult = strIOResult.getvalue()
    strIOResult.close()
    return strResult

# Interface base class for the various RoutingConfigEntry implementations
class RoutingConfigEntryInterface(InviteMsgMatcher):

  def getPriority(self):
    return 0

  def isValid(self):
    return False

  # InviteMsgMatcher interface
  # To implement in the derived class
  def matchInvite(self, in_inviteRequest):
    return False

class RoutingConfigEntryPatternToMatchImpl(RoutingConfigEntryInterface):

  def __init__(self,
               in_uiPriority = 0,
               in_inviteMatcher = None):
    # The priority associated to this entry, 0 is the highest
    # priority (this is used to give a priority to an entry
    # relative to other entries, and for entries where
    # inviteMatcher matched).
    self.priority = in_uiPriority

    # An instance of the pattern class (e.g. request-uri, From
    # header, To header) to match with the initial SIP INVITE.
    self.inviteMatcher = in_inviteMatcher

    self._bIsValid = None

  def getPriority(self):
    return self.priority

  def isValid(self):
    if (self._bIsValid is not None):
      return self._bIsValid

    self._bIsValid = True

    if ((not isinstance(self.priority, int)) or (self.priority < 0)):
      PyLoggingFunctions.log_error(fileScopeLogger,
                "Setting of Priority in Config Entry must be greater or equal "
                "than 0")
      self._bIsValid = False
    elif ((self.inviteMatcher is not None) and
          (not issubclass(self.inviteMatcher.__class__,
                          PatternInviteMsgMatcher))):
      PyLoggingFunctions.log_error(fileScopeLogger,
                "Setting of Invite Matcher in Config Entry is invalid")
      self._bIsValid = False
    elif ((self.inviteMatcher is not None) and
          (not self.inviteMatcher.isValid())):
      self._bIsValid = False

    return self._bIsValid

  def matchInvite(self, in_inviteRequest):
    t = TraceScope(fileScopeLogger,
                   "RoutingConfigEntryPatternToMatchImpl.matchInvite")
    bRv = True
    if (self.inviteMatcher is not None):
        bRv = self.inviteMatcher.matchInvite(in_inviteRequest)
    PyLoggingFunctions.log_debug(fileScopeLogger, "matchInvite = %i", bRv)
    return bRv

class RoutingConfigEntry(RoutingConfigEntryPatternToMatchImpl):

  """
  This class contains all the details about an entry among others to be able
  to offer the load-balancing and the failover within the end-user's network
  topology.
  """
  def __init__(self,
               in_szUniqueName,
               in_listRouteToServers,
               in_uiWeight = 0,
               in_uiPriority = 0,
               in_inviteMatcher = None,
               in_bRouteInParallel = False,
               in_uiTimeoutMs = 5000):

    RoutingConfigEntryPatternToMatchImpl.__init__(self, in_uiPriority,
                                                  in_inviteMatcher)
    
    # All entries must have a unique name.
    self.uniqueName = in_szUniqueName

    # The list of RouteTo class's instance to try when
    # invietMatcher member matchs the request-uri of the
    # initial SIP INVITE.
    self.routeToServers = in_listRouteToServers

    # The weight associated to this entry, 100 means than 100% will be routed
    # to this entry (this is used to give some weight to an entry relative to
    # other entries, and for entries where inviteMatcher matched).
    self.weight = in_uiWeight
    
    # When set to True the list of servers in routeToServers will be tried in
    # parallel, otherwise the list of servers will be tried sequentially.
    self.routeInParallel = in_bRouteInParallel

    # The timeout in milliseconds allocated for connecting with one server
    # from the list of servers.
    self.timeout = in_uiTimeoutMs

  def _isRouteToServersValid(self):
    if ((not isinstance(self.routeToServers, list)) or
        (len(self.routeToServers) < 1)):
      PyLoggingFunctions.log_error(fileScopeLogger,
                "Setting of RouteTo Servers in RoutingConfigEntry is empty")
      return False

    for server in self.routeToServers:
      if (not isinstance(server, RouteTo)):
        PyLoggingFunctions.log_error(fileScopeLogger,
                  "Setting of RouteTo Servers in RoutingConfigEntry contains "
                  "an invalid entry")
        return False
      elif (not server.isValid()):
        return False

    return True

  def isValid(self):
    if (self._bIsValid is not None):
      return self._bIsValid

    if (not RoutingConfigEntryPatternToMatchImpl.isValid(self)):
      self._bIsValid = False
    elif (not isinstance(self.uniqueName, str)):
      PyLoggingFunctions.log_error(fileScopeLogger,
                "Setting of Unique Name in RoutingConfigEntry must be a string")
      self._bIsValid = False
    elif (not self._isRouteToServersValid()):
      self._bIsValid = False
    elif ((not isinstance(self.weight, int)) or (self.weight < 0)):
      PyLoggingFunctions.log_error(fileScopeLogger,
                "Setting of Weight in RoutingConfigEntry must be greater or "
                "equal than 0")
      self._bIsValid = False
    elif (not isinstance(self.routeInParallel, bool)):
      PyLoggingFunctions.log_error(fileScopeLogger,
                "Setting of Route In Parallel in RoutingConfigEntry must be a "
                "boolean")
      self._bIsValid = False
    elif ((not isinstance(self.timeout, int)) or (self.timeout <= 0)):
      PyLoggingFunctions.log_error(fileScopeLogger,
                "Setting of Timeout in RoutingConfigEntry must be greater "
                "than 0")
      self._bIsValid = False

    return self._bIsValid

  def __str__(self):
    strIOResult = StringIO()

    strIOResult.write("RoutingConfigEntry members:\n")

    strIOResult.write("Unique Name = %s\n" % (self.uniqueName))
    if ((self.routeToServers is not None) and \
        (isinstance(self.routeToServers, list))):
      for server in self.routeToServers:
        strIOResult.write("%s" % (server))
    strIOResult.write("Weight = %s\n" % (self.weight))
    strIOResult.write("Priority = %s\n" % (self.priority))
    strIOResult.write("InviteMatcher = %s\n" % (self.inviteMatcher))
    strIOResult.write("Route in Parallel = %s\n" % (self.routeInParallel))
    strIOResult.write("Timeout in ms = %s\n" % (self.timeout))

    strResult = strIOResult.getvalue()
    strIOResult.close()
    return strResult

class RejectConfigEntry(RoutingConfigEntryPatternToMatchImpl):

  def __init__(self,
               in_uiSipErrorCode = 603,
               in_szSipErrorDesc = "Declined by Routing Configuration",
               in_uiPriority = 0,
               in_inviteMatcher = None):

    RoutingConfigEntryPatternToMatchImpl.__init__(self, in_uiPriority,
                                                  in_inviteMatcher)

    # The SIP error code to take for rejecting the incoming SIP
    # INVITE message.
    self.sipErrorCode = in_uiSipErrorCode

    # The SIP error description to use when rejecting an incoming SIP INVITE
    # message.
    self.sipErrorDesc = in_szSipErrorDesc

  def isValid(self):
    if (self._bIsValid is not None):
      return self._bIsValid

    if (not RoutingConfigEntryPatternToMatchImpl.isValid(self)):
      self._bIsValid = False
    elif ((not isinstance(self.sipErrorCode, int)) or (self.sipErrorCode <= 0)):
      PyLoggingFunctions.log_error(fileScopeLogger,
                "Setting of SIP Error code in RejectConfigEntry must be "
                "greater than 0")
      self._bIsValid = False
    elif (not isinstance(self.sipErrorDesc, str)):
      PyLoggingFunctions.log_error(fileScopeLogger,
                "Setting of SIP Error Description in RejectConfigEntry must "
                "be a string")
      self._bIsValid = False

    return self._bIsValid

  def __str__(self):
    strIOResult = StringIO()

    strIOResult.write("RejectConfigEntry members:\n")

    strIOResult.write("SIP Error Code = %s\n" % (self.sipErrorCode))
    strIOResult.write("SIP Error Description = %s\n" % (self.sipErrorDesc))
    strIOResult.write("Priority = %s\n" % (self.priority))
    strIOResult.write("%s" % (self.inviteMatcher))
    
    strResult = strIOResult.getvalue()
    strIOResult.close()
    return strResult

## !!MIRROR FILE, DO NOT MODIFY!!
