#
#  @file   RoutingSession.py
#  @author Francis Moreau
#  date    Wed June 11 2008
#  
#  Description:
#
#  Type your description here
#
#  Copyright (C) 2008 Paraxip Technologies.
#  All Rights Reserved.
#

import paraxip.infra.PyLoggingFunctions as PyLoggingFunctions
from paraxip.infra.Assert import *
from netborderPyInfra import *
from sip_router.Route import Route
from sip_router.RouteSet import RouteSet
from sip_router.RoutingAction import *
from sip_router.RoutingConfigEntry import *
from cStringIO import StringIO
import random
import weakref

fileScopeLogger = netborderPyInfra.CallLogger(
  "netborder.sip-router.RoutingSession.py")

# Force declaration of class attributes in __init__ method
__pychecker__ = "initattr"

class RoutingFailureHistory:
  def __init__(self, in_bTimeoutReached=False, in_uiSipErrorCode=0):
    __pychecker__ = 'unusednames=t'
    t = TraceScope(fileScopeLogger, "RoutingFailureHistory.__init__")

    self.timeoutReached = in_bTimeoutReached
    self.sipErrorCode = in_uiSipErrorCode

class RoutingSession:
  def __init__(self, in_szCallId, in_inviteRequest, in_routingEngine):
    __pychecker__ = 'unusednames=t'
    t = TraceScope(fileScopeLogger, "RoutingSession.__init__")

    self._routingEngineWR = weakref.ref(in_routingEngine)

    self._strCallId = in_szCallId
    self._initialInvite = in_inviteRequest

    self._callLogger = CallLogger("netborder.sip-router.RoutingSession.py")
    # XXX where is callEnd() performed ?
    self._callLogger.callStart(self._strCallId)

    self._routeMethodHasBeenCalled = False

    self._listRoutingEntriesPerPriority = None
    self._uiCurPositionInRoutingEntries = 0

    # Hash table of RoutingFailureHistory result per destination server:port
    self._hashRoutingFailureHistory = {}

  def _get_routingEngine(self):
    return self._routingEngineWR()
  _routingEngine = property(_get_routingEngine, None)

  def route(self):
    """
    This method must be called one time at the beginning of the routing
    process. Then, onRouteTimeout and onRouteFailure are called to know the
    next routing step.
    """
    __pychecker__ = 'unusednames=t'
    t = TraceScope(fileScopeLogger, "RoutingSession.route")

    if (isLogAssert(self._callLogger, not self._routeMethodHasBeenCalled)):
      routingAction = self._rejectWithServerInternalError()
      self._logRoutingAction(routingAction)
      return routingAction

    self._routeMethodHasBeenCalled = True

    self._listRoutingEntriesPerPriority = \
      self._routingEngine.getApplicableRoutingEntriesPerPriority(
        self._initialInvite)

    # Log the list of routing configuration entries that match the SIP INVITE
    # of this session
    strList = None
    strIOList = StringIO()
    for priority in self._listRoutingEntriesPerPriority:
      for routingEntry in priority:
        strIOList.write("%s" % (routingEntry))
    strList = strIOList.getvalue()
    strIOList.close()
    PyLoggingFunctions.log_info(self._callLogger, "List of routing configuration entries that "
                               "match this SIP INVITE:\n%s", strList)

    return self._logGetNextRoutingAction()

  def onRouteTimeout(self, in_route):
    __pychecker__ = 'unusednames=t'
    t = TraceScope(fileScopeLogger, "RoutingSession.onRouteTimeout")

    if (isLogAssert(self._callLogger, self._routeMethodHasBeenCalled)):
      routingAction = self._rejectWithServerInternalError()
      self._logRoutingAction(routingAction)
      return routingAction

    PyLoggingFunctions.log_info(self._callLogger, "This Server timed out:\n%s", in_route)

    # TODO: Support parallel routing. In this case we should return
    #       ContinueRoutingAction and wait all the results before returning any
    #       other RoutingAction.

    self._updateRoutingFailureHistory(in_route, True)

    return self._logGetNextRoutingAction()

  def onRouteFailure(self, in_route, in_sipMessage):
    __pychecker__ = 'unusednames=t'
    t = TraceScope(fileScopeLogger, "RoutingSession.onRouteFailure")

    if (isLogAssert(self._callLogger, self._routeMethodHasBeenCalled)):
      routingAction = self._rejectWithServerInternalError()
      self._logRoutingAction(routingAction)
      return routingAction

    PyLoggingFunctions.log_info(self._callLogger,
             "This Server failed with a SIP error code of %d:\n%s",
             in_sipMessage.getStatusCode(),
             in_route)

    # TODO: Support parallel routing. In this case we should return
    #       ContinueRoutingAction and wait all the results before returning any
    #       other RoutingAction.

    if (in_sipMessage.getStatusCode() == 486): # Busy Here
      # Final decision, no more routing needed
      routingAction = AbandonRoutingAction(486, "Busy Here")
      self._logRoutingAction(routingAction)
      return routingAction
    else:
      self._updateRoutingFailureHistory(in_route, False, in_sipMessage)
      return self._logGetNextRoutingAction()

  def _rejectWithServerInternalError(self):
    __pychecker__ = 'unusednames=t'
    t = TraceScope(fileScopeLogger,
                   "RoutingSession._rejectWithServerInternalError")

    rejectConfigEntry = RejectConfigEntry(500, "Server Internal Error")
    return RejectRoutingAction(rejectConfigEntry)

  def _createNewResultRoutingAction(self, in_routingConfigEntry):
    __pychecker__ = 'unusednames=t'
    t = TraceScope(fileScopeLogger,
                   "RoutingSession._createNewResultRoutingAction")

    # TODO: Support parallel routing. For now take the first one.
    routeToConfig = in_routingConfigEntry.routeToServers[0]
    # Fill a RouteSet
    route = Route(routeToConfig, in_routingConfigEntry.timeout)
    routeSet = RouteSet()
    routeSet.addRoute(route)

    # Return a NewResultRoutingAction
    return NewResultRoutingAction(routeSet)

  def _logRoutingAction(self, in_routingAction):
    PyLoggingFunctions.log_info(self._callLogger,
             "RoutingSession returns this Routing Action:\n%s", in_routingAction)

  def _logGetNextRoutingAction(self):
    routingAction = self._getNextRoutingAction()
    self._logRoutingAction(routingAction)
    return routingAction

  def _getNextRoutingAction(self):
    __pychecker__ = 'unusednames=t'
    t = TraceScope(fileScopeLogger, "RoutingSession._getNextRoutingAction")
    if(len(self._listRoutingEntriesPerPriority) == 0):
      # Return a RejectRoutingAction
      # As specified in rfc 3261, section 16.5:
      #   "If the target set remains empty after applying all of the above, the
      #   proxy MUST return an error response, which SHOULD be the 480
      #   (Temporarily Unavailable) response."
      rejectConfigEntry = RejectConfigEntry(480, 
                         "Invite message does not match any RoutingConfigEntry")
      return RejectRoutingAction(rejectConfigEntry)

    # Check if we already been through _listRoutingEntriesPerPriority entries
    if (self._uiCurPositionInRoutingEntries >= \
        len(self._listRoutingEntriesPerPriority)):
      # Return a RejectRoutingAction
      rejectConfigEntry = RejectConfigEntry(480, 
                                            "No target has accepted the call")
      return RejectRoutingAction(rejectConfigEntry)

    # Get the list of routing entries for the current priority
    currentListRoutingEntries = \
      self._listRoutingEntriesPerPriority[self._uiCurPositionInRoutingEntries]
    if ((currentListRoutingEntries is None) or
        (len(currentListRoutingEntries) < 1)):
      # Return a RejectRoutingAction
      rejectConfigEntry = RejectConfigEntry(480, 
                                            "No target has accepted the call")
      return RejectRoutingAction(rejectConfigEntry)

    else:
      # There is at least one entry for this priority

      # First look if there is a RejectConfigEntry for the current priority
      for entry in currentListRoutingEntries:
        if (issubclass(entry.__class__, RejectConfigEntry)):
          # Return a RejectRoutingAction
          return RejectRoutingAction(entry)

      uiCumulativeWeight = 0

      # Load balance the traffic between the remaining entries with for the
      # current priority (the entries that have not already been tried)
      remainingRoutingEntries = []
      for entry in currentListRoutingEntries:
        if (issubclass(entry.__class__, RoutingConfigEntry)):
          # TODO: Support parallel routing. For now take the first one.
          routeToConfig = entry.routeToServers[0]
          if (not self._isServerExistInRoutingFailureHistory(routeToConfig)):
            remainingRoutingEntries.append(entry)
            # Count the cumulative weight of all remaining routing entries
            uiCumulativeWeight += entry.weight
        else:
          # TODO: PyLoggingFunctions.log_warn... invalid type...
          pass

      uiRemainingRoutingEntriesSize = len(remainingRoutingEntries)
      if (uiRemainingRoutingEntriesSize < 1):
        # All the servers for the current priority have already been tried
        # during this session, go to the next priority        
        self._uiCurPositionInRoutingEntries += 1
        PyLoggingFunctions.log_info(self._callLogger,
                 "All the targets for the current priority (%d) have been tried"
                 ", trying next priority (if any)" ,
                  currentListRoutingEntries[0].priority)
        # Recursive call
        return self._getNextRoutingAction()

      elif (uiRemainingRoutingEntriesSize == 1):
        # There is only one entry available for the current priority
        return self._createNewResultRoutingAction(remainingRoutingEntries[0])

      else:            
        PyLoggingFunctions.log_info(self._callLogger,
                  "Load balancing between priority %d RoutingConfigEntry: %s",
                  remainingRoutingEntries[0].priority,
                  self._getListOfEntryName(remainingRoutingEntries))
             
        # There are multiple entries available for the current priority

        # Here is how the weight is respected with a random value:
        #
        # Let's say there are 3 routing configuration entries; one with a weight
        # of 30, a second one with a weight of 65, and a third one with 5.
        # Therefore, the first one handles the random numbers from 1 to 30
        # inclusively, the second one from 31 to 95 still inclusively, and the
        # last one from 96 to 100. This way the sum of weight of routing
        # configuration entries with the same priority could be smaller than
        # 100, or bigger too.
        if (uiCumulativeWeight < 1):
          PyLoggingFunctions.log_warn(self._callLogger, "Invalid weight distribution for "
                   "RoutingConfigEntry: %s", 
                   self._getListOfEntryName(remainingRoutingEntries))
          return self._rejectWithServerInternalError()
        randomWeight = random.randint(1, uiCumulativeWeight)

        # All the remaining entries for the current priority are of
        # RoutingConfigEntry type
        currentWeightBeginningRange = 1
        for entry in remainingRoutingEntries:
          if ((currentWeightBeginningRange + entry.weight) > randomWeight):
            return self._createNewResultRoutingAction(entry)
          currentWeightBeginningRange += entry.weight

    # TODO: logAssert...
    return self._rejectWithServerInternalError()

  def _updateRoutingFailureHistory(self,
                                   in_route,
                                   in_bTimeoutReached=False,
                                   in_sipMessage=None):
    __pychecker__ = 'unusednames=t'
    t = TraceScope(fileScopeLogger,
                   "RoutingSession._updateRoutingFailureHistory")

    strDestServerKey = \
      self._getRoutingFailureHistoryKey(in_route.getRouteToConfig())

    logAssert(self._callLogger,
              strDestServerKey not in self._hashRoutingFailureHistory)

    if (in_bTimeoutReached):
      self._hashRoutingFailureHistory[strDestServerKey] = \
        RoutingFailureHistory(True)
    else:
      self._hashRoutingFailureHistory[strDestServerKey] = \
        RoutingFailureHistory(False, in_sipMessage.getStatusCode())

  def _isServerExistInRoutingFailureHistory(self, in_routeToConfig):
    __pychecker__ = 'unusednames=t'
    t = TraceScope(fileScopeLogger,
                   "RoutingSession._isServerExistInRoutingFailureHistory")

    return self._getRoutingFailureHistoryKey(in_routeToConfig) in \
           self._hashRoutingFailureHistory

  def _getRoutingFailureHistoryKey(self, in_routeToConfig):
    __pychecker__ = 'unusednames=t'
    t = TraceScope(fileScopeLogger,
                   "RoutingSession._getRoutingFailureHistoryKey")

    strIOKey = StringIO()
    strIOKey.write("%s:%s" % (in_routeToConfig.serverAddress,
                              in_routeToConfig.serverPort))
    strKey = strIOKey.getvalue()
    strIOKey.close()
    return strKey

  def _getListOfEntryName(self,in_routingEntryList):
    strListOfEntryName = ""
    for entry in in_routingEntryList:          
      if(entry != in_routingEntryList[0]):
        # Add a comma in front for all entries except the first one
        strListOfEntryName = strListOfEntryName + ","
      strListOfEntryName = strListOfEntryName + entry.uniqueName 
    return strListOfEntryName
## !!MIRROR FILE, DO NOT MODIFY!!
